/*
 * Conductivity.h
 *
 * Created on: 15 Apr 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * Architecture:
 *
 *  +-----------------+
 *  | Conductivity    |
 *  +-----------------+
 *  | readRaw()       |  readRaw() passed as function pointer to helper
 *  | formatString()  |  formatString() passed as function pointer to helper
 *  +-----------------+
 *  | helper          |  helper is instance of OnePointCalibrationHelper
 *  +-----------------+
 *           |
 *  +---------------------------+
 *  | OnePointCalibrationHelper |
 *  +---------------------------+
 *  | setCalibValues()          |  setCalibValues() saves raw and ideal values,
 *  |                           |                   calls back readRaw()
 *  | readString()              |  readString() calls back readRaw() and formatString(),
 *  |                           |               uses OnePointCalibration to calculate
 *  |                           |               calibrated value
 *  +---------------------------+
 *  | calibration               |  calibration is instance of OnePointCalibration
 *  +---------------------------+
 *               |
 *  +----------------------+
 *  | OnePointCalibration  |
 *  +----------------------+
 *  | calculateValue()     |  calculateValue() calculates calibrated and
 *  |                      |                   temperature compensated value
 *  +----------------------+
 *  | raw                  |  value measured in reference liquid during calibration
 *  | ideal                |  value from reference liquid (buffer solution)
 *  | calibTemp            |  temperature during calibration
 *  +----------------------+
 */

#ifndef CONDUCTIVITY_H_
#define CONDUCTIVITY_H_

#include "OnePointCalibrationHelper.h"


class Conductivity {
  public:
    // Init sensor
    static void begin();

    // Read conductivity from sensor
    static float readRaw();

    // Convert voltage value to temperature-compensated tds value
    static float readTdsValue(float temp);

    // Roughly calculate from TDS ppm to electrical conductivity with
    // 1 ppm (mg/l) = 2 µS/cm
    float calculateFromTdsValue(float temp);

    // Format conductivity and answer string with unit (= true) or not
    static String formatString(float conductivity, bool withUnit);

    // Helper with functions set up for conductivity
    static OnePointCalibrationHelper helper;

  private:
    static char buf[];

    // Answer value needed to calculate ecValue
    static float voltageFormula(float voltage);

    static int getMedianValue(int arr[], int size);
};

#endif /* CONDUCTIVITY_H_ */
