/*
 * Config.h
 *
 * Created on: 13 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 * 
 * ESP32 Pinout and connections                                                  
 *    USED ESP32 PINS                       PWR            SD                  RTC, LCD
 *                                     5V  HOLD   --   --  CS   --   --   --   ---I2C---
 *    +------------------------------------------------------------------------------------+
 *    |                       o    o    o    o    o    o    o    o    o    o    o    o     |
 *    |                      BAT  EN   USB  #13  #12  #27  #33  #15  #32  #14  SCL  SDA    |
 *    |                                     GPIO GPIO GPIO GPIO GPIO GPIO GPIO             |
 *    |                                                                                    |
 *    |                           ESP32 Feather Huzzah (Adafruit)                          |
 *    |                                                                                    |
 *    |                       GPIO GPIO GPI  GPI  GPI  GPIO                                |
 *    |                       #26  #25  #34  #39  #36  #4                            GPIO  |
 *    |   RST  3V   NC   GND  A0   A1   A2   A3   A4   A5   SCK  MO   MI   RX   TX   #21   |
 *    |    o    o    o    o    o    o    o    o    o    o    o    o    o    o    o    o    |
 *    +------------------------------------------------------------------------------------+
 *            3.3V       GND  --   --   Cond Turb LCD  Tch   ----SPI-----  Serial1   Temp
 *                                                Btn  Btn     SD card     option
 */

#ifndef CONFIG_H_
#define CONFIG_H_

#include <Arduino.h>
#include "WbPrintBuffer.h"
#include "Timeout.h"
#include "StopWatch.h"
#include "WbPreferences.h"


// Preferences key strings: do not change
#define TEMPERATURE_STR "temperature"
#define TEMP_RAW_LOW_STR "tempRawLow"
#define TEMP_RAW_HIGH_STR "tempRawHigh"
#define TEMP_IDEAL_LOW_STR "tempIdealLow"
#define TEMP_IDEAL_HIGH_STR "tempIdealHigh"
#define TEMP_LAST_DATE_STR "tempLastDate"
#define CONDUCTIVITY_STR "conductivity"
#define COND_RAW_STR "condRaw"
#define COND_IDEAL_STR "condIdeal"
#define COND_CALIB_TEMP_STR "condCalibTemp"
#define COND_LAST_DATE_STR "condLastDate"
#define TURBIDITY_STR "turbidity"
#define TURB_RAW_LOW_STR "turbRawLow"
#define TURB_RAW_HIGH_STR "turbRawHigh"
#define TURB_IDEAL_LOW_STR "turbIdealLow"
#define TURB_IDEAL_HIGH_STR "turbIdealHigh"
#define TURB_LAST_DATE_STR "turbLastDate"


class Config {
  public:
    static const String VERSION_NO;
    static const String VERSION;
    static const String INSTITUTION;

    // VARS
    // In- and out-redirection for Print objects
    // like Serial, Serial1, PrintBuffer
    static Stream* inStream;
    static Print* outStream;
    // PrintBuffer to build strings in buffer for webpage
    static WbPrintBuffer pb;
    // Stop watch used for development testing
    static StopWatch stopWatch;

    // CONSTANTS
    // Button, touch and power
    static const int BUTTON_PIN;
    static const int TOUCH_BUTTON_PIN;
    static const int POWER_HOLD_PIN;
 
    // Unused pins
    static const int UNUSED_A0_PIN;
    static const int UNUSED_A1_PIN;
    static const int UNUSED_27_PIN;
    static const int UNUSED_12_PIN;
    static const int UNUSED_14_PIN;
    static const int UNUSED_15_PIN;
    static const int UNUSED_32_PIN;

    // SPI [SD card]
    // Uses ESP32 SCK, MI, and MO pins: thus nothing to set
    // as library default values are used.
    // SD card chip select on ESP32 Feather Huzzah
    static const int SD_CS_PIN;

    // LCD PINS
    // Uses predefined ESP32 SDA and SCL pins: thus nothing to set.
    // I2C address for LCD display used from library
    // No I2C pull-ups needed as integrated.

    // I2C PINS [RTC: PCF8523]
    // Uses predefined ESP32 SDA and SCL pins: thus nothing to set.
    // I2C address for RTC used from library
    // No I2C pull-ups needed with Adalogger feather board as integrated.

    // SERIAL1 PINS
    // Uses predefined ESP32 TX and RX pins: thus nothing to set.

    // ONE_WIRE FOR TEMPERATURE
    // Data wire of DS18B20 (Digital Temperature Sensor)
    static const uint8_t TEMP_PIN;
    // CONDUCTIVITY ANALOG PIN
    static const int CONDUCTIVITY_PIN;
    // TURBIDITY ANALOG PIN
    static const int TURBIDITY_PIN;

    static const String WEB_TITLE;

    // Defines the logger's associated default logfile name
    // Format 8.3
    static const char DEFAULT_LOGFILE_NAME[];
    // Defines the lcd display mode logger's associated default logfile name
    static const String LCD_DEFAULT_LOGFILE_NAME;

    static const char DAYS_OF_WEEK[];

    // Logfile column headers
    static const String DATE_HEADER;
    static const String TIME_HEADER;
    static const String TEMP_HEADER;
    static const String CONDUCTIVITY_HEADER;
    static const String TURBIDITY_HEADER;

    // CONDUCTIVITY A/D-RANGE
    static const float CONDUCTIVITY_ADC_RANGE;
    // CONDUCTIVITY ADC AREF: 3.3 V on ESP32
    static const float CONDUCTIVITY_AREF;
    static const float CONDUCTIVITY_CALIB_DEFAULT_TEMP;

    // MONITOR AND LOGGER TIMER
    // Minimum interval in seconds for monitor, logger
    static const int TIMER_MIN_INTERVAL;

    // LCD TIMER
    // Interval in seconds for LCD display mode
    static const int LCD_TIMER_INTERVAL;

    // Maximum text length of textarea contents on website
    static const long MAX_RESPONSETEXT_LENGTH;

    // Define pin mode of pins
    static void setPinMode();
};

#endif /* CONFIG_H_ */
