/*
 * Interpreter.h
 *
 * Created on: 5 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * Input and Output are separately redirectable:
 *  Interpreter::setInputStream
 *  Interpreter::setOutputStream
 */

#ifndef INTERPRETER_H_
#define INTERPRETER_H_

// #define INTERPRETER_TESTING

#include "Config.h"
#include <Arduino.h>


// Up to MAX parameters types are possible.
// Now: 'i'  'f'  '"'
//  'i'  for integer
//  'f'  for float
//  '"'  for string with a closing '"'
#define MAX 3


/*
 * The interpreter expects command words. The format of the command with
 * parameters or without is as follows:
 *  <cmd>
 *  <cmd> s<param>
 *  <cmd> s<param1>s<param2>
 *  <cmd> s<param1>s<param2>s<param3>
 * Separators s in front of a <param> are used to identify the type of
 * the param.
 *  'i'  for integer
 *  'f'  for float
 *  '"'  for string with a closing '"'
 * Each part s<param> may be followed by blanks. At least one blank after
 *  <cmd> must separate a following parameter.
 *
 * Examples:
 *  help
 *  settime i12 i12 i10
 *  time
 *  echo "hello" "dolly"
 */
class Interpreter {
  public:
    // Array to store int parameter values.
    static int intParams[MAX];

    // Array to store float parameter values.
    static float floatParams[MAX];

    // Array to store String parameters
    static String stringParams[MAX];

	  // Interpret command with possible parameters.
    static void interpret();

    // Set command input to be interpreted by interpret()
    static void setInput(String commmand);

	  // Read command and interpret its.
    static void readAndInterpret();

    // Test if expected parameters are present.
    static bool expectedParams(char c0, char c1, char c2);

    // Set echoOn flag on or off
    static void setEcho(bool flag);

    // Test if echoOn flag is on
    static bool isEchoOn();

  private:
    // Initial command input
    static String input;

    // Initial command input used to echo to monitor
    static String echo;

    // Input string used for string separator detection of '"'
    // as with "mystring"
    static String changedInput;

    // Contains the command word.
    static String cmdWord;

    // Separator types
    static const char SEPARATOR_TYPES[];

    // Array to store separator types in encountered order.
    // '0' stands for no separator
    static char separators[MAX];

    // Array to store start-'"' locations of strings.
    static unsigned char startLocations[MAX];

    // Array to store locations of 'i' and 'f' separators and
    // end-'"' separators of strings.
    static unsigned char locations[MAX];

    // Flag whether to echo commandline on monitor or not
    static bool echoOn;

    // Print required parameter of the command.
    static void printRequiredParams(char c);

    /*
     * Find command word (the first string) till a blank or \n is
     * encountered.
     */
    static void findCommandWord();

    /*
     * Find all separators in String 'changedInput'. A maximum of MAX separators
     * are allowed.
     * Memorise the order of separators in array 'separators' using the
     * separator type symbols ('i', 'f' etc.) from array 'separatorTypes'.
     * In the case of string-type ('"') the second location (end separator
     * location) is memorised in array 'locations'.
     *
     * Example1:
     * input: <command> f1.1"abc"d22
     * separators:     ['f', '"', 'i']  -> found and order of separators
     * locations:      [  1    9   10]  -> end-'"' found at location 9,
     *                                     'f' found at location 1,
     *                                     'i' found at location 10
     * Example2:
     * input: <cmd> "abc" i10
     * separators:     ['"', 'i', '0']  -> found and order of separators
     * locations:      [  5    7  255]  -> end-'"' found at location 5,
     *                                     'i' found at location 7
     *
     * value: 255 = unsigned char = |-1|: location irrelevant
     */
    static void findSeparators();

    /*
     * Find missing start-'"' (now '?') separator locations and
     * memorise them in array 'startLocations'.
     * Example1:
     * input: <command> f1.1"abc"d22
     * startLocations: [255    5  255]  -> start-'"' found at location 5
     *
     * Example2:
     * input: <cmd> "abc" i10
     * startLocations: [  1  255  255]  -> start-'"' found at location 1
     *
     * value: 255 = unsigned char = |-1|: location irrelevant
     */
    static void findStringstartSeparators();

    /*
     * Find and convert parameters. Store them in the appropriate
     * array.
     */
    static void findParameters();

    // Execute the function the command is associated with.
    static void executeCommand();

    // Send 'unknown command'.
    static void printNoSuchCommand();

    #if defined INTERPRETER_TESTING
    // Used for debugging
    static void space();
    static void printSeparators();
    static void printSeparatorInfo();
    static void printParameterInfo();
    #endif  // INTERPRETER_TESTING

    // Interpreter c'tor: private
    Interpreter();
};

#endif /* INTERPRETER_H_ */
