/*
 * LcdButton.h
 *
 * Created on: 7 May 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * 
 * The complete button circuit has the following features:
 * - Power-on-by-button
 * - Normal button presses
 * - Touch mode: Activate LCD backlight
 * 
 * (Top View)                +------------------------------------------> 5V USB ESP32
 *                           |
 *                     +-----|-----+  FPQ27P06 P Channel MOSFET
 *                     | S   D   G |
 *                     +=|=======|=+            2N2222
 *                       |  ___  |           +-----------+
 *       +---------------+-|___|-+-----------| C       E |-------+
 *       |               | R = 100k          \     B     /       |
 *       |             \                       ----|----        GND
 *       |      Button  \                          |
 *       |               |                         |
 *       |               +----> Touch Breakout     |
 *       |               |                         |
 *       |               |     ____                |    ____ 
 *       |               +----|____|------I>|------+---|____|-------|<I----< GPIO #13
 *    5V |               |    R = 22k    Diode     |   R = 330     Diode     PWR HOLD
 *    +  |              ---                        |
 *    -------           | |  R = 100k              |
 *     =====            ---                        |
 *       |               |                        ---           
 *       |               +------> GPI #36 / A4    | |  R = 100k 
 *       |               |        LCD Btn         ---
 *       |              ---                        |
 *       |              | |  R = 220k              |
 *       |              ---                        |
 *       |               |                         |
 *      GND             GND                       GND
 *  
 * 
 * CP. also:
 * https://learn.upesy.com/en/programmation/Arduino-ESP32/what-are-interrupts-in-esp32-with-examples-for-arduino-code.html
 */

#ifndef LCDBUTTON_H_
#define LCDBUTTON_H_


class LcdButton {
  public:
    // Initialise button class
    static void begin();

    static void enable();

    // Disable button (disable interrupt)
    // THIS DISABLES OTHER INTERRUPTS AS WELL!
    // Cp. data sheet
    static void disable();

    static void setChanged(bool flag);
    static void setDown(bool flag);
    static void setUp(bool flag);
    
    static bool hasChanged();
    static bool isDown();
    static bool isUp();

    // Only used in polling mode
    static void powerOffIfWbServerActive();

  private:
    static void btnIsr();

    // Flags whether button has changed, is down, up
    static volatile bool changed;
    static volatile bool down;
    static volatile bool up;
};

#endif /* LCDBUTTON_H_ */
