/*
 * LcdLogger.h
 *
 * Created on: 20 May 2021
 * Author: ruedi.mueller@fhnw.ch
 * 
 * The class LcdLoggger realises one logger for the LCD display mode (as all
 * variables and functions are static). It instantiates a timer which determines
 * when data is logged.
 * To log data the following conditions must be met:
 * In setup()   LcdLogger::begin();   must be called to initialise classes
 * and instances for the LCD display mode
 */

#ifndef LCDLOGGER_H_
#define LCDLOGGER_H_

#include <Arduino.h>
#include <SPI.h>
#include <SD.h>
#include <RTClib.h>
#include "Timer.h"


/**
 * The class LcdLogger provides services to log data to the inserted SD card.
 * It is used in lcd display mode and logs all 3 sensor values.
 */
class LcdLogger {
  public:
    // The logger's timer
    static Timer timer;

    // Locking flag and millis var used for LcdButton disabling
    static bool buttonLocked;
    static long lockingTimepoint;

    // Create file logger and its associated lcd display mode file.
    static void begin();

    // Answer if filename has a correct format
    static bool isFilenameCorrect(String filename);

    // Create a new numbered file (from default name) with a higher number.
    // Answer if creation succeeded.
    // Valid filenames in WbPreferences and used names for logging:
    // WbPreferences  -->   logging name
    //   t.csv                t00.csv
    //   t0.csv               t000.csv
    //   name.csv             name00.csv
    //   name0.csv            name00.csv
    //   name00.csv           name00.csv
    //   sample.csv           sample00.csv
    //   example.csv          exampl00.csv
    //   example0.csv         exampl00.csv
    //   example7.csv         exampl00.csv
    //   exampl00.csv         exampl00.csv
    //   t                    t00
    //   ex.a                 ex00.a
    //   exa.long             exa00.lon
    static bool createFile();

    // Get name of the associated file
    static String getFilename();

    // Set name of the associated file
    static void setFilename(String filename);

    // Write all column headers into the csv logfile
    static void writeFileHeaders();

    // Write all values into the csv logfile
    static void writeFileValues();

    // Enable logging
    static void enable();

    // Disable logging
    static void disable();

    // Answer true if logging is enabled
    static bool isEnabled();

    // Handle LCD display mode: to be used in loop()
    static void handleLcdDisplayMode();

  private:
      // Used for createFile
    static char buf[];
    // Flag to mark enabled logging
    static bool enabled;
    // File system object
    static File root;
    // The file name of the logfile
    static String filename;
    // The file associated to this logger
    static File logfile;
    // Write contents
    static void writeFile(String data);
};

#endif /* LCDLOGGER_H_ */
