/*
 * LcdStatesDiagram.h
 *
 * Created on: 19 May 2021
 * Author: ruedi.mueller@fhnw.ch
 * 
 * This is the class where all states and transition functions are defined.
 *                                                       
 * State diagram:                                        .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .
 *                                                       .    legend:                                                         .
 *            o===============o                          .    o===============o                                               .
 *            I               I LONG_PRESS               .    I    state      I                                               .
 *            I   lgSplash    I offZeroMillis            .    o===============o                                               .
 *            I               I--------> 1)              .            | transition condition (time [millis] or other event)   .
 *            I               I                          .            | transition function                                   .
 *            I               I                          .            |                                                       .
 *            o===============o                          .            v to next state                                         .
 *                   | SHORT_PRESS                       .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .            
 *                   | pLgReadyScreen1                                  
 *                   |
 *                   |
 *                   |   +-------------------------------------------------------------------------------------------------------+
 *                   |   |   +-------------------------------------------------------------------------------------------------+ |
 *                   v   v   v                                                                                                 | |
 *               o===============o 2000                      o==========o 2000       o==========o 2000           o==========o  | |
 *               I               I pLgReadyScreen2           I          I pLg..3     I          I pLg..2         I          I  | |
 *               I   lgReady1    I-------------------------->I lgReady2 I----------->I lgReady3 I--------------->I lgReady4 I  | |
 *               I               I                           I          I            I          I                I          I  | |
 *               I               I                           I          I            I          I           2000 I          I  | |
 *               I               I                           I          I            I          I      pLgReady~ I          I  | |
 *               I               I                           I          I            I          I       ~Screen1 I          I  | |
 *               I               I<-------------------+      I          I            I          I            +---I          I  | |
 *               o===============o                    |      o==========o            o==========o            |   o==========o  | |
 *       LONG_PRESS |        | SHORT_PRESS            |  LONG_~|    | SHORT_PRESS      |    |                |     |    |      | |
 *                  v        | pLgRunScreenWriteFile  | ~PRESS v    | pLgRunScreen~    v    |                |     v    |      | |
 *                 2)        |                        |       3)    | ~WriteFile      4)    |                |    5)    |      | |
 *                           |                        |             |                       |                |          |      | |
 *           +---------------+                        +-----<-------|-----------------------|----------------+          |      | |
 *           |  +-------------------------------------<-------------+                       |                           |      | |
 *           |  |  +----------------------------------<-------------------------------------+                           |      | |
 *           |  |  |  +-------------------------------<-----------------------------------------------------------------+      | |
 *           |  |  |  |   +--------------------------------------------<---------------------------------------------------+   | |
 *           |  |  |  |   |                                                                                                |   | |
 *           v  v  v  v   v                                                                                                |   | |
 *         o===============o 2000                    o===============o 2000                o===============o 2000          |   | |
 *         I               I pLgRunScreen2           I               I pLgRunScreen3       I               I pLgRunScreen1 |   | |
 *         I  lgRunBlink1  I------------------------>I  lgRunBlink2  I-------------------->I  lgRunBlink3  I----->---------+   | |
 *         I               I                         I               I                     I               I                   | |
 *         I               I WRITE_FILE              I               I WRITE_FILE          I               I WRITE_FILE        | |
 *         I               I pLgRunScreenWriteValues I               I pLgRunScreenWrit... I               I pLgRunScreenWr... | |
 *         I               I------->--------+        I               I----->----+          I               I----->----+        | |
 *         I               I                |        I               I          |          I               I          |        | |
 *         I               I<---------------+        I               I<---------+          I               I<---------+        | |
 *         I               I                         I               I                     I               I                   | |
 *         I               I 1000                    I               I 1000                I               I 1000              | |
 *         I               I pLgRunBlinkOff          I               I pLgRunBlinkOff      I               I pLgRunBlinkOff    | |
 *         I               I------->--------+        I               I----->----+          I               I----->----+        | |
 *         I      3: 1:    I                |        I 3:   1:       I          |          I               I          |        | |
 *         I      4: 2:    I                |        I 4:   2:       I          |          I               I          |        | |
 *         o===============o                |        o===============o          |          o===============o          |        | |
 *          ^ ^ ^  |  |  ^                  |          |     |    ^             |            |     |    ^             |        | |
 *          | | |  |  |  |                  |          |     |    |             |            |     |    |             |        | |
 *          | | |  v  |  +------------------+          v     |    +-------------+            v     |    +-------------+        | |
 *          | | |  |  |                                |     |                               |     |                           | |
 *   +------|-|-|--+  |                                |     |                  3:LONG_PRESS |     |                           | |
 *   | +----|-|-|-----|--------------------------------+     |                 4:p2Countdown |     |                           | |
 *   | | +--|-|-|-----|--------------------------------------|-------------------------------+     |                           | |
 *   | | |  | | |     |     +--------------------------------+                                     | 1:SHORT_PRESS             | |
 *   | | |  | | |     |     |   +------------------------------------------------------------------+ 2:pLgPause1               | |
 *   | | |  | | |     |     |   |                                                                                              | |
 *   | | |  | | |     v     v   v                                                                                              | |
 *   | | |  | | |   o===============o 2000                o===============o                                                    | |
 *   | | |  | | |   I               I pLgPause2           I               I                    LONG_PRESS:                     | |
 *   | | |  | | |   I   lgPause1    I-------------------->I   lgPause2    I                    > 1000 ms                       | |
 *   | | |  | | |   I               I                     I               I                                                    | |
 *   | | |  | | |   I               I                2000 I               I                    SHORT_PRESS:                    | |
 *   | | |  | | |   I               I           pLgPause1 I               I                    < 1000 ms                       | |
 *   | | |  | | |   I               I<--------------------I               I                                                    | |
 *   | | |  | | |   I               I                     I               I                    DOWN:                           | |
 *   | | |  | | |   I               I LONG_PRESS          I               I LONG_PRESS         Button not released             | |
 *   | | |  | | |   I               I p2Countdown         I               I p2Countdown                                        | |
 *   | | |  | | |   I               I--------------+      I               I------------+         RELEASED:                     | |
 *   | | |  | | |   I               I<----------+  |      I               I            |         Button released               | |
 *   | | |  | | |   o===============o           |  |      o===============o            |                                       | |
 *   | | |  | | |       | SHORT_PRESS           |  |          | SHORT_PRESS            |                                       | |
 *   | | |  | | |       | pLgRunScreen1         |  |          | pLgRunScreen1          |                                       | |
 *   | | |  | | +-------+                       |  |          |                        |                                       | |
 *   | | |  | |                                 |  +----------|--------------------+   |                                       | |
 *   | | |  | |                                 +-------------|------+             |   |                                       | |
 *   | | |  | +-----------------------------------------------+      |             |   |                                       | |
 *   | | |  |                                                        |             |   |                                       | |
 *   | | |  |                  +------------------------------+      |             |   |   +--------------------------+        | |
 *   | | |  |                  |    +----------------------+  |      |             |   |   |    +------------------+  |        | |
 *   | | |  |                  |    |                      |  |      |             |   |   |    |                  |  |        | |
 *   | | |  |                  v    v                      |  |      |             v   v   v    v                  |  |        | |
 *   | | |  |             o================o 1000 and DOWN |  |      |          o==================o 1000 and DOWN |  |        | |
 *   | | |  |             I                I p1Countdown   |  |      |          I                  I p1Countdown   |  |        | |
 *   | | |  +------<------I lgRunCountdown I---------------+  |      +----<-----I lgPauseCountdown I---------------+  |        | |
 *   | | |       RELEASED I                I                  |        RELEASED I                  I                  |        | |
 *   | | |  pLgRunScreen1 I                I 2000 and DOWN    | pLgPauseScreen1 I                  I 2000 and DOWN    |        | |
 *   | | |                I                I p0Countdown      |                 I                  I p0Countdown      |        | |
 *   | | +--------------->I                I------------------+                 I                  I------------------+        | |
 *   | +----------------->I                I                                    I                  I                           | |
 *   +------------------->I                I                                    I                  I                           | |
 *                        o================o                                    o==================o                           | |
 *                                | 2500                                                 | 2500                                | |
 *                                | pLgReadyScreen1                                      | pLgReadyScreen1                     | |
 *                                |                                                      +-------------------------------------+ |
 *                                +----------------------------------------------------------------------------------------------+
 *
 * 
 *                             1)                              
 *                              |                              
 *                              v                              
 *                          o===============o                  
 *                 RELEASED I               I 2000, 3000, 4000, 5000,
 *                      off I    lgWifi     I 6000, 7000, 8000
 *             +------------I               I on, off, on, off, on, off, on  
 *             |            I               I-----------------+
 *             v            I               I                 |
 *         lgSplash         I               I                 |
 *                          I               I<----------------+
 *                          I               I                  
 *                          I               I 9000            
 *                          I               I pLgWifiScreen   
 *                          I               I----------+       
 *                          I               I          |       
 *                          o===============o          =       
 *                                                  lgSplash (blocked: restart needed)
 * 
 *                             2)                                                                    3)
 *                              |                                                                     |
 *                              v                                                                     v
 *                          o===============o                                                    o===============o
 *                 RELEASED I               I 2000, 3000, 4000, 5000,                   RELEASED I               I 2000, 3000, 4000, 5000,
 *          pLgReadyScreen3 I   lgPwrOff1   I 6000, 7000, 8000                   pLgReadyScreen2 I   lgPwrOff2   I 6000, 7000, 8000
 *             +------------I               I pLgPwrOffScreen3, pwr,                +------------I               I pLgPwrOffScreen2, pwr,
 *             |            I               I pLgPwrOffScreen2, pwr,                |            I               I pLgPwrOffScreen1, pwr,
 *             |            I               I pLgPwrOffScreen1, pwr,                |            I               I pLgPwrOffScreen2, pwr,
 *             |            I               I pLgPwrOffScreen2                      |            I               I pLgPwrOffScreen3
 *             |            I               I-----------------+                     |            I               I-----------------+
 *             v            I               I                 |                     v            I               I                 |
 *         lgReady1         I               I                 |                 lgReady2         I               I                 |
 *                          I               I<----------------+                                  I               I<----------------+
 *                          I               I                                                    I               I
 *                          I               I 9000                                               I               I 9000
 *                          I               I poweroff                                           I               I poweroff
 *                          I               I----------+                                         I               I----------+
 *                          I               I          |                                         I               I          |
 *                          o===============o          =                                         o===============o          =
 *                                                 powered off                                                           powered off
 *  * 
 *                             4)                                                                    5)
 *                              |                                                                     |
 *                              v                                                                     v
 *                          o===============o                                                    o===============o
 *                 RELEASED I               I 2000, 3000, 4000, 5000,                   RELEASED I               I 2000, 3000, 4000, 5000,
 *          pLgReadyScreen1 I   lgPwrOff3   I 6000, 7000, 8000                   pLgReadyScreen2 I   lgPwrOff4   I 6000, 7000, 8000
 *             +------------I               I pLgPwrOffScreen1, pwr,                +------------I               I pLgPwrOffScreen2, pwr,
 *             |            I               I pLgPwrOffScreen2, pwr,                |            I               I pLgPwrOffScreen3, pwr,
 *             |            I               I pLgPwrOffScreen3, pwr,                |            I               I pLgPwrOffScreen2, pwr,
 *             |            I               I pLgPwrOffScreen2                      |            I               I pLgPwrOffScreen1
 *             |            I               I-----------------+                     |            I               I-----------------+
 *             v            I               I                 |                     v            I               I                 |
 *         lgReady3         I               I                 |                 lgReady4         I               I                 |
 *                          I               I<----------------+                                  I               I<----------------+
 *                          I               I                                                    I               I
 *                          I               I 9000                                               I               I 9000
 *                          I               I poweroff                                           I               I poweroff
 *                          I               I----------+                                         I               I----------+
 *                          I               I          |                                         I               I          |
 *                          o===============o          =                                         o===============o          =
 *                                                 powered off                                                           powered off
 * 
 */
#ifndef LCDSTATESDIAGRAM_H_
#define LCDSTATESDIAGRAM_H_

#include "LcdState.h"


class LcdStatesDiagram {
  public:
    // States
    static LcdState lgSplash;

    static LcdState lgWifi;

    static LcdState lgReady1;
    static LcdState lgReady2;
    static LcdState lgReady3;
    static LcdState lgReady4;

    static LcdState lgPwrOff1;
    static LcdState lgPwrOff2;
    static LcdState lgPwrOff3;
    static LcdState lgPwrOff4;

    static LcdState lgRunBlink1;
    static LcdState lgRunBlink2;
    static LcdState lgRunBlink3;

    static LcdState lgPause1;
    static LcdState lgPause2;

    static LcdState lgRunCountdown;
    static LcdState lgPauseCountdown;

    // State arrays
    static LcdState* lgSplash_connectedStates[];   // states array
    static int lgSplash_conditions[];              // transition conditions array
    static void (* lgSplash_actions[])();          // transition functions array

    static LcdState* lgWifi_connectedStates[];
    static int lgWifi_conditions[];
    static void (* lgWifi_actions[])();

    static LcdState* lgReady1_connectedStates[];
    static int lgReady1_conditions[];
    static void (* lgReady1_actions[])();

    static LcdState* lgReady2_connectedStates[];
    static int lgReady2_conditions[];
    static void (* lgReady2_actions[])();

    static LcdState* lgReady3_connectedStates[];
    static int lgReady3_conditions[];
    static void (* lgReady3_actions[])();

    static LcdState* lgReady4_connectedStates[];
    static int lgReady4_conditions[];
    static void (* lgReady4_actions[])();

    static LcdState* lgPwrOff1_connectedStates[];
    static int lgPwrOff1_conditions[];
    static void (* lgPwrOff1_actions[])();

    static LcdState* lgPwrOff2_connectedStates[];
    static int lgPwrOff2_conditions[];
    static void (* lgPwrOff2_actions[])();

    static LcdState* lgPwrOff3_connectedStates[];
    static int lgPwrOff3_conditions[];
    static void (* lgPwrOff3_actions[])();

    static LcdState* lgPwrOff4_connectedStates[];
    static int lgPwrOff4_conditions[];
    static void (* lgPwrOff4_actions[])();

    static LcdState* lgRunBlink1_connectedStates[];
    static int lgRunBlink1_conditions[];
    static void (* lgRunBlink1_actions[])();

    static LcdState* lgRunBlink2_connectedStates[];
    static int lgRunBlink2_conditions[];
    static void (* lgRunBlink2_actions[])();

    static LcdState* lgRunBlink3_connectedStates[];
    static int lgRunBlink3_conditions[];
    static void (* lgRunBlink3_actions[])();

    static LcdState* lgPause1_connectedStates[];
    static int lgPause1_conditions[];
    static void (* lgPause1_actions[])();

    static LcdState* lgPause2_connectedStates[];
    static int lgPause2_conditions[];
    static void (* lgPause2_actions[])();

    static LcdState* lgRunCountdown_connectedStates[];
    static int lgRunCountdown_conditions[];
    static void (* lgRunCountdown_actions[])();

    static LcdState* lgPauseCountdown_connectedStates[];
    static int lgPauseCountdown_conditions[];
    static void (* lgPauseCountdown_actions[])();

    // State transition condition functions
    // These functions can be exchanged using
    // FSM::setConditionFunction(buttonShortCondFunction);
    // Set FSM to be susceptible to short button presses
    static int buttonShortCondFunction();
    // Set FSM to be susceptible to long button presses
    static int buttonLongCondFunction();
    // Set FSM to be susceptible to released button
    static int buttonReleasedCondFunction();
    // Set FSM to be susceptible to write action
    static int writeCondFunction();

    // Init state connections, conditions, transition actions and array lengths.
    static void initStates();

    // Perform next step within the button and lcd scenario
    // if waterbox is used via button and lcd menu
    static void fsmExecute();

    // Set flag to show that lcd display mode is active
    static void setActive(bool flag);
    // Answer whether lcd display mode is active
    static bool isActive();

    // Set/reset flag whether sensor data can be collected or not
    static void collectOn();
    static void collectOff();
    static bool canCollect();

    // Power off ESP32
    static void poweroff();

  private:
    // Flag whether FSM/LcdStatesDiagram can run
    static bool active;

    // Flag to signal if sensor data can be collected
    static bool collect;

    static void on();
    static void off();
    static void offZeroMillis();
    static void pwr();
};

#endif /* LCDSTATESDIAGRAM_H_ */
