/*
 * Logger.h
 *
 * Author: ruedi.mueller@fhnw.ch
 *
 * The class Logger provides functionality to log values
 * in a csv file on a SD card.
 *
 * SD card preparation:
 * Formatted, FAT32. Only filenames of format 8.3 accepted.
 *
 * Usage:
 * Generally a logger (and only one is possible) is set up
 * with an associated logfile. A name of 2 to 6 letters and
 * the extension .csv is recommended.
 *
 * > Logger::begin("sample.csv");
 * sets up the logger. If no sample00.csv is found on the SD
 * card, this will be the first name. After a reset and a new
 * setup a new file sample01.csw will be associated with the
 * logger.
 *
 * The associated file is where all log data will be written.
 *
 * The values to be written must be provided as Strings. The
 * csv file's first row shows the meaning of the data as seen
 * in Excel tables (column titles). The lines below depict the
 * data values.
 * To define the header coulumn simply use statements as follows:
 * > Logger::addHeader("date");
 * > Logger::addHeader("time");
 * To write them into the associated file use
 * > Logger::writeFileTitles();
 *
 * To collect data to be written use the header and the data as below:
 * > Logger::addValue("time", "22:23:44");
 * > Logger::addValue("date", "01-Jan-2020");
 * The order of adding values does not matter as it was defined by
 * the addHeader functions - left to right.
 *
 * If the order of titles in the header line do not matter you can
 * also use the addValue function without using addHeader functions.
 * In this case a missing header will be added behind the scenes.
 * Make sure to writeFileTitles only once and after having added
 * all needed values..
 *
 * To write the data values use
 * > Logger::writeFileValues()
 * This would be repeatedly done if logging is on.
 *
 * Some added functions are ready to use:
 * > Logger::getFilename()
 * > Logger::getFilename(String filename)
 * > Logger::printFileTitles()
 *
 * How to prepare the logger:
 * > loggersettings
 * show the current values of the logger.
 * To set the start time and the stop time write
 * > setloggerstarttime i22i33
 * > setloggerstoptime i23i44
 * The logger interval, the period of time after which values
 * are logged to the logfile, is set as follows:
 * > setloggerinterval i0i10i0
 * Logger setting for this example show the following values:
 * > loggersettings
 *  start logger: 10/02/2021  22:33:00
 *  stop logger:  10/02/2021  23:44:00
 *  log interval:             00:10:00
 * If required logging over longer periods is possible. Use
 * the commands 'setloggerstartdate' and 'setloggerstopdate'
 * but only aftr having set the respective time.
 * Logger intervals can be prolonged as well. Use the command
 * 'addloggerintervaldays' but only after having set the
 * interval time.
 * Use
 * > resetloggertimes
 * to reset the times if you like to start again.
 *
 *
 * Sources:
 * adafruit.com; sparkfun Electronics
 *
 * History:
 * 10 Feb 2021  rm   init
 */

#ifndef LOGGER_H_
#define LOGGER_H_

#include <Arduino.h>
#include <SPI.h>
#include <SD.h>
#include <RTClib.h>
#include "Timer.h"


/**
 * The class Logger provides services to log data to the inserted SD card.
 * Other services as reading from and writing to the card are available.
 */
class Logger {
  public:
    // The logger's timer
    static Timer timer;

	  // Create file logger and its associated file.
    static void begin(String defaultFilename);

    // Answer if filename has a correct format
    static bool isFilenameCorrect(String filename);

    // Create a new numbered file (from default name) with a higher number.
    // Answer if creation succeeded.
    static bool createFile(String filename);
    
    // Get name of the associated file
    static String getFilename();

    // Set the name of the logfile
    static void setFilename(String filename);

    // Write all selected column headers into the csv logfile
    static void writeFileHeaders();

    // Write all selected values into the csv logfile
    static void writeFileValues();

    // Print the current filename's contents:
    // the first 4 lines, the last last 3 lines
    static void printPartialFile();

    // Print SD card info
    static void printSdCardInfo();

    // Flag whether temperature is to be logged or not
    static void setTempOn(bool flag);
    // Answer true if temp logging is on
    static bool isTempOn();

    // Flag whether conductivity is to be logged or not
    static void setConductivityOn(bool flag);
    // Answer true if conductivity logging is on
    static bool isConductivityOn();

    // Flag whether turbidity is to be logged or not
    static void setTurbidityOn(bool flag);
    // Answer true if turbidity logging is on
    static bool isTurbidityOn();

    // Print temperature sensor settings
    static void printTempSettings();

    // Print conductivity sensor settings
    static void printConductivitySettings();

    // Print turbidity sensor settings
    static void printTurbiditySettings();

    // Handle logger in loop()
    static void handleLogger();

    static void directory();

  private:
    // Used for createFile
    static char buf[];
    
    // File system object
    static File root;
    // The file name of the logfile
    static String filename;
    // The file associated to this logger
    static File logfile;

    // Memorise whether temperature is to be logged
    static bool tempOn;
    // Memorise whether conductivity is to be logged
    static bool conductivityOn;
    // Memorise whether turbidity is to be logged
    static bool turbidityOn;
    // Write contents
    static void writeFile(String data);

    static char* getDateTime(File file);
};

#endif /* LOGGER_H_ */
