/*
 * Temperature.h
 *
 * Created on: 18 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * Hardware:
 *                              +---------+
 *    GND   --------------------|         |
 *    3.3 V -----------+--------|         |
 *                     |        | DS18B20 |
 *                  R 4.7 k     |         |
 *                     |        |         |    Data Pin Number:
 *    Data Pin --------+--------|         |    Config::TEMP_PIN  GPIO #21
 *                              +---------+
 *
 * More information:
 * https://randomnerdtutorials.com/esp32-ds18b20-temperature-arduino-ide/
 * 
 * This class uses a timeout timer to prevent that temperatures a read in too quick
 * a succession which would lead to a crashed One Wire Bus.
 * Temperatures values are updated after 4 to 6 seconds.
 */

#ifndef TEMPERATURE_H_
#define TEMPERATURE_H_

#include <OneWire.h>
#include <DallasTemperature.h>
#include "TwoPointCalibrationHelper.h"
#include "Timeout.h"


// More than one IC on the same bus possible.
// If only one: 0 refers to the first IC on the wire.
// This is assumed for this project.
class Temperature {
  public:
    // Start up the library
	  static void begin();

	  // Read temperature from sensor
    static float readRaw();

    // Format temperature to %f.1 and answer string with unit (= true) or not
    static String formatString(float temp, bool withUnit);

    // Two-Point-Calibration Helper with functions to set up the temp sensor
    static TwoPointCalibrationHelper helper;

  private:
    static char buf[];

    // Setup a oneWire instance
    static OneWire oneWire;

    // Pass oneWire reference to Dallas Temperature
    static DallasTemperature sensor;

    // Buffer for last read temperature value
    static float lastValue;

    // Timeout which determines when to read a new lastValue
    // As reading is synchronised with FSM the defined period
    // denotes a miniumum. Add 2 secs.
    static Timeout timeout;
};

#endif /* TEMPERATURE_H_ */
