/*
 * Timer.h
 *
 * Created on: 18 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * NOTE:
 * Using char buf[50] for sprintf within a function causes
 * errors. Therefore realised as class variable.
 */

#ifndef TIMER_H_
#define TIMER_H_

#include <Arduino.h>
#include <RTClib.h>


/*
 * The class Timer is responsible for storing start times, stop times
 * and interval periods.
 * IMPORTANT:
 * Must be used in loop() in a way that the timer can regularly check
 * millis.
 */
class Timer {
  public:
    // C'tor with name where called from
    Timer(String name);

    // Set preferences for the timer
    void setPrefs();

    // Set time when reading should begin today.
    // Use setStartDate if date is later than today.
    void setStartTimeToday(int hh, int mm);

    // Set date when reading should begin later than today.
    // Otherwise use only setStartTimeToday
    void setStartDate(int dd, int mm, int yyyy);

    // Set time when reading should end today.
    // Use setStopDate if date is later than today.
    void setStopTimeToday(int hh, int mm);

    // Set date when reading should end later than today.
    // Otherwise use only setStopTimeToday
    void setStopDate(int dd, int mm, int yyyy);

    // Set the time interval (used for monitor and logger)
    void setInterval(int mm, int ss);

    // Set the time interval (used for monitor and logger)
    void setInterval(int hh, int mm, int ss);

    // Add the day interval (used for monitor and logger)
    // The time interval must be set beforehand.
    void setIntervalDays(int dd);

    // Reset all reading times
    void resetTimes();

    // Print the settings for the reading
    void printSettings();

    // Read interval from preferences
    TimeSpan readInterval();

    // Write interval to preferences
    void writeInterval(TimeSpan ts);

    // Enable to run
    void run();

    // Halt timer immediately
    void halt();

    // Test whether the timer is running
    bool isRunning();

    // Test whether we are within the measuring period.
    // If yes: true is answered and the next DateTime will be set.
    // If no: false will be answered.
    // If we are before or after the measuring period the running
    // flag will be disabled.
    bool isWithinMeasuringPeriod();

    // Getters (by value) and setters
    DateTime getStartDateTime();
    void setStartDateTime(DateTime startDateTime);
    DateTime getStopDateTime();
    void setStopDateTime(DateTime stopDateTime);
    TimeSpan& getInterval();
    void setInterval(TimeSpan interval);

    // Enable to run (LCD display mode only).
    void lcdRun();
    // Disable to run (LCD display mode only).
    void lcdHalt();
    // Set the logDateTime to now so that logging is started
    // instantly (LCD display mode only).
    void setLogDateTimeNow();
    // Answer if measured values are to be logged (LCD display mode only).
    bool isTimeToLog();

  private:
    // Used for sprintf conversions
    static char buf[];

	  // Where called from (Monitor or Logger)
	  String whereFrom;
    // Date and time when to start with reading
    DateTime startDateTime;
    // Date and time when to stop with reading
    DateTime stopDateTime;
    // At what intervals to read
    TimeSpan interval;
    // Flag to memorise whether reading is enabled to run or not
    bool running;

    // This is the DateTime when the next printing/logging will take place
    DateTime nextLogDateTime;

    // Set dates and times back
    void reset();

    // Print that timer is still running
    void printRunningError();

    // Print the date and time of dateTime
    void printDateTime(DateTime dateTime);

    // Print the time of dateTime only
    void printTime(DateTime dateTime);

    // Print the interval time
    void printInterval();

    // Print the interval time, different format
    void printIntervalModif();
};

#endif /* TIMER_H_ */
