/*
 * Turbidity.h
 *
 * Created on: 27 Jun 2023
 * Author: ruedi.mueller@fhnw.ch
 *
 * The turbidity sensor is able to detect suspended particles
 * in water by measuring the light transmittance and scattering
 * rate which changes with the amount of total suspended solids
 * in water.
 * 
 * WITHOUT CALIBRATION
 *
 * Settings:
 * Make sure the on-board switch "A<->D" on the breakout is set to
 * "A" for analog readings.
 */

#ifndef TURBIDITY_H_
#define TURBIDITY_H_

#include <Arduino.h>


class Turbidity {
  public:
    // Start up the library
    static void begin();

    // Read newturbidity from sensor
    static float readRaw();

    static String readString(bool withUnit);

    static float depth();

    static void printVoltageArray();
    static void printVisibilityArray();

    // Turbidity out-of-range values
    static const float TOO_HIGH;
    static const float TOO_LOW;

    static const float NEVER_USED;

  private:
    // Return array index where the volt value is greater equal the array value
    static int getIndex(float volt);

    // Partially map (interpolate in a linear way) volt to visibility depth (Secchi disk).
    // Test if value is too high or too low in which case an out-of-range value is answered.
    // Answer a manually interpolated value for cases where consecutive identical voltages
    // exist in the voltage array.
    // In all other cases interpolate and answer the visibility value.
    static float myMap(float volt);

    // Format turbidity
    static String formatString(float turbidity, bool withUnit);

    static void printArray(float* arr);

    static char buf[];

    // Lookup arrays (ordered values from high voltage to low voltage). All
    // arrays must contain the same number of values.
    // Sensor comparation voltage
    static float voltage[];
    // Visibility using Secchi disk
    static float visibility[];
    // Where to interpolate and where not
    static float interpolate[];
    // Size of arrays
    static const int TURB_TABLE_SIZE;
};

#endif /* TURBIDITY_H_ */
