/*
 * Commands.cpp
 *
 * Created on: 20 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 *
 * CAUTION with buf size BUF_SIZE.
 */

#include "Commands.h"
#include "Interpreter.h"
#include "Config.h"
#include "WbRTC.h"
#include "WbLCD.h"
#include "WbPreferences.h"
#include "LcdLogger.h"
#include "LcdStatesDiagram.h"
#include "LcdFSM.h"
#include "WbServer.h"
#include "LcdButton.h"
#include "LcdText.h"
#include "Monitor.h"
#include "Logger.h"
#include "LogComposer.h"
#include "Temperature.h"
#include "Conductivity.h"
#include "Turbidity.h"
#include "Streaming.h"


#define BUF_SIZE 100

// Tabulators for printHelp() etc.
#define HELP_TAB 17
#define SENSORS_TAB 35
#define DATETIME_TAB 32
#define LOGFILE_TAB 28
#define LOGGER_TAB 43
#define MONITOR_TAB 44
#define LCD_TAB 37
#define MISC_TAB 39
#define DEVEL_TAB 56

// MENU_TABLE, TEXT_TABLE and FP_TABLE arrays MUST have all the same size
#define TABLE_SIZE (sizeof(MENU_TABLE) / sizeof(MENU_TABLE[0]))
// Alias for shorter text
#define C Commands

// BE SURE TO HAVE BIG ENOUGH A BUFFER
// Used to build text strings in RAM --> Must be as long as
// longest text string + 1.
char Commands::buf[BUF_SIZE];

boolean Commands::toPowerOff = false;


// Order is crucial!
void (* const Commands::FP_TABLE[])() = {
  // --- HELP --------------------------------------------------------------------------------------
  C::printHelpPT, C::printHelpSensorsPT, C::printHelpTimeDatePT, C::printHelpLogfilePT,
  C::printHelpLoggerPT, C::printHelpMonitorPT, C::printHelpLcdPT, C::printHelpMiscPT,
  C::printHelpDevelPT,
  // --- SENSORS -----------------------------------------------------------------------------------
  C::printTempPT, C::printTempCalibValuesPT, C::setTempCalibLowValuesPT, C::setTempCalibHighValuesPT,
  C::printConductivityPT, C::printConductivityCalibValuesPT, C::setConductivityCalibValuesPT, C::printTurbidityPT, 
  C::printTurbidityCalibValuesPT,
  // --- DATE AND TIME -----------------------------------------------------------------------------
  C::printDatePT, C::printTimePT, C::setDatePT, C::setTimePT,
  // --- LOGFILE -----------------------------------------------------------------------------------
  C::printLogfileNamePT, C::setLogfileNamePT, C::printPartialLogfilePT,
  // --- LOGGER ------------------------------------------------------------------------------------
  C::setLoggerStartTimeTodayPT, C::setLoggerStartDatePT, C::setLoggerStopTimeTodayPT, C::setLoggerStopDatePT,
  C::setLoggerIntervalPT, C::setLoggerIntervalDaysPT, C::resetLoggerPT, C::printLoggerSettingsPT,
  C::logTempOnPT, C::logTempOffPT, C::logConductivityOnPT, C::logConductivityOffPT,
  C::logTurbidityOnPT, C::logTurbidityOffPT, C::loggerRunPT, C::loggerHaltPT,
  // --- MONITOR -----------------------------------------------------------------------------------
  C::setMonitorStartTimeTodayPT, C::setMonitorStartDatePT, C::setMonitorStopTimeTodayPT, C::setMonitorStopDatePT,
  C::setMonitorIntervalPT, C::setMonitorIntervalDaysPT, C::resetMonitorPT, C::printMonitorSettingsPT,
  C::monitorTempOnPT, C::monitorTempOffPT, C::monitorConductivityOnPT, C::monitorConductivityOffPT,
  C::monitorTurbidityOnPT, C::monitorTurbidityOffPT, C::monitorRunPT, C::monitorHaltPT,
  // --- LCD DISPLAY AND BUTTON MODE ---
  C::storeLcdLogfileNamePT, C::setLcdLoggerIntervalPT, C::setBacklightTimeoutPT, C::printLcdSettingsPT,
  // --- MISCELLANEOUS -----------------------------------------------------------------------------
  C::printVersionInstitutionPT, C::printSdCardInfoPT, C::setEchoOnPT, C::setEchoOffPT,
  C::setSerialPT, C::wifiPT, C::setNetPT, C::setApPT,
  C::setNetCredentialsPT, C::setApCredentialsPT, C::printLicensePT, C::poweroffPT,
  // --- DEVELOPMENT -------------------------------------------------------------------------------
  C::setMonitorSettingsPT, C::setLoggerSettingsPT, C::testrunMonitorPT, C::testrunLoggerPT,
  C::copyMonitorDatesToLoggerPT, C::fakesetTempCalibLowValuesPT, C::fakesetTempCalibHighValuesPT, C::fakesetConductivityCalibValuesPT,
  C::printPrefsPT, C::resetPrefsPT,
  C::enableSerial1PT, C::disableSerial1PT, C::setSerial1PT, C::startLcdPT,
  C::stopLcdPT, C::connectWifiPT, C::disconnectWifiPT, C::directoryPT,
  C::test0PT, C::test1PT, C::test2PT, C::test3PT,
  C::test4PT, C::test5PT, C::test6PT, C::test7PT,
  C::test8PT, C::test9PT,
};

// --- Tables --------------------------------------------------------------------------------------
// Number of commands in each group.
// To be adapted if changes are made
#define HH_N  9   // meaning: 9 commands in HELP group
#define SE_N  9   // 9 in SENSORS group etc.
#define DT_N  4   // DATE AND TIME
#define LF_N  3   // LOGFILE
#define LO_N 16   // LOGGER
#define MO_N 16   // MONITOR
#define LB_N  4   // LCD DISPLAY AND BUTTON MODE
#define MS_N 12   // MISCELLANEOUS
#define DE_N 28   // DEVELOPMENT

// Total number of array slots
#define ADDED_NUMS(n0, n1, n2, n3, n4, n5, n6, n7, n8) \
                  (n0 + n1 + n2 + n3 + n4 + n5 + n6 + n7 + n8)
#define TOTAL_NUM ADDED_NUMS(HH_N, SE_N, DT_N, LF_N, LO_N, MO_N, LB_N, MS_N, DE_N)


// Number and order is crucial!
const char* const Commands::MENU_TABLE[TOTAL_NUM] PROGMEM = {
  // --- HELP ---
  MENU_HELP_00, MENU_HELP_01, MENU_HELP_02, MENU_HELP_03,
  MENU_HELP_04, MENU_HELP_05, MENU_HELP_06, MENU_HELP_07,
  MENU_HELP_08,
  // --- SENSORS ---
  MENU_SENSORS_00, MENU_SENSORS_01, MENU_SENSORS_02, MENU_SENSORS_03,
  MENU_SENSORS_04, MENU_SENSORS_05, MENU_SENSORS_06, MENU_SENSORS_07,
  MENU_SENSORS_08,
  // --- DATE AND TIME ---
  MENU_DATETIME_00, MENU_DATETIME_01, MENU_DATETIME_02, MENU_DATETIME_03,
  // --- LOGFILE ---
  MENU_LOGFILE_00, MENU_LOGFILE_01, MENU_LOGFILE_02,
  // --- LOGGER ---
  MENU_LOGGER_00, MENU_LOGGER_01, MENU_LOGGER_02, MENU_LOGGER_03,
  MENU_LOGGER_04, MENU_LOGGER_05, MENU_LOGGER_06, MENU_LOGGER_07,
  MENU_LOGGER_08, MENU_LOGGER_09, MENU_LOGGER_10, MENU_LOGGER_11,
  MENU_LOGGER_12, MENU_LOGGER_13, MENU_LOGGER_14, MENU_LOGGER_15,
  // --- MONITOR ---
  MENU_MONITOR_00, MENU_MONITOR_01, MENU_MONITOR_02, MENU_MONITOR_03,
  MENU_MONITOR_04, MENU_MONITOR_05, MENU_MONITOR_06, MENU_MONITOR_07,
  MENU_MONITOR_08, MENU_MONITOR_09, MENU_MONITOR_10, MENU_MONITOR_11,
  MENU_MONITOR_12, MENU_MONITOR_13, MENU_MONITOR_14, MENU_MONITOR_15,
  // --- LCD DISPLAY/BUTTON ---
  MENU_LCD_00, MENU_LCD_01, MENU_LCD_02, MENU_LCD_03,
  // --- MICELLANEOUS ---
  MENU_MISC_00, MENU_MISC_01, MENU_MISC_02, MENU_MISC_03,
  MENU_MISC_04, MENU_MISC_05, MENU_MISC_06, MENU_MISC_07,
  MENU_MISC_08, MENU_MISC_09, MENU_MISC_10, MENU_MISC_11,
  // --- DEVELOPMENT ---
  MENU_DEVEL_00, MENU_DEVEL_01, MENU_DEVEL_02, MENU_DEVEL_03,
  MENU_DEVEL_04, MENU_DEVEL_05, MENU_DEVEL_06, MENU_DEVEL_07,
  MENU_DEVEL_08, MENU_DEVEL_09, MENU_DEVEL_10, MENU_DEVEL_11,
  MENU_DEVEL_12, MENU_DEVEL_13, MENU_DEVEL_14, MENU_DEVEL_15,
  MENU_DEVEL_16, MENU_DEVEL_17, MENU_DEVEL_18, MENU_DEVEL_19,
  MENU_DEVEL_20, MENU_DEVEL_21, MENU_DEVEL_22, MENU_DEVEL_23,
  MENU_DEVEL_24, MENU_DEVEL_25, MENU_DEVEL_26, MENU_DEVEL_27,
};

// Number and order are crucial!
const char* const Commands::TEXT_TABLE[] PROGMEM = {
  // --- HELP ---
  TEXT_HELP_00, TEXT_HELP_01, TEXT_HELP_02, TEXT_HELP_03,
  TEXT_HELP_04, TEXT_HELP_05, TEXT_HELP_06, TEXT_HELP_07,
  TEXT_HELP_08,
  // --- SENSORS ---
  TEXT_SENSORS_00, TEXT_SENSORS_01, TEXT_SENSORS_02, TEXT_SENSORS_03,
  TEXT_SENSORS_04, TEXT_SENSORS_05, TEXT_SENSORS_06, TEXT_SENSORS_07,
  TEXT_SENSORS_08,
  // --- TIME / DATE ---
  TEXT_DATETIME_00, TEXT_DATETIME_01, TEXT_DATETIME_02, TEXT_DATETIME_03,
  // --- LOGFILE ---
  TEXT_LOGFILE_00, TEXT_LOGFILE_01, TEXT_LOGFILE_02,
  // --- LOGGER ---
  TEXT_LOGGER_00, TEXT_LOGGER_01, TEXT_LOGGER_02, TEXT_LOGGER_03,
  TEXT_LOGGER_04, TEXT_LOGGER_05, TEXT_LOGGER_06, TEXT_LOGGER_07,
  TEXT_LOGGER_08, TEXT_LOGGER_09, TEXT_LOGGER_10, TEXT_LOGGER_11,
  TEXT_LOGGER_12, TEXT_LOGGER_13, TEXT_LOGGER_14, TEXT_LOGGER_15,
  // --- MONITOR ---
  TEXT_MONITOR_00, TEXT_MONITOR_01, TEXT_MONITOR_02, TEXT_MONITOR_03,
  TEXT_MONITOR_04, TEXT_MONITOR_05, TEXT_MONITOR_06, TEXT_MONITOR_07,
  TEXT_MONITOR_08, TEXT_MONITOR_09, TEXT_MONITOR_10, TEXT_MONITOR_11,
  TEXT_MONITOR_12, TEXT_MONITOR_13, TEXT_MONITOR_14, TEXT_MONITOR_15,
  // --- LCD DISPLAY/BUTTON ---
  TEXT_LCD_00, TEXT_LCD_01, TEXT_LCD_02, TEXT_LCD_03,
  // --- MICELLANEOUS ---
  TEXT_MISC_00, TEXT_MISC_01, TEXT_MISC_02, TEXT_MISC_03,
  TEXT_MISC_04, TEXT_MISC_05, TEXT_MISC_06, TEXT_MISC_07,
  TEXT_MISC_08, TEXT_MISC_09, TEXT_MISC_10, TEXT_MISC_11,
  // --- DEVELOPMENT ---
  TEXT_DEVEL_00, TEXT_DEVEL_01, TEXT_DEVEL_02, TEXT_DEVEL_03,
  TEXT_DEVEL_04, TEXT_DEVEL_05, TEXT_DEVEL_06, TEXT_DEVEL_07,
  TEXT_DEVEL_08, TEXT_DEVEL_09, TEXT_DEVEL_10, TEXT_DEVEL_11,
  TEXT_DEVEL_12, TEXT_DEVEL_13, TEXT_DEVEL_14, TEXT_DEVEL_15,
  TEXT_DEVEL_16, TEXT_DEVEL_17, TEXT_DEVEL_18, TEXT_DEVEL_19,
  TEXT_DEVEL_20, TEXT_DEVEL_21, TEXT_DEVEL_22, TEXT_DEVEL_23,
  TEXT_DEVEL_24, TEXT_DEVEL_25, TEXT_DEVEL_26, TEXT_DEVEL_27,
};


// Commands array in flash memory (PROGMEM)
// https://www.arduino.cc/reference/en/language/variables/utilities/progmem/
// --- HELP ----------------------------------------------------------------------
const char Commands::MENU_HELP_00[] PROGMEM = "help";
const char Commands::TEXT_HELP_00[] PROGMEM = "show help commands";

const char Commands::MENU_HELP_01[] PROGMEM = "helpsensors";
const char Commands::TEXT_HELP_01[] PROGMEM = "show sensor commands";

const char Commands::MENU_HELP_02[] PROGMEM = "helpdatetime";
const char Commands::TEXT_HELP_02[] PROGMEM = "show date and time commands";

const char Commands::MENU_HELP_03[] PROGMEM = "helplogfile";
const char Commands::TEXT_HELP_03[] PROGMEM = "show logfile commands";

const char Commands::MENU_HELP_04[] PROGMEM = "helplogger";
const char Commands::TEXT_HELP_04[] PROGMEM = "show logger commands";

const char Commands::MENU_HELP_05[] PROGMEM = "helpmonitor";
const char Commands::TEXT_HELP_05[] PROGMEM = "show monitoring commands";

const char Commands::MENU_HELP_06[] PROGMEM = "helplcdmode";
const char Commands::TEXT_HELP_06[] PROGMEM = "show lcd display mode commands";

const char Commands::MENU_HELP_07[] PROGMEM = "helpmisc";
const char Commands::TEXT_HELP_07[] PROGMEM = "show miscellaneous commands";

const char Commands::MENU_HELP_08[] PROGMEM = "helpdevel";
const char Commands::TEXT_HELP_08[] PROGMEM = "show developer commands";

// --- SENSORS -------------------------------------------------------------------
const char Commands::MENU_SENSORS_00[] PROGMEM = "temp";
const char Commands::TEXT_SENSORS_00[] PROGMEM = "show the current temperature";

const char Commands::MENU_SENSORS_01[] PROGMEM = "tempcalib";
const char Commands::TEXT_SENSORS_01[] PROGMEM = "show temperature calibration values";

const char Commands::MENU_SENSORS_02[] PROGMEM = "settempcaliblow f<ideal>";
const char Commands::TEXT_SENSORS_02[] PROGMEM = "set ideal temp calibration low value";

const char Commands::MENU_SENSORS_03[] PROGMEM = "settempcalibhigh f<ideal>";
const char Commands::TEXT_SENSORS_03[] PROGMEM = "set ideal temp calibration high value";

const char Commands::MENU_SENSORS_04[] PROGMEM = "conductivity";
const char Commands::TEXT_SENSORS_04[] PROGMEM = "show the current electrical conductivity";

const char Commands::MENU_SENSORS_05[] PROGMEM = "conductivitycalib";
const char Commands::TEXT_SENSORS_05[] PROGMEM = "show conductivity calibration value";

const char Commands::MENU_SENSORS_06[] PROGMEM = "setconductivitycalib f<ideal>";
const char Commands::TEXT_SENSORS_06[] PROGMEM = "set ideal conductivity calibration value";

const char Commands::MENU_SENSORS_07[] PROGMEM = "turbidity";
const char Commands::TEXT_SENSORS_07[] PROGMEM = "show the current turbidity";

const char Commands::MENU_SENSORS_08[] PROGMEM = "turbiditycalib";
const char Commands::TEXT_SENSORS_08[] PROGMEM = "show turbidity calibration values";

// --- DATE / TIME ---------------------------------------------------------------
const char Commands::MENU_DATETIME_00[] PROGMEM = "date";
const char Commands::TEXT_DATETIME_00[] PROGMEM = "show the current RTC date";

const char Commands::MENU_DATETIME_01[] PROGMEM = "time";
const char Commands::TEXT_DATETIME_01[] PROGMEM = "show the current RTC time";

const char Commands::MENU_DATETIME_02[] PROGMEM = "setdate i<dd> i<mm> i<yyyy>";
const char Commands::TEXT_DATETIME_02[] PROGMEM = "set the RTC date";

const char Commands::MENU_DATETIME_03[] PROGMEM = "settime i<hh> i<mm> i<ss>";
const char Commands::TEXT_DATETIME_03[] PROGMEM = "set the RTC time";

// --- LOGFILE -------------------------------------------------------------------
const char Commands::MENU_LOGFILE_00[] PROGMEM = "logfilename";
const char Commands::TEXT_LOGFILE_00[] PROGMEM = "show name of logfile";

const char Commands::MENU_LOGFILE_01[] PROGMEM = "setlogfilename \"<name>\"";
const char Commands::TEXT_LOGFILE_01[] PROGMEM = "set name of logfile";

const char Commands::MENU_LOGFILE_02[] PROGMEM = "logfilecontents";
const char Commands::TEXT_LOGFILE_02[] PROGMEM = "show logfile (partial) contents";

// --- LOGGER --------------------------------------------------------------------
const char Commands::MENU_LOGGER_00[] PROGMEM = "setloggerstarttime i<hh> i<mm>";
const char Commands::TEXT_LOGGER_00[] PROGMEM = "set logging start time for date today";

const char Commands::MENU_LOGGER_01[] PROGMEM = "setloggerstartdate i<dd> i<mm> i<yyyy>";
const char Commands::TEXT_LOGGER_01[] PROGMEM = "set logging start date if not today";

const char Commands::MENU_LOGGER_02[] PROGMEM = "setloggerstoptime i<hh> i<mm>";
const char Commands::TEXT_LOGGER_02[] PROGMEM = "set logging stop time for date today";

const char Commands::MENU_LOGGER_03[] PROGMEM = "setloggerstopdate i<dd> i<mm> i<yyyy>";
const char Commands::TEXT_LOGGER_03[] PROGMEM = "set logging stop date if not today";

const char Commands::MENU_LOGGER_04[] PROGMEM = "setloggerinterval i<hh> i<mm> i<ss>";
const char Commands::TEXT_LOGGER_04[] PROGMEM = "set logging interval";

const char Commands::MENU_LOGGER_05[] PROGMEM = "setloggerintervaldays i<dd>";
const char Commands::TEXT_LOGGER_05[] PROGMEM = "set days of logging interval";

const char Commands::MENU_LOGGER_06[] PROGMEM = "resetlogger";
const char Commands::TEXT_LOGGER_06[] PROGMEM = "reset all logging times and dates";

const char Commands::MENU_LOGGER_07[] PROGMEM = "loggersettings";
const char Commands::TEXT_LOGGER_07[] PROGMEM = "show logging settings";

const char Commands::MENU_LOGGER_08[] PROGMEM = "setloggertempon";
const char Commands::TEXT_LOGGER_08[] PROGMEM = "include temp in logging";

const char Commands::MENU_LOGGER_09[] PROGMEM = "setloggertempoff";
const char Commands::TEXT_LOGGER_09[] PROGMEM = "exclude temp from logging";

const char Commands::MENU_LOGGER_10[] PROGMEM = "setloggerconductivityon";
const char Commands::TEXT_LOGGER_10[] PROGMEM = "include conductivity in logging";

const char Commands::MENU_LOGGER_11[] PROGMEM = "setloggerconductivityoff";
const char Commands::TEXT_LOGGER_11[] PROGMEM = "exclude conductivity from logging";

const char Commands::MENU_LOGGER_12[] PROGMEM = "setloggerturbidityon";
const char Commands::TEXT_LOGGER_12[] PROGMEM = "include turbidity in logging";

const char Commands::MENU_LOGGER_13[] PROGMEM = "setloggerturbidityoff";
const char Commands::TEXT_LOGGER_13[] PROGMEM = "exclude turbidity from logging";

const char Commands::MENU_LOGGER_14[] PROGMEM = "loggerrun";
const char Commands::TEXT_LOGGER_14[] PROGMEM = "run logging";

const char Commands::MENU_LOGGER_15[] PROGMEM = "loggerhalt";
const char Commands::TEXT_LOGGER_15[] PROGMEM = "halt logging";

// --- MONITOR -------------------------------------------------------------------
const char Commands::MENU_MONITOR_00[] PROGMEM = "setmonitorstarttime i<hh> i<mm>";
const char Commands::TEXT_MONITOR_00[] PROGMEM = "set monitoring start time for date today";

const char Commands::MENU_MONITOR_01[] PROGMEM = "setmonitorstartdate i<dd> i<mm> i<yyyy>";
const char Commands::TEXT_MONITOR_01[] PROGMEM = "set monitoring start date if not today";

const char Commands::MENU_MONITOR_02[] PROGMEM = "setmonitorstoptime i<hh> i<mm>";
const char Commands::TEXT_MONITOR_02[] PROGMEM = "set monitoring stop time for date today";

const char Commands::MENU_MONITOR_03[] PROGMEM = "setmonitorstopdate i<dd> i<mm> i<yyyy>";
const char Commands::TEXT_MONITOR_03[] PROGMEM = "set monitoring stop date if not today";

const char Commands::MENU_MONITOR_04[] PROGMEM = "setmonitorinterval i<hh> i<mm> i<ss>";
const char Commands::TEXT_MONITOR_04[] PROGMEM = "set monitoring interval";

const char Commands::MENU_MONITOR_05[] PROGMEM = "addmonitorintervaldays i<dd>";
const char Commands::TEXT_MONITOR_05[] PROGMEM = "add days to monitoring interval";

const char Commands::MENU_MONITOR_06[] PROGMEM = "resetmonitor";
const char Commands::TEXT_MONITOR_06[] PROGMEM = "reset all monitoring times and dates";

const char Commands::MENU_MONITOR_07[] PROGMEM = "monitorsettings";
const char Commands::TEXT_MONITOR_07[] PROGMEM = "show monitoring settings";

const char Commands::MENU_MONITOR_08[] PROGMEM = "setmonitortempon";
const char Commands::TEXT_MONITOR_08[] PROGMEM = "include temp in monitoring";

const char Commands::MENU_MONITOR_09[] PROGMEM = "setmonitortempoff";
const char Commands::TEXT_MONITOR_09[] PROGMEM = "exclude temp from monitoring";

const char Commands::MENU_MONITOR_10[] PROGMEM = "setmonitorconductivityon";
const char Commands::TEXT_MONITOR_10[] PROGMEM = "include conductivity in monitoring";

const char Commands::MENU_MONITOR_11[] PROGMEM = "setmonitorconductivityoff";
const char Commands::TEXT_MONITOR_11[] PROGMEM = "exclude conductivity from monitoring";

const char Commands::MENU_MONITOR_12[] PROGMEM = "setmonitorturbidityon";
const char Commands::TEXT_MONITOR_12[] PROGMEM = "include turbidity in monitoring";

const char Commands::MENU_MONITOR_13[] PROGMEM = "setmonitorturbidityoff";
const char Commands::TEXT_MONITOR_13[] PROGMEM = "exclude turbidity from monitoring";

const char Commands::MENU_MONITOR_14[] PROGMEM = "monitorrun";
const char Commands::TEXT_MONITOR_14[] PROGMEM = "run monitoring";

const char Commands::MENU_MONITOR_15[] PROGMEM = "monitorhalt";
const char Commands::TEXT_MONITOR_15[] PROGMEM = "halt monitoring";

// ---LCD ---------- -------------------------------------------------------------
const char Commands::MENU_LCD_00[] PROGMEM = "setlcdfilename \"<name>\"";
const char Commands::TEXT_LCD_00[] PROGMEM = "set logfile name for lcd display mode ";

const char Commands::MENU_LCD_01[] PROGMEM = "setlcdloggerinterval i<mm> i<ss>";
const char Commands::TEXT_LCD_01[] PROGMEM = "set logging interval for lcd display mode (>= 10 sec)";

const char Commands::MENU_LCD_02[] PROGMEM = "setbacklighttimeout i<min>";
const char Commands::TEXT_LCD_02[] PROGMEM = "set timeout period for lcd backlight; -1: no timeout";

const char Commands::MENU_LCD_03[] PROGMEM = "lcdsettings";
const char Commands::TEXT_LCD_03[] PROGMEM = "show settings for lcd display mode";

// --- MISCELLANEOUS -------------------------------------------------------------
const char Commands::MENU_MISC_00[] PROGMEM = "version";
const char Commands::TEXT_MISC_00[] PROGMEM = "show version";

const char Commands::MENU_MISC_01[] PROGMEM = "cardinfo";
const char Commands::TEXT_MISC_01[] PROGMEM = "show SD card info";

const char Commands::MENU_MISC_02[] PROGMEM = "setechoon";
const char Commands::TEXT_MISC_02[] PROGMEM = "set echo input in monitor window";

const char Commands::MENU_MISC_03[] PROGMEM = "setechooff";
const char Commands::TEXT_MISC_03[] PROGMEM = "suppress echo in monitor window";

const char Commands::MENU_MISC_04[] PROGMEM = "serial";
const char Commands::TEXT_MISC_04[] PROGMEM = "communicate via Serial";

const char Commands::MENU_MISC_05[] PROGMEM = "wifi";
const char Commands::TEXT_MISC_05[] PROGMEM = "show wifi settings";

const char Commands::MENU_MISC_06[] PROGMEM = "setnet";
const char Commands::TEXT_MISC_06[] PROGMEM = "set wifi connection to local network";

const char Commands::MENU_MISC_07[] PROGMEM = "setap";
const char Commands::TEXT_MISC_07[] PROGMEM = "set wifi as access point";

const char Commands::MENU_MISC_08[] PROGMEM = "setnetcredentials \"<ssid>\" \"<pwd>\"";
const char Commands::TEXT_MISC_08[] PROGMEM = "set network credentials";

const char Commands::MENU_MISC_09[] PROGMEM = "setapcredentials \"<ssid>\" \"<pwd>\"";
const char Commands::TEXT_MISC_09[] PROGMEM = "set access point credentials";

const char Commands::MENU_MISC_10[] PROGMEM = "license";
const char Commands::TEXT_MISC_10[] PROGMEM = "show license text";

const char Commands::MENU_MISC_11[] PROGMEM = "poweroff";
const char Commands::TEXT_MISC_11[] PROGMEM = "power off the system";

// --- DEVEL ---------------------------------------------------------------------
const char Commands::MENU_DEVEL_00[] PROGMEM = "setmonitorsettings";
const char Commands::TEXT_DEVEL_00[] PROGMEM = "set test times for monitor testing";

const char Commands::MENU_DEVEL_01[] PROGMEM = "setloggersettings";
const char Commands::TEXT_DEVEL_01[] PROGMEM = "set test times for logger testing";

const char Commands::MENU_DEVEL_02[] PROGMEM = "testrunmonitor";
const char Commands::TEXT_DEVEL_02[] PROGMEM = "run monitor with all sensors on";

const char Commands::MENU_DEVEL_03[] PROGMEM = "testrunlogger";
const char Commands::TEXT_DEVEL_03[] PROGMEM = "run logger with all sensors on";

const char Commands::MENU_DEVEL_04[] PROGMEM = "copymonitordatestologger";
const char Commands::TEXT_DEVEL_04[] PROGMEM = "copy monitor dates and times to logger";

const char Commands::MENU_DEVEL_05[] PROGMEM = "fakesettempcaliblow f<sensor> f<ideal>";
const char Commands::TEXT_DEVEL_05[] PROGMEM = "fake and set temp calib low values";

const char Commands::MENU_DEVEL_06[] PROGMEM = "fakesettempcalibhigh f<sensor> f<ideal>";
const char Commands::TEXT_DEVEL_06[] PROGMEM = "fake and set temp calib high values";

const char Commands::MENU_DEVEL_07[] PROGMEM = "fakesetconductivitycalib f<sensor> f<ideal> f<temp>";
const char Commands::TEXT_DEVEL_07[] PROGMEM = "fake and set conductivity calib values";

const char Commands::MENU_DEVEL_08[] PROGMEM = "prefs";
const char Commands::TEXT_DEVEL_08[] PROGMEM = "show preferences";

const char Commands::MENU_DEVEL_09[] PROGMEM = "clearprefs";
const char Commands::TEXT_DEVEL_09[] PROGMEM = "CAUTION: clear all preferences";

const char Commands::MENU_DEVEL_10[] PROGMEM = "enableserial1";
const char Commands::TEXT_DEVEL_10[] PROGMEM = "enable Serial1 (after restart)";

const char Commands::MENU_DEVEL_11[] PROGMEM = "disableserial1";
const char Commands::TEXT_DEVEL_11[] PROGMEM = "disable Serial1 (after restart)";

const char Commands::MENU_DEVEL_12[] PROGMEM = "serial1";
const char Commands::TEXT_DEVEL_12[] PROGMEM = "communicate via Serial1";

const char Commands::MENU_DEVEL_13[] PROGMEM = "startlcdmode";
const char Commands::TEXT_DEVEL_13[] PROGMEM = "enable lcd display mode";

const char Commands::MENU_DEVEL_14[] PROGMEM = "stoplcdmode";
const char Commands::TEXT_DEVEL_14[] PROGMEM = "disable lcd display mode";

const char Commands::MENU_DEVEL_15[] PROGMEM = "startwifimode";
const char Commands::TEXT_DEVEL_15[] PROGMEM = "connect to wifi";

const char Commands::MENU_DEVEL_16[] PROGMEM = "stopwifimode";
const char Commands::TEXT_DEVEL_16[] PROGMEM = "deactivate wifi";

const char Commands::MENU_DEVEL_17[] PROGMEM = "directory";
const char Commands::TEXT_DEVEL_17[] PROGMEM = "show sd card root files";

const char Commands::MENU_DEVEL_18[] PROGMEM = "test0";
const char Commands::TEXT_DEVEL_18[] PROGMEM = "show temperature values";

const char Commands::MENU_DEVEL_19[] PROGMEM = "test1";
const char Commands::TEXT_DEVEL_19[] PROGMEM = "show conductivity values";

const char Commands::MENU_DEVEL_20[] PROGMEM = "test2";
const char Commands::TEXT_DEVEL_20[] PROGMEM = "show turbidity values";

const char Commands::MENU_DEVEL_21[] PROGMEM = "test3";
const char Commands::TEXT_DEVEL_21[] PROGMEM = "set LCD backlight on";

const char Commands::MENU_DEVEL_22[] PROGMEM = "test4";
const char Commands::TEXT_DEVEL_22[] PROGMEM = "set LCD backlight off";

const char Commands::MENU_DEVEL_23[] PROGMEM = "test5";
const char Commands::TEXT_DEVEL_23[] PROGMEM = "show system (esp32) date and time";

const char Commands::MENU_DEVEL_24[] PROGMEM = "test6";
const char Commands::TEXT_DEVEL_24[] PROGMEM = "calibrated TDS value";

const char Commands::MENU_DEVEL_25[] PROGMEM = "test7";
const char Commands::TEXT_DEVEL_25[] PROGMEM = "conductivity test";

const char Commands::MENU_DEVEL_26[] PROGMEM = "test8";
const char Commands::TEXT_DEVEL_26[] PROGMEM = "temperature test";

const char Commands::MENU_DEVEL_27[] PROGMEM = "test9 i<param>";
const char Commands::TEXT_DEVEL_27[] PROGMEM = "test parameter passing";

// -------------------------------------------------------------------------------

String Commands::getMenuTable(int i) {
  return FPSTR(MENU_TABLE[i]);
}


String Commands::getTextTable(int i) {
  return FPSTR(TEXT_TABLE[i]);
}


int Commands::find(String string) {
  for (unsigned int i = 0; i < TABLE_SIZE; ++ i) {
    String str = getMenuTable(i);
    int pos = str.indexOf(' ');
    if (pos != -1) {
      // Get rid of param information
      str = str.substring(0, pos);
    }
    if (str.equals(string)) {
      return i;
    }
  }
  return -1;
}


void Commands::executeFunction(int i) {
  Commands::FP_TABLE[i]();
}


// --- Command function maxros ---------------------------------------------------------------------
// The command functions are called via pointers.
// Up to 3 'parameters' can be passed to them via
// global arrays accessible within the functions.
// The macros test whether the required number and
// type of 'parameters' were passed. If not, the function
// is exited.
// An error message is issued to the active output.
// Functions with such parameter testing are named
// function PT.
#define RETURN_IF_PARAMS  \
  if (!Interpreter::expectedParams('0', '0', '0')) \
    return;

#define RETURN_IF_NOT_ONE_INT  \
  if (!Interpreter::expectedParams('i', '0', '0')) \
    return;

#define ONE_INT \
  Interpreter::intParams[0]

#define RETURN_IF_NOT_TWO_INTS  \
  if (!Interpreter::expectedParams('i', 'i', '0')) \
    return;

#define TWO_INTS \
  Interpreter::intParams[0], Interpreter::intParams[1]

#define RETURN_IF_NOT_THREE_INTS  \
  if (!Interpreter::expectedParams('i', 'i', 'i')) \
    return;

#define THREE_INTS \
  Interpreter::intParams[0], Interpreter::intParams[1], Interpreter::intParams[2]

#define RETURN_IF_NOT_ONE_FLOAT  \
  if (!Interpreter::expectedParams('f', '0', '0')) \
    return;

#define RETURN_IF_NOT_TWO_FLOATS  \
  if (!Interpreter::expectedParams('f', 'f', '0')) \
    return;

#define RETURN_IF_NOT_THREE_FLOATS  \
  if (!Interpreter::expectedParams('f', 'f', 'f')) \
    return;

#define RETURN_IF_NOT_ONE_STRING  \
  if (!Interpreter::expectedParams('"', '0', '0')) \
    return;

#define RETURN_IF_NOT_TWO_STRINGS  \
  if (!Interpreter::expectedParams('"', '"', '0')) \
    return;


// --- ---------------------------------------------------------------------------------------------
void Commands::printSpaces(int num) {
  for (int n = 0; n < num; ++n) {
    Config::outStream->print(' ');
  }
}


void Commands::printDots(int num) {
  if (num >= 3) {
    num -= 2;
  }
  Config::outStream->print(' ');
  for (int n = 0; n < num; ++n) {
    Config::outStream->print('.');
  }
  Config::outStream->print(' ');
}


void Commands::printHelpLine(int i, int tab) {
  String s = getMenuTable(i);
  Config::outStream->print(s);
  printDots(tab - s.length());
  s = getTextTable(i);
  Config::outStream->println(s);
}


// --- HELP ----------------------------------------------------------------------------
void Commands::printHelpPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== HELP ==="));
  for (unsigned int i = 0; i < HH_N; ++i) {
    printHelpLine(i, HELP_TAB);
  }
}


void Commands::printHelpSensorsPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== SENSORS ==="));
  for (unsigned int i = HH_N; i < HH_N + SE_N; ++i) {
    printHelpLine(i, SENSORS_TAB);
  }
}


void Commands::printHelpTimeDatePT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== DATE AND TIME ==="));
  for (
    unsigned int i = HH_N + SE_N;
                 i < HH_N + SE_N + DT_N;
    ++i)
  {
    printHelpLine(i, DATETIME_TAB);
  }
}


void Commands::printHelpLogfilePT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== LOGFILE ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N;
                 i < HH_N + SE_N + DT_N + LF_N;
    ++i)
  {
    printHelpLine(i, LOGFILE_TAB);
  }
}


void Commands::printHelpLoggerPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== LOGGER ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N + LF_N;
                 i < HH_N + SE_N + DT_N + LF_N + LO_N;
    ++i)
  {
    printHelpLine(i, LOGGER_TAB);
  }
}


void Commands::printHelpMonitorPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== MONITOR ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N + LF_N + LO_N;
                 i < HH_N + SE_N + DT_N + LF_N + LO_N + MO_N;
    ++i)
  {
    printHelpLine(i, MONITOR_TAB);
  }
}


void Commands::printHelpLcdPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== LCD DISPLAY MODE ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N + LF_N + LO_N + MO_N;
                 i < HH_N + SE_N + DT_N + LF_N + LO_N + MO_N + LB_N;
    ++i)
  {
    printHelpLine(i, LCD_TAB);
  }
}


void Commands::printHelpMiscPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== MISCELLANEOUS ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N + LF_N + LO_N + MO_N + LB_N;
                 i < HH_N + SE_N + DT_N + LF_N + LO_N + MO_N + LB_N + MS_N;
    ++i)
  {
    printHelpLine(i, MISC_TAB);
  }
}


void Commands::printHelpDevelPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("=== DEVELOPMENT ==="));
  for (
    unsigned int i = HH_N + SE_N + DT_N + LF_N + LO_N + MO_N + LB_N + MS_N;
                 i < HH_N + SE_N + DT_N + LF_N + LO_N + MO_N + LB_N + MS_N + DE_N;
    ++i)
  {
    printHelpLine(i, DEVEL_TAB);
  }
}


// --- SENSORS -------------------------------------------------------------------------
// Temperature
void Commands::temp() {
  Config::outStream->print(F("  "));
  Config::outStream->println(Temperature::helper.readString(true));
}


void Commands::printTempPT() {
  RETURN_IF_PARAMS
  printCurrentTime();
  temp();
}


void Commands::printTempCalibValuesPT() {
  RETURN_IF_PARAMS
  Temperature::helper.printCalibValues();
}



void Commands::setTempCalibLowValuesPT() {
  RETURN_IF_NOT_ONE_FLOAT
  // f<ideal>
  float idealLow = Interpreter::floatParams[0];
  Temperature::helper.setCalibLowValues(idealLow);
}


void Commands::setTempCalibHighValuesPT() {
  RETURN_IF_NOT_ONE_FLOAT
  // f<ideal>
  float idealHigh = Interpreter::floatParams[0];
  Temperature::helper.setCalibHighValues(idealHigh);
}


void Commands::printMonitorTemp() {
  if (Monitor::isTempOn()) {
    temp();
  }
}


// Conductivity (electrical) -----------------------------------------------------------
void Commands::conductivity() {
  Config::outStream->print(F("  "));
  Config::outStream->print(Conductivity::helper.readString(true));

  // Print at which temperature
  Config::outStream->print(F("  @ "));
  float idealTempValue = Temperature::helper.readCalibratedValue();
  Temperature::helper.printFormattedLn(idealTempValue, true);
}


void Commands::printConductivityPT() {
  RETURN_IF_PARAMS
  printCurrentTime();
  conductivity();
}


void Commands::printConductivityCalibValuesPT() {
  RETURN_IF_PARAMS
  Conductivity::helper.printCalibValues();
}


void Commands::setConductivityCalibValuesPT() {
  RETURN_IF_NOT_ONE_FLOAT
  // f<ideal>
  float idealLow = Interpreter::floatParams[0];
  Conductivity::helper.setCalibValues(idealLow);
}


void Commands::printMonitorConductivity() {
  if (Monitor::isConductivityOn()) {
    if (Monitor::isTempOn()) {
      Config::outStream->print(F("  "));
      Config::outStream->println(Conductivity::helper.readString(true));
    }
    else {
      conductivity();
    }
  }
}


// Turbidity ---------------------------------------------------------------------------
void Commands::turbidity() {
  Config::outStream->print(F("  "));
  Config::outStream->println(Turbidity::readString(true));
}


void Commands::printTurbidityPT() {
  RETURN_IF_PARAMS
  printCurrentTime();
  turbidity();
}


void Commands::printTurbidityCalibValuesPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("hard coded xyxcalibration table:"));
  Config::outStream->println(F("sensor voltage [V]:"));
  Turbidity::printVoltageArray();
  Config::outStream->println(F("Secchi disk visibility [cm]:"));
  Turbidity::printVisibilityArray();
}


void Commands::printMonitorTurbidity() {
  if (Monitor::isTurbidityOn()) {
    turbidity();
  }
}


// Date / Time -------------------------------------------------------------------------
void Commands::printCurrentTime() {
  // Save current time
  WbRTC::saveCurrentDateTime();
  delay(10);
  int n = WbRTC::savedDayOfTheWeek();
  char dayName[] = {
    Config::DAYS_OF_WEEK[3*n],
    Config::DAYS_OF_WEEK[3*n + 1],
    Config::DAYS_OF_WEEK[3*n + 2],
    '\0'
  };
  sprintf(
    buf,
    "%s, %2d/%02d/%4d  %2d:%02d:%02d",
    dayName,
    WbRTC::savedDay(), WbRTC::savedMonth(), WbRTC::savedYear(),
    WbRTC::savedHour(), WbRTC::savedMinute(), WbRTC::savedSecond()
  );
  Config::outStream->print(buf);
}
void Commands::printCurrentTimeLn() {
  printCurrentTime();
  Config::outStream->println();
}


void Commands::printDatePT() {
  RETURN_IF_PARAMS
  printCurrentTimeLn();
}


void Commands::printTimePT() {
  RETURN_IF_PARAMS
  printCurrentTimeLn();
}


void Commands::setDatePT() {
  RETURN_IF_NOT_THREE_INTS
  // Allow only if logger and/or monitor is/are not running
  if (!Logger::timer.isRunning() || !Monitor::timer.isRunning() || !LcdLogger::timer.isRunning()) {
    WbRTC::setDate(THREE_INTS);
    printCurrentTimeLn();
  }
  else {
    Config::outStream->print(F("error: (lcd) logger and/or monitor running"));
  }
}


void Commands::setTimePT() {
  RETURN_IF_NOT_THREE_INTS
  // Allow only if logger and/or monitor is/are not running
  if (!Logger::timer.isRunning() || !Monitor::timer.isRunning() || !LcdLogger::timer.isRunning()) {
    WbRTC::setTime(THREE_INTS);
    printCurrentTimeLn();
  }
  else {
    Config::outStream->print(F("error: (lcd) logger and/or monitor running"));
  }
}


// Logfile -----------------------------------------------------------------------------
void Commands::printLogfileNamePT() {
  RETURN_IF_PARAMS
  Config::outStream->println(Logger::getFilename());
}


void Commands::setLogfileNamePT() {
  RETURN_IF_NOT_ONE_STRING
  String filename = Interpreter::stringParams[0];
  filename.trim();
  String fn = filename;
  if (Logger::isFilenameCorrect(filename)) {
    Logger::setFilename(fn);
    Config::outStream->print(F("logfile name: "));
    Config::outStream->println(Logger::getFilename());
  }
  // incorrect case handled within isFilenameCorrect()
}


void Commands::printPartialLogfilePT() {
  RETURN_IF_PARAMS
  Logger::printPartialFile();
}


// Logger ------------------------------------------------------------------------------
void Commands::setLoggerStartTimeTodayPT() {
  RETURN_IF_NOT_TWO_INTS
  Logger::timer.setStartTimeToday(TWO_INTS);
}


void Commands::setLoggerStartDatePT() {
  RETURN_IF_NOT_THREE_INTS
  Logger::timer.setStartDate(THREE_INTS);
}


void Commands::setLoggerStopTimeTodayPT() {
  RETURN_IF_NOT_TWO_INTS
  Logger::timer.setStopTimeToday(TWO_INTS);
}


void Commands::setLoggerStopDatePT() {
  RETURN_IF_NOT_THREE_INTS
  Logger::timer.setStopDate(THREE_INTS);
}


void Commands::setLoggerIntervalPT() {
  RETURN_IF_NOT_THREE_INTS
  Logger::timer.setInterval(THREE_INTS);
}


void Commands::setLoggerIntervalDaysPT() {
  RETURN_IF_NOT_ONE_INT
  Logger::timer.setIntervalDays(ONE_INT);
}


void Commands::resetLoggerPT() {
  RETURN_IF_PARAMS
  Logger::timer.resetTimes();
}


void Commands::printLoggerRunning() {
  Config::outStream->println(F("logger running"));
}


void Commands::printLoggerSettingsPT() {
  RETURN_IF_PARAMS
  Config::outStream->print(F("log file: "));
  Config::outStream->println(Logger::getFilename());
  Logger::timer.printSettings();
  Logger::printTempSettings();
  Logger::printConductivitySettings();
  Logger::printTurbiditySettings();
}


void Commands::logTempOnPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setTempOn(true);
    Config::outStream->println(F("temp logging is on"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::logTempOffPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setTempOn(false);
    Config::outStream->println(F("temp logging is off"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::logConductivityOnPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setConductivityOn(true);
    Config::outStream->println(F("conductivity logging is on"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::logConductivityOffPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setConductivityOn(false);
    Config::outStream->println(F("conductivity logging is off"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::logTurbidityOnPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setTurbidityOn(true);
    Config::outStream->println(F("turbidity logging is on"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::logTurbidityOffPT() {
  RETURN_IF_PARAMS
  if (!Logger::timer.isRunning()) {
    Logger::setTurbidityOn(false);
    Config::outStream->println(F("turbidity logging is off"));
  }
  else {
    printLoggerRunning();
  }
}


void Commands::collectLogfileHeaders() {
  LogComposer::initHeaders();
  // Optional
  if (Logger::isTempOn()) {
    LogComposer::addHeader(Config::TEMP_HEADER);
  }
  if (Logger::isConductivityOn()) {
    LogComposer::addHeader(Config::CONDUCTIVITY_HEADER);
  }
  if (Logger::isTurbidityOn()) {
    LogComposer::addHeader(Config::TURBIDITY_HEADER);
  }
}


void Commands::loggerRunPT() {
  RETURN_IF_PARAMS
  bool before = Logger::timer.isRunning();   // ignore latency
  Logger::timer.run();                       // ignore latency

  bool after = Logger::timer.isRunning();
  if (!before && after) {
    // Create csv-file with file headers
    if (Logger::createFile(Logger::getFilename())) {
      collectLogfileHeaders();
      Logger::writeFileHeaders();
      Config::outStream->print(F("created log file "));
    }
    else {
      Config::outStream->print(F("error: could not create log file "));
    }
    Config::outStream->println(Logger::getFilename());
  }
}


void Commands::loggerHaltPT() {
  RETURN_IF_PARAMS
  Logger::timer.halt();
  Config::outStream->println(F("logger manually halted"));
}


// Monitor -----------------------------------------------------------------------------
void Commands::setMonitorStartTimeTodayPT() {
  RETURN_IF_NOT_TWO_INTS
  Monitor::timer.setStartTimeToday(TWO_INTS);
}


void Commands::setMonitorStartDatePT() {
  RETURN_IF_NOT_THREE_INTS
  Monitor::timer.setStartDate(THREE_INTS);
}


void Commands::setMonitorStopTimeTodayPT() {
  RETURN_IF_NOT_TWO_INTS
  Monitor::timer.setStopTimeToday(TWO_INTS);
}


void Commands::setMonitorStopDatePT() {
  RETURN_IF_NOT_THREE_INTS
  Monitor::timer.setStopDate(THREE_INTS);
}


void Commands::setMonitorIntervalPT() {
  RETURN_IF_NOT_THREE_INTS
  Monitor::timer.setInterval(THREE_INTS);
}


void Commands::setMonitorIntervalDaysPT() {
  RETURN_IF_NOT_ONE_INT
  Monitor::timer.setIntervalDays(ONE_INT);
}


void Commands::resetMonitorPT() {
  RETURN_IF_PARAMS
  Monitor::timer.resetTimes();
}


void Commands::printMonitorSettingsPT() {
  RETURN_IF_PARAMS
  Monitor::timer.printSettings();
  Monitor::printTempSettings();
  Monitor::printConductivitySettings();
  Monitor::printTurbiditySettings();
}


void Commands::monitorTempOnPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setTempOn(true);
  Config::outStream->println(F("temp monitoring is on"));
}


void Commands::monitorTempOffPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setTempOn(false);
  Config::outStream->println(F("temp monitoring is off"));
}


void Commands::monitorConductivityOnPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setConductivityOn(true);
  Config::outStream->println(F("conductivity monitoring is on"));
}


void Commands::monitorConductivityOffPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setConductivityOn(false);
  Config::outStream->println(F("conductivity monitoring is off"));
}


void Commands::monitorTurbidityOnPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setTurbidityOn(true);
  Config::outStream->println(F("turbidity monitoring is on"));
}


void Commands::monitorTurbidityOffPT() {
  RETURN_IF_PARAMS
  // Less restrictive than logger, allow changes while monitor is running
  Monitor::setConductivityOn(false);
  Config::outStream->println(F("turbidity monitoring is off"));
}


void Commands::monitorRunPT() {
  RETURN_IF_PARAMS
  Monitor::timer.run();
}


void Commands::monitorHaltPT() {
  RETURN_IF_PARAMS
  Monitor::timer.halt();
  Config::outStream->println(F("monitor manually halted"));
}


// LCD display and button mode ---------------------------------------------------------
void Commands::storeLcdLogfileNamePT() {
  RETURN_IF_NOT_ONE_STRING
  String filename = Interpreter::stringParams[0];
  filename.trim();
  if (LcdLogger::isFilenameCorrect(filename)) {
    LcdLogger::setFilename(filename);
    WbPreferences::setLcdLogfileName(filename);
    Config::outStream->print(F("lcd display mode logfile name: "));
    Config::outStream->println(LcdLogger::getFilename());
  }
  else {
    Config::outStream->print(F("error: incorrrect display mode logfile name: "));
    Config::outStream->println(filename);
  }
}


void Commands::printLcdInterval() {
  int seconds = WbPreferences::getLcdTimerInterval();
  int minutes = seconds / 60;
  if (minutes != 0) {
    Config::outStream->print(minutes);
    Config::outStream->print(F(" min "));
  }
  Config::outStream->print(seconds % 60);
  Config::outStream->println(F(" sec"));
}


void Commands::setLcdLoggerIntervalPT() {
  RETURN_IF_NOT_TWO_INTS
  // Allow only if lcd display mode logger is not running
  if (!LcdLogger::timer.isRunning()) {
    int seconds = Interpreter::intParams[0] * 60 + Interpreter::intParams[1];
    WbPreferences::setLcdTimerInterval(seconds);
    Config::outStream->print(F("lcd display mode logging interval: "));
    printLcdInterval();
  }
  else {
    Config::outStream->print(F("error: lcd logger running"));
  }
}


void Commands::setBacklightTimeoutPT() {
  RETURN_IF_NOT_ONE_INT
  // Allow only if lcd display mode logger is not running
  if (!LcdLogger::timer.isRunning()) {
    int minutes = Interpreter::intParams[0];
    WbPreferences::setBacklightTimeout(minutes);
    Config::outStream->print(F("lcd backlight timeout: "));
    minutes = WbPreferences::getBacklightTimeout();
    if (minutes < 0) {
      Config::outStream->println(F("no timeout"));
      return;
    }
    Config::outStream->print(minutes);
    Config::outStream->println(F(" min"));
    WbLCD::backlightOn();
  }
  else {
    Config::outStream->print(F("error: lcd logger running"));
  }
}


void Commands::printLcdSettingsPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("lcd display mode settings "));
  Config::outStream->print(F("file name: "));
  Config::outStream->println(LcdLogger::getFilename());
  Config::outStream->print(F("logging interval: "));
  printLcdInterval();
  Config::outStream->print(F("lcd backlight timeout: "));
  Config::outStream->print(WbPreferences::getBacklightTimeout());
  Config::outStream->println(F(" min, < 0: no timeout"));
  Config::outStream->println(F("button short-press: < 1 sec (not settable)"));
  Config::outStream->println(F("button long-press:  > 3 sec (not settable)"));
}


// Miscellaneous -----------------------------------------------------------------------
void Commands::printVersionInstitution() {
  Config::outStream->println(Config::VERSION);
  Config::outStream->println(Config::INSTITUTION);
}


void Commands::printVersionInstitutionPT() {
  RETURN_IF_PARAMS
  printVersionInstitution();
}


void Commands::printSdCardInfoPT() {
  RETURN_IF_PARAMS
  Logger::printSdCardInfo();
}


void Commands::setEchoOnPT() {
  RETURN_IF_PARAMS
  Interpreter::setEcho(true);
  Config::outStream->println(F("echo set on"));
}


void Commands::setEchoOffPT() {
  RETURN_IF_PARAMS
  Interpreter::setEcho(false);
  Config::outStream->println(F("echo set off"));
}


void Commands::enableSerial1PT() {
  RETURN_IF_PARAMS
  WbPreferences::setSerial1Switch(SERIAL1_ENABLED);
  Config::outStream->println(F("Serial1 enabled after restart"));
  Config::outStream->println(F("Make sure Serial1 is connected."));
}


void Commands::disableSerial1PT() {
  RETURN_IF_PARAMS
  WbPreferences::setSerial1Switch(SERIAL1_DISABLED);
  Config::outStream->println(F("Serial1 disabled after restart"));
}


void Commands::printConnectedTo(String str) {
  String s = F("Connected to ");
  Serial.println(s + str);
  if (WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) {
    Serial1.println(s + str);
  }
}


void Commands::connectSerial(String str, Stream* stream) {
  Config::inStream = stream;
  Config::outStream = stream;
  printConnectedTo(str);
}


void Commands::setSerial() {
  connectSerial(F("Serial"), &Serial);
  if (!WbServer::active) {
    LcdText::printSerialActive();
  }
}


void Commands::setSerialPT() {
  RETURN_IF_PARAMS
  setSerial();
}


void Commands::wifiPT() {
  RETURN_IF_PARAMS
  WbPreferences::printWifi();
}


void Commands::connectWifiPT() {
  RETURN_IF_PARAMS
  // TODO Test if alredy active
  WbServer::begin();
  Config::outStream->println(F("wifi connected"));

  // LcdLogger::stop();   // TODO
  // LcdButton::stop();

  LcdStatesDiagram::setActive(false);
}


// void Commands::disconnectWifiPT() {
//   // RETURN_IF_PARAMS
//   //TODO Test if disconnected
//   WbServer::disconnect();
//   Config::outStream->println(F("wifi disconnected"));

//   LcdLogger::begin();
//   LcdButton::begin();

//   LcdStatesDiagram::setActive(true);
//   LcdFSM::resetMillis();
// }
void Commands::disconnectWifiPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("Not implemented. Restart instead."));
}


void Commands::setNetPT() {
  RETURN_IF_PARAMS
  WbPreferences::setWifiMethod(NETWORK_METHOD);
  WbPreferences::printWifiMethod();
}


void Commands::setApPT() {
  RETURN_IF_PARAMS
  WbPreferences::setWifiMethod(ACCESSPOINT_METHOD);
  WbPreferences::printWifiMethod();
}


void Commands::setNetCredentialsPT() {
  RETURN_IF_NOT_TWO_STRINGS
  WbPreferences::setNetSsid(Interpreter::stringParams[0]);
  WbPreferences::setNetPwd(Interpreter::stringParams[1]);
  WbPreferences::printNetCredentials();
}


void Commands::setApCredentialsPT() {
  RETURN_IF_NOT_TWO_STRINGS
  WbPreferences::setApSsid(Interpreter::stringParams[0]);
  WbPreferences::setApPwd(Interpreter::stringParams[1]);
  WbPreferences::printApCredentials();
}


void Commands::printLicensePT() {
  RETURN_IF_PARAMS
  Config::outStream->print(F("MIT License\n"));
  
  Config::outStream->print(F("Copyright (c) 2021-23 ruedi.mueller@fhnw.ch\n\n"));

  Config::outStream->print(F("Permission is hereby granted, free of charge, to any person obtaining a copy "));
  Config::outStream->print(F("of this software and associated documentation files (the \"Software\"), to deal "));
  Config::outStream->print(F("in the Software without restriction, including without limitation the rights "));
  Config::outStream->print(F("to use, copy, modify, merge, publish, distribute, sublicense, and/or sell "));
  Config::outStream->print(F("copies of the Software, and to permit persons to whom the Software is "));
  Config::outStream->print(F("furnished to do so, subject to the following conditions:\n\n"));
  
  Config::outStream->print(F("The above copyright notice and this permission notice shall be included in all "));
  Config::outStream->print(F("copies or substantial portions of the Software.\n\n"));

  Config::outStream->print(F("THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR "));
  Config::outStream->print(F("IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, "));
  Config::outStream->print(F("FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE "));
  Config::outStream->print(F("AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER "));
  Config::outStream->print(F("LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, "));
  Config::outStream->print(F("OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE "));
  Config::outStream->print(F("SOFTWARE.\n"));
}


void Commands::poweroffPT() {
  RETURN_IF_PARAMS
  Config::outStream->println(F("Power off the system..."));
  toPowerOff = true;
}


// Devel -------------------------------------------------------------------------------
void Commands::setLoggerSettingsPT() {
  RETURN_IF_PARAMS
  WbRTC::saveCurrentDateTime();
  // Set time to start 1 min from now,
  TimeSpan offset = TimeSpan(0, 0, 1, 0);
  DateTime dateTime = WbRTC::getSavedCurrentDateTime() + offset;
  Logger::timer.setStartTimeToday(dateTime.hour(), dateTime.minute());
  Logger::timer.setStartDate(dateTime.day(), dateTime.month(), dateTime.year());

  // Set time to stop 3 min from now,
  offset = TimeSpan(0, 0, 3, 0);
  dateTime = WbRTC::getSavedCurrentDateTime() + offset;
  Logger::timer.setStopTimeToday(dateTime.hour(), dateTime.minute());
  Logger::timer.setStopDate(dateTime.day(), dateTime.month(), dateTime.year());

  // Set to interval 10 secs
  Logger::timer.setInterval(0, 0, 10);
}


void Commands::setMonitorSettingsPT() {
  RETURN_IF_PARAMS
  WbRTC::saveCurrentDateTime();
  // Set time to start 1 min from now,
  TimeSpan offset = TimeSpan(0, 0, 1, 0);
  DateTime dateTime = WbRTC::getSavedCurrentDateTime() + offset;
  // As for devel: Accept double printing of date and time
  Monitor::timer.setStartTimeToday(dateTime.hour(), dateTime.minute());
  Monitor::timer.setStartDate(dateTime.day(), dateTime.month(), dateTime.year());

  // Set time to stop 3 min from now,
  offset = TimeSpan(0, 0, 3, 0);
  dateTime = WbRTC::getSavedCurrentDateTime() + offset;
  // As for devel: Accept double printing of date and time
  Monitor::timer.setStopTimeToday(dateTime.hour(), dateTime.minute());
  Monitor::timer.setStopDate(dateTime.day(), dateTime.month(), dateTime.year());

  // Set to interval 10 secs
  Monitor::timer.setInterval(0, 0, 10);
}


void Commands::testrunMonitorPT() {
  RETURN_IF_PARAMS
  Monitor::setTempOn(true);
  Monitor::setConductivityOn(true);
  Monitor::setTurbidityOn(true);
  setMonitorSettingsPT();
  printMonitorSettingsPT();
  monitorRunPT();
  printCurrentTimeLn();
}


void Commands::testrunLoggerPT() {
  RETURN_IF_PARAMS
  Logger::setTempOn(true);
  Logger::setConductivityOn(true);
  Logger::setTurbidityOn(true);
  Logger::setFilename("logtst.csv");
  setLoggerSettingsPT();
  printLoggerSettingsPT();
  loggerRunPT();
  printCurrentTimeLn();
}


void Commands::copyMonitorDatesToLoggerPT() {
  RETURN_IF_PARAMS
  Logger::timer.setStartDateTime(Monitor::timer.getStartDateTime());
  Logger::timer.setStopDateTime(Monitor::timer.getStopDateTime());
  Logger::timer.setInterval(Monitor::timer.getInterval());
  printMonitorSettingsPT();
  printLoggerSettingsPT();
}


void Commands::fakesetTempCalibLowValuesPT() {
  RETURN_IF_NOT_TWO_FLOATS
  // f<sensor> f<ideal>
  float rawLow = Interpreter::floatParams[0];
  float idealLow = Interpreter::floatParams[1];
  Temperature::helper.fakesetCalibLowValues(rawLow, idealLow);
}


void Commands::fakesetTempCalibHighValuesPT() {
  RETURN_IF_NOT_TWO_FLOATS
  // f<sensor> f<ideal>
  float rawHigh = Interpreter::floatParams[0];
  float idealHigh = Interpreter::floatParams[1];
  Temperature::helper.fakesetCalibHighValues(rawHigh, idealHigh);
}


void Commands::fakesetConductivityCalibValuesPT() {
  RETURN_IF_NOT_THREE_FLOATS
  // f<sensor> f<ideal> f<temp>
  float raw = Interpreter::floatParams[0];
  float ideal = Interpreter::floatParams[1];
  float temp = Interpreter::floatParams[2];
  Conductivity::helper.fakesetCalibValues(raw, ideal, temp);
}


void Commands::printPrefsPT() {
  RETURN_IF_PARAMS
  WbPreferences::printPrefs();
}


void Commands::resetPrefsPT() {
  RETURN_IF_PARAMS
  WbPreferences::clearPrefs();
  Config::outStream->println(F("All prefs cleared"));
}


void Commands::setSerial1() {
  if (WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) {
    connectSerial(F("Serial1"), &Serial1);
    if (!WbServer::active) {
      LcdText::printSerial1Active();
    }
  }
  else {
    Config::outStream->println(F("invalid: Serial1 disabled in preferences"));
  }
}


void Commands::setSerial1PT() {
  RETURN_IF_PARAMS
  setSerial1();
}


void Commands::printMessage(String str) {
  Serial.println(str);
  if (WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) {
    Serial1.println(str);
  }
}


void Commands::startLcdPT() {
  RETURN_IF_PARAMS
  // Enable FSM
  LcdStatesDiagram::setActive(true);
  printMessage("lcd display mode enabled");
  LcdFSM::resetMillis();
}


void Commands::stopLcdPT() {
  RETURN_IF_PARAMS
  // Disable FSM
  LcdStatesDiagram::setActive(false);
  printMessage("lcd display mode disabled");
}


void Commands::directoryPT() {
  RETURN_IF_PARAMS
  Logger::directory();
}


void Commands::test0PT() {
  RETURN_IF_PARAMS
  printCurrentTimeLn();
  Config::outStream->print(F("uncalibrated current temperature value: "));
  float temp = Temperature::readRaw();
  Config::outStream->println(temp);
  Config::outStream->print(F("uncalibrated formatted current temperature value with units: "));
  Config::outStream->println(Temperature::formatString(temp, true));
  Config::outStream->print(F("uncalibrated formatted current temperature value without units: "));
  Config::outStream->println(Temperature::formatString(temp, false));
  Config::outStream->print(F("calibrated current temperature:"));
  Commands::temp();
  // Config::outStream->println(F("test0 inactive"));
}


void Commands::test1PT() {
  RETURN_IF_PARAMS
  printCurrentTimeLn();
  Config::outStream->print(F("uncalibrated current conductivity value: "));
  float cd = Conductivity::readRaw();
  Config::outStream->println(cd);
  Config::outStream->print(F("uncalibrated formatted current conductivity value with units: "));
  Config::outStream->println(Conductivity::formatString(cd, true));
  Config::outStream->print(F("uncalibrated formatted current conductivity value without units: "));
  Config::outStream->println(Conductivity::formatString(cd, false));
  Config::outStream->print(F("calibrated current conductivity, temperature compensated:"));
  conductivity();
  // Config::outStream->println(F("test1 inactive"));
}


void Commands::test2PT() {
  RETURN_IF_PARAMS
  printCurrentTimeLn();
  Config::outStream->print(F("raw voltage from sensor: "));
  Config::outStream->println(Turbidity::readRaw());
  Config::outStream->print(F("visibility interpolated: "));
  Config::outStream->println(Turbidity::depth());
  // Config::outStream->println(F("test2 inactive"));
}


void Commands::test3PT() {
  RETURN_IF_PARAMS
  WbLCD::backlightOn();
  Config::outStream->println(F("LCD backlight on"));
  // Config::outStream->println(F("test3 inactive"));
}


void Commands::test4PT() {
  RETURN_IF_PARAMS
  WbLCD::backlightOff();
  Config::outStream->println(F("LCD backlight off"));
  // Config::outStream->println(F("test4 inactive"));
}


void Commands::test5PT() {
  RETURN_IF_PARAMS
  WbRTC::printSystemTime();
  // Config::outStream->println(F("test5 inactive"));
}


void Commands::test6PT() {   // // TODO
  RETURN_IF_PARAMS
  Config::outStream->print(F("Calibrated TDS value: "));
  float temp = Temperature::readRaw();
  float tds = Conductivity::readTdsValue(temp);
  Config::outStream->print(tds);
  Config::outStream->print(" ppm  @ ");
  Config::outStream->print(temp);
  Config::outStream->println("°C");

  Config::outStream->print(F("Calibrated approx. electrical conductivity: "));
  float ec = 2 * tds;
  Config::outStream->print(ec);
  Config::outStream->println(" µS/cm");

  // Config::outStream->println(F("test6 inactive"));
}


#include "ConductivityTest.h"
void Commands::test7PT() {
  RETURN_IF_PARAMS
  ConductivityTest::measureTDS();
  Config::outStream->print("voltage: ");
  Config::outStream->print(ConductivityTest::getAverageVoltage(), 1);
  Config::outStream->print(" V   TDS value: ");
  Config::outStream->print(ConductivityTest::getTdsValue(), 0);
  Config::outStream->print(" ppm    @ ");
  Config::outStream->print(ConductivityTest::getTemperature(), 1);
  Config::outStream->println("°C");

  // // Config::outStream->println(F("test7 inactive"));
}


#include "TemperatureTest.h"
void Commands::test8PT() {
  RETURN_IF_PARAMS

  printTemp();
  // Config::outStream->println(F("test8 inactive"));
}


void Commands::test9PT() {
  RETURN_IF_NOT_ONE_INT
  Config::outStream->print(F("passed int parameter is "));
  Config::outStream->println(ONE_INT);
  // Config::outStream->println(F("test9 inactive"));
}
