/*
 * Interpreter.cpp
 *
 * Created on: 5 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "Interpreter.h"
#include "Commands.h"
#include "Config.h"
#include "InputStreamReader.h"
#include "Streaming.h"


String Interpreter::input;
String Interpreter::echo;
bool Interpreter::echoOn = true;
String Interpreter::changedInput;
String Interpreter::cmdWord;
const char Interpreter::SEPARATOR_TYPES[] = { 'i', 'f', '"' };
char Interpreter::separators[MAX];
unsigned char Interpreter::startLocations[MAX];
unsigned char Interpreter::locations[MAX];
int Interpreter::intParams[MAX];
float Interpreter::floatParams[MAX];
String Interpreter::stringParams[MAX];


void Interpreter::findCommandWord() {
  echo = input;
  if (input.length() == 0) {
    cmdWord = "";
    return;
  }
  int idx;
  if ((idx = input.indexOf(F(" "))) == -1) {
    // String alone
    cmdWord = input;
  }
  else {
    // More chars after string
    cmdWord = input.substring(0, idx);
  }
  cmdWord.trim();

  // Strip off command word for further parsing
  input = input.substring(idx);
}


void Interpreter::findSeparators() {
  memset(locations, -1, sizeof(locations));
  changedInput = input;
  bool exchange = false;

  // String manipulation
  for (unsigned int i = 0; i < changedInput.length(); ++i) {
    if (changedInput[i] == '"')   exchange = !exchange;
    // Exchange start-'"' and all string chars by '?' except end-'"'
    // Example: "mystring" --> ????????"
    if (exchange)   changedInput[i] = '?';
  }

  int occurrence = 0;
  memset(separators, '0', sizeof(separators));
  for (unsigned int i = 0; i < changedInput.length(); ++i) {
    // Check with all separator types
    for (int j = 0; j < MAX; ++j) {
      if (changedInput[i] == SEPARATOR_TYPES[j]) {
        if (occurrence == MAX) {
          // More than MAX separators found
          return;
        }
        // Memorise order and location of separators
        separators[occurrence] = SEPARATOR_TYPES[j];
        locations[occurrence++] = i;
      }
    }
  }
}


void Interpreter::findStringstartSeparators() {
  memset(startLocations, -1, sizeof(startLocations));
  int strPtr = 0;
  for (int i = 0; i < MAX; ++i) {
    if (separators[i] == '"') {
      bool first = true;
      for ( ; strPtr < changedInput.length(); ++strPtr) {
        if (changedInput[strPtr] == '?' && first) {
          // Start of string: first '?'
          startLocations[i] = strPtr;
          first = false;
        }
        if (changedInput[strPtr] == '"') {
          // End of string: trailing '"'
          first = true;
          strPtr++;
          break;
        }
      }
    }
  }
}


void Interpreter::findParameters() {
  memset(intParams, 0, sizeof(intParams));
  memset(floatParams, 0, sizeof(floatParams));
  for (int i = 0; i < MAX; ++i)   { stringParams[i] = ""; }
  for (int i = 0; i < MAX; ++i) {
    if (separators[i] == 'i') {
      intParams[i] = input.substring(locations[i] + 1).toInt();
    }
    if (separators[i] == 'f') {
      floatParams[i] = input.substring(locations[i] + 1).toFloat();
    }
    if (separators[i] == '"') {
      stringParams[i] = input.substring(startLocations[i] + 1, locations[i]);
    }
  }
}


void Interpreter::executeCommand() {
  int cmdIdx = 0;
  if ((cmdIdx = Commands::find(cmdWord)) >= 0) {
    // Used for debugging only:
    // Serial << F("cmdIdx: ") << cmdIdx << endl;
    // Serial << F("getMenuTable(cmdIdx): ") << Commands::getMenuTable(cmdIdx) << endl;
    Commands::executeFunction(cmdIdx);
  }
  else {
    if (cmdWord.length() == 0) {
      Config::outStream->println();
    }
    else {
      printNoSuchCommand();
    }
  }
}


void Interpreter::interpret() {
  #if defined INTERPRETER_TESTING
  Serial << F("input: ") << input << endl;
  #endif  // INTERPRETER_TESTING
  // Parse word command
  findCommandWord();
  #if defined INTERPRETER_TESTING
  Serial << F("cmdWord: ") << cmdWord << endl;
  #endif  // INTERPRETER_TESTING
  findSeparators();
  findStringstartSeparators();
  #if defined INTERPRETER_TESTING
  printSeparatorInfo();
  #endif  // INTERPRETER_TESTING
  findParameters();
  #if defined INTERPRETER_TESTING
  printParameterInfo();
  #endif  // INTERPRETER_TESTING
  if (echoOn) {
    Config::outStream->print(F(">>> "));
    Config::outStream->println(echo);
  }
  executeCommand();
}


void Interpreter::setInput(String command) {
  command.trim();
  input = command;

  if (input.endsWith(F("\r"))) {
    input = input.substring(0, input.length() - 1);
    input.trim();
  }
}


void Interpreter::readAndInterpret() {
  bool hasInput = false;
  // Read input (from monitor, terminal program, app) non-blocking,
  // i.e. zero or one char at a time.
  // Input is only accepted from current inStream!
  InputStreamReader::readInput();
  if (InputStreamReader::isComplete()) {
    input = InputStreamReader::getInput();
    hasInput = true;
  }

  // Check if complete command is available
  if (hasInput) {
    input.trim();
    // Accept 'Both NL & CR', 'Carriage return' or 'Newline'
    // when using Arduino Serial Monitor on Ubuntu
    if (input.endsWith(F("\r"))) {
      input = input.substring(0, input.length() - 1);
      input.trim();
    }
    interpret();
  }
}


void Interpreter::printNoSuchCommand() {
  Config::outStream->print(F("unknown command: "));
  Config::outStream->println(cmdWord);
}


void Interpreter::printRequiredParams(char c) {
  if (c == 'i')   Config::outStream->print(F("i<int> "));
  else if (c == 'f')   Config::outStream->print(F("f<float> "));
  else if (c == '"')   Config::outStream->print(F("\"<string>\" "));
}


bool Interpreter::expectedParams(char c0, char c1, char c2) {
  if (separators[0] != c0 || separators[1] != c1 || separators[2] != c2) {
    Config::outStream->print(F("Parameter error: <cmd> "));
    printRequiredParams(c0);
    printRequiredParams(c1);
    printRequiredParams(c2);
    Config::outStream->println(F("required"));
    return false;
  }
  return true;
}


void Interpreter::setEcho(bool flag) {
  echoOn = flag;
}


bool Interpreter::isEchoOn() {
  return echoOn;
}


#if defined INTERPRETER_TESTING
void Interpreter::space() {
  Serial << F(" ");
}


void Interpreter::printSeparators() {
  Serial << F("separators[]: ");
  Serial << separators[0]; space();
  Serial << separators[1]; space();
  Serial << separators[2] << endl;
}


void Interpreter::printSeparatorInfo() {
  Serial.println(F("---------------"));
  Serial.print(F("input: ")); Serial.println(input);
  Serial.print(F("changedInput: ")); Serial.println(changedInput);
  printSeparators();
  Serial.print(F("startLocations[]: "));
  Serial.print(startLocations[0]); space();
  Serial.print(startLocations[1]); space();
  Serial.println(startLocations[2]);
  Serial.print(F("locations[]: "));
  Serial.print(locations[0]); space();
  Serial.print(locations[1]); space();
  Serial.println(locations[2]);
  Serial.println(F("---------------"));
}


void Interpreter::printParameterInfo() {
  Serial.print(F("intParams[]: "));
  Serial.print(intParams[0]); space();
  Serial.print(intParams[1]); space();
  Serial.println(intParams[2]);
  Serial.print(F("floatParams[]: "));
  Serial.print(floatParams[0]); space();
  Serial.print(floatParams[1]); space();
  Serial.println(floatParams[2]);
  Serial.print(F("stringParams[]: \""));
  Serial.print(stringParams[0]);
  Serial.print(F("\" \""));
  Serial.print(stringParams[1]);
  Serial.print(F("\" \""));
  Serial.print(stringParams[2]);
  Serial.println(F("\""));
}
#endif  // INTERPRETER_TESTING
