/*
 * LcdLogger.cpp
 *
 * Created on: 20 May 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "LcdLogger.h"
#include <Arduino.h>
#include "Streaming.h"
#include "Config.h"
#include <SD.h>
#include "WbPreferences.h"
#include "WbRTC.h"
#include "WbLCD.h"
#include "LcdFSM.h"
#include "LcdState.h"
#include "LcdStatesDiagram.h"
#include "LcdLogComposer.h"
#include "LcdText.h"
#include "LcdButton.h"
#include "LcdSensorValuesCollector.h"
#include "WbPreferences.h"
#include "Temperature.h"
#include "Turbidity.h"
#include "Conductivity.h"

#define LOCKING_INTERVAL 3000   // Time in millis to wait until LcdButton is unlocked


// Class variables
char LcdLogger::buf[20];
File LcdLogger::root;
Timer LcdLogger::timer = Timer(PSTR("lcd display mode logger"));
String LcdLogger::filename;
File LcdLogger::logfile;
bool LcdLogger::enabled;
bool LcdLogger::buttonLocked = false;
long LcdLogger::lockingTimepoint;


void LcdLogger::begin() {
  // Set filename from preferences
  filename = WbPreferences::getLcdLogfileName();
  Serial << F("LCD logfile set: ") << filename << endl;
  // Set timer preferences
  timer.setPrefs();
  // Init state connections, conditions, transition actions and array lengths
  LcdStatesDiagram::initStates();
  Serial << F("FSM states initialised\n");
  // Start FSM
  LcdFSM::begin(&LcdStatesDiagram::lgSplash);
  LcdFSM::setConditionFunction(LcdFSM::millisCondFunction);
  Serial << F("FSM set\n");
  // Display LCD welcome lines
  LcdText::pSplashScreen();
  Serial << F("LCD welcome screen written\n");
}


bool LcdLogger::isFilenameCorrect(String filename) {
  int n = filename.indexOf(PSTR("."));
  if (n != -1) {
    // Extension
    String ext = filename.substring(n);
    // Serial << F("ext: ") << ext << F(", ext.length: ") << ext.length() << endl;   // DEBUG
    if (ext.length() > 4) {
      Config::outStream->printf(PSTR("error: '%s' extension is > 3 chars long\n"), filename.c_str());
      return false;
    }
  }
  else {
    // No extension
    n = filename.length();
  }
  String name = filename.substring(0, n);
  if (name.length() < 1 || name.length() > 8) {
    Config::outStream->printf(PSTR("error: '%s' is < 1 or > 8 chars long\n"), filename.c_str());
    return false;
  }
  return true;
}


bool LcdLogger::createFile() {
  String filename = WbPreferences::getLcdLogfileName();
  int n = filename.indexOf(PSTR("."));
  if (n == -1) {
    n = filename.length();
  }

  // Format: name
  String name = filename.substring(0, n);
  // Format: .ext
  String dotExt = filename.substring(n, n + 4);

  if (n == 8) {
    name = name.substring(0, n - 2);
  }
  else if (n == 7) {
    if (isDigit(name.charAt(n - 1))) {
      name = name.substring(0, n - 1);
      if (isDigit(name.charAt(n - 2))) {
        name = name.substring(0, n - 2);
      }
    }
    else {
      name = name.substring(0, n - 1);
    }
  }
  else if (n > 2) {
    // Strip off up to two trailing digits
    if (isDigit(name.charAt(n - 1))) {
      name = name.substring(0, n - 1);
      if (isDigit(name.charAt(n - 2))) {
        name = name.substring(0, n - 2);
      }
    }
  }

  // Create a file with a new number < 100 like
  // /name00.ext .. /name09.ext, /name10.ext .. /name99.ext
  int i = 0;
  while (i < 100) {
    filename = name + (i < 10 ? "0" : "") + i + dotExt;
    // Increment for next loop, or to set i > 99
    i++;
    if (!SD.exists(String('/') + filename)) {
      // Found inexistent file name
      break;
    }
    else {
      // Serial << F("File exists: ") << filename << endl;   // DEBUG
      continue;
    }
  }
  // Filename candidate
  LcdLogger::filename = filename;

  // Number within range?
  if (i > 99) {
    Config::outStream->printf(PSTR("error: '%s' number > 99\n"), filename.c_str());
    return false;
  }

  // Try to create file
  logfile = SD.open(String('/') + filename, FILE_WRITE);
  if (!logfile) {
    // Cannot open: write eror
    Config::outStream->printf(PSTR("error: could not create '%s'\n"), filename.c_str());
    return false;
  }

  // Written, all ok
  logfile.close();
  return true;
}


void LcdLogger::writeFileHeaders() {
  String headersString = LcdLogComposer::getHeadersString();
  writeFile(headersString);
}


String LcdLogger::getFilename() {
  return filename;
}


void LcdLogger::setFilename(String filename) {
  LcdLogger::filename = filename;
}


void LcdLogger::writeFile(String data) {
  File file = SD.open(String('/') + filename, FILE_APPEND);
  if (file) {
    file.println(data);
    file.close();
  }
  else {
    Config::outStream->printf(PSTR("could not write to file %s\n"), filename.c_str());
  }
}


void LcdLogger::writeFileValues() {
  LcdLogComposer::initValues();
  // Current date and time added here
  WbRTC::saveCurrentDateTime();
  LcdLogComposer::addValue(WbRTC::getSavedDateString());
  LcdLogComposer::addValue(WbRTC::getSavedTimeString());

  LcdLogComposer::addValue(Temperature::helper.readString(false));
  LcdLogComposer::addValue(Conductivity::helper.readString(false));
  LcdLogComposer::addValue(Turbidity::readString(false));

  String valuesString = LcdLogComposer::getValuesString();
  writeFile(valuesString);
}


void LcdLogger::enable() {
  enabled = true;
}


void LcdLogger::disable() {
  enabled = false;
}


bool LcdLogger::isEnabled() {
  return enabled;
}


void LcdLogger::handleLcdDisplayMode() {
  // If LCD display and button are active execute FSM to set lcd display scenario
  // in action.
  if (LcdStatesDiagram::isActive()) {
    LcdStatesDiagram::fsmExecute();

    if (LcdLogger::timer.isRunning() && LcdLogger::timer.isTimeToLog()) {
      // Enable writing of all 3 (collected) sensor values.
      // Writing is done *within* the fsm for synchronising reasons
      LcdLogger::enable();
    }

    if (LcdStatesDiagram::canCollect()) {
      // Collect sensor data per sensor in a round robin manner once the fsm
      // time slot is open.
      // The slot is opened by a time event in FsmStatesDiagram and closed
      // there as well.
      LcdSensorValuesCollector::collect();
    }

    if (buttonLocked) {
      // LcdButton was locked by LcdText::p0Countdown(). Check if it can be unlocked
      // after the defined locking interval.
      if ((millis() - lockingTimepoint) > LOCKING_INTERVAL) {
        // Unlock LcdButton
        LcdButton::enable();
        buttonLocked = false;
      }
    }
  }
  WbLCD::backlightCheck();
}
