/*
 * LcdStatesDiagram.cpp
 *
 * Created on: 19 May 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "LcdStatesDiagram.h"
#include "LcdText.h"
#include "LcdState.h"
#include "LcdButton.h"
#include "LcdButtonTimer.h"
#include "LcdFSM.h"
#include "LcdLogger.h"
#include "Streaming.h"
#include "Config.h"


#define L LcdText
#define S LcdStatesDiagram


// State transition condition functions
// These functions can be exchanged using
// FSM::setConditionFunction(buttonShortCondFunction);

// Button press codes
//                       // Code -1 reserved! Do not use!
#define SHORT_PRESS -2   // -2, -3, -4, -5: 'unreachable' millis numbers
#define LONG_PRESS -3    // used as code for
#define RELEASED -4      // button events
#define WRITE_FILE -5    // and write file event


bool LcdStatesDiagram::active = false;
bool LcdStatesDiagram::collect = false;


// Define global states
LcdState LcdStatesDiagram::lgSplash(PSTR("lg_splash"));

LcdState LcdStatesDiagram::lgWifi(PSTR("lg_wifi"));

LcdState LcdStatesDiagram::lgReady1(PSTR("lg_ready_1"));
LcdState LcdStatesDiagram::lgReady2(PSTR("lg_ready_2"));
LcdState LcdStatesDiagram::lgReady3(PSTR("lg_ready_3"));
LcdState LcdStatesDiagram::lgReady4(PSTR("lg_ready_4"));

LcdState LcdStatesDiagram::lgPwrOff1(PSTR("lg_pwr_off_1"));
LcdState LcdStatesDiagram::lgPwrOff2(PSTR("lg_pwr_off_2"));
LcdState LcdStatesDiagram::lgPwrOff3(PSTR("lg_pwr_off_3"));
LcdState LcdStatesDiagram::lgPwrOff4(PSTR("lg_pwr_off_4"));

LcdState LcdStatesDiagram::lgRunBlink1(PSTR("lg_blink_1"));
LcdState LcdStatesDiagram::lgRunBlink2(PSTR("lg_blink_2"));
LcdState LcdStatesDiagram::lgRunBlink3(PSTR("lg_blink_3"));

LcdState LcdStatesDiagram::lgPause1(PSTR("lg_pause_1"));
LcdState LcdStatesDiagram::lgPause2(PSTR("lg_pause_2"));

LcdState LcdStatesDiagram::lgRunCountdown(PSTR("lg_run_countdown"));
LcdState LcdStatesDiagram::lgPauseCountdown(PSTR("lg_pause_countdown"));


// Set up Diagram
// CAUTION: States which include RELEASED conditions *MUST* be included in LcdStatesDiagram::fsmExecute()

// State 'lgSplash' arrays                                      waited first 1000 ms to detect long press
LcdState* S::lgSplash_connectedStates[] = { &lgReady1,          &lgWifi };
int S::lgSplash_conditions[] =            { SHORT_PRESS,        LONG_PRESS };
void (* S::lgSplash_actions[])() =        { L::pLgReadyScreen1, offZeroMillis };
#define LG_SPLASH (sizeof(lgSplash_conditions) / sizeof(lgSplash_conditions[0]))

// State 'lgWifi' arrays                                                                                 waited <8 secs
LcdState* S::lgWifi_connectedStates[] = { &lgWifi, &lgWifi, &lgWifi, &lgWifi, &lgWifi, &lgWifi, &lgWifi, &lgSplash,        &lgSplash };
int S::lgWifi_conditions[] =            { 2000,    3000,    4000,    5000,    6000,    7000,    8000,    9000,             RELEASED };
void (* S::lgWifi_actions[])() =        { on,      off,     on,      off,     on,      off,     on,      L::pLgWifiScreen, off };
#define LG_WIFI (sizeof(lgWifi_conditions) / sizeof(lgWifi_conditions[0]))

// State 'lgReady1' arrays                                                                            waited first 1000 ms to detect long press
LcdState* S::lgReady1_connectedStates[] = { &lgReady1,  &lgReady2,          &lgRunBlink1,             &lgPwrOff1 };
int S::lgReady1_conditions[] =            { 200,        2000,               SHORT_PRESS,              LONG_PRESS };
void (* S::lgReady1_actions[])() =        { collectOff, L::pLgReadyScreen2, L::pLgRunScreenWriteFile, L::pLgReadyScreen2 };
#define LG_READY_1 (sizeof(lgReady1_conditions) / sizeof(lgReady1_conditions[0]))

// State 'lgReady2' arrays                                                                            waited first 1000 ms to detect long press
LcdState* S::lgReady2_connectedStates[] = { &lgReady2,  &lgReady3,          &lgRunBlink1,             &lgPwrOff2 };
int S::lgReady2_conditions[] =            { 200,        2000,               SHORT_PRESS,              LONG_PRESS };
void (* S::lgReady2_actions[])() =        { collectOff, L::pLgReadyScreen3, L::pLgRunScreenWriteFile, L::pLgReadyScreen3 };
#define LG_READY_2 (sizeof(lgReady2_conditions) / sizeof(lgReady2_conditions[0]))

// State 'lgReady3' arrays                                                                            waited first 1000 ms to detect long press
LcdState* S::lgReady3_connectedStates[] = { &lgReady3,  &lgReady4,          &lgRunBlink1,             &lgPwrOff3 };
int S::lgReady3_conditions[] =            { 200,        2000,               SHORT_PRESS,              LONG_PRESS };
void (* S::lgReady3_actions[])() =        { collectOff, L::pLgReadyScreen2, L::pLgRunScreenWriteFile, L::pLgReadyScreen2 };
#define LG_READY_3 (sizeof(lgReady3_conditions) / sizeof(lgReady3_conditions[0]))

// State 'lgReady4' arrays                                                                            waited first 1000 ms to detect long press
LcdState* S::lgReady4_connectedStates[] = { &lgReady4,  &lgReady1,          &lgRunBlink1,             &lgPwrOff4 };
int S::lgReady4_conditions[] =            { 200,        2000,               SHORT_PRESS,              LONG_PRESS };
void (* S::lgReady4_actions[])() =        { collectOff, L::pLgReadyScreen1, L::pLgRunScreenWriteFile, L::pLgReadyScreen1 };
#define LG_READY_4 (sizeof(lgReady4_conditions) / sizeof(lgReady4_conditions[0]))

// State 'lgPwrOff1' arrays                                                                                                                                          waited <8 secs
LcdState* S::lgPwrOff1_connectedStates[] = { &lgPwrOff1,          &lgPwrOff1, &lgPwrOff1,          &lgPwrOff1, &lgPwrOff1,          &lgPwrOff1, &lgPwrOff1,          &lgSplash, &lgReady1 };
int S::lgPwrOff1_conditions[] =            { 2000,                3000,       4000,                5000,       6000,                7000,       8000,                9000,      RELEASED };
void (* S::lgPwrOff1_actions[])() =        { L::pLgPwrOffScreen3, pwr,        L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen1, pwr,        L::pLgPwrOffScreen2, poweroff,  L::pLgReadyScreen3 };
#define LG_PWR_OFF_1 (sizeof(lgPwrOff1_conditions) / sizeof(lgPwrOff1_conditions[0]))

// State 'lgPwrOff2' arrays                                                                                                                                          waited <8 secs
LcdState* S::lgPwrOff2_connectedStates[] = { &lgPwrOff2,          &lgPwrOff2, &lgPwrOff2,          &lgPwrOff2, &lgPwrOff2,          &lgPwrOff2, &lgPwrOff2,          &lgSplash, &lgReady2 };
int S::lgPwrOff2_conditions[] =            { 2000,                3000,       4000,                5000,       6000,                7000,       8000,                9000,      RELEASED };
void (* S::lgPwrOff2_actions[])() =        { L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen1, pwr,        L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen3, poweroff,  L::pLgReadyScreen2 };
#define LG_PWR_OFF_2 (sizeof(lgPwrOff2_conditions) / sizeof(lgPwrOff2_conditions[0]))

// State 'lgPwrOff3' arrays                                                                                                                                          waited <8 secs
LcdState* S::lgPwrOff3_connectedStates[] = { &lgPwrOff3,          &lgPwrOff3, &lgPwrOff3,          &lgPwrOff3, &lgPwrOff3,          &lgPwrOff3, &lgPwrOff3,          &lgSplash, &lgReady3 };
int S::lgPwrOff3_conditions[] =            { 2000,                3000,       4000,                5000,       6000,                7000,       8000,                9000,      RELEASED };
void (* S::lgPwrOff3_actions[])() =        { L::pLgPwrOffScreen1, pwr,        L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen3, pwr,        L::pLgPwrOffScreen2, poweroff,  L::pLgReadyScreen1 };
#define LG_PWR_OFF_3 (sizeof(lgPwrOff3_conditions) / sizeof(lgPwrOff3_conditions[0]))

// State 'lgPwrOff4' arrays                                                                                                                                          waited <8 secs
LcdState* S::lgPwrOff4_connectedStates[] = { &lgPwrOff4,          &lgPwrOff4, &lgPwrOff4,          &lgPwrOff4, &lgPwrOff4,          &lgPwrOff4, &lgPwrOff4,          &lgSplash, &lgReady4 };
int S::lgPwrOff4_conditions[] =            { 2000,                3000,       4000,                5000,       6000,                7000,       8000,                9000,      RELEASED };
void (* S::lgPwrOff4_actions[])() =        { L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen3, pwr,        L::pLgPwrOffScreen2, pwr,        L::pLgPwrOffScreen1, poweroff,  L::pLgReadyScreen2 };
#define LG_PWR_OFF_4 (sizeof(lgPwrOff4_conditions) / sizeof(lgPwrOff4_conditions[0]))

// State 'lgRunBlink1' arrays                                off                                  write data                                waited first 1000 ms to detect long press
LcdState* S::lgRunBlink1_connectedStates[] = { &lgRunBlink1, &lgRunBlink1,      &lgRunBlink2,     &lgRunBlink1,               &lgPause1,    &lgRunCountdown };
int S::lgRunBlink1_conditions[] =            { 200,          1000,              2000,             WRITE_FILE,                 SHORT_PRESS,  LONG_PRESS };
void (* S::lgRunBlink1_actions[])() =        { collectOff,   L::pLgRunBlinkOff, L::pLgRunScreen2, L::pLgRunScreenWriteValues, L::pLgPause1, L::p2Countdown };
#define LG_RUN_BLINK_1 (sizeof(lgRunBlink1_conditions) / sizeof(lgRunBlink1_conditions[0]))

// State 'lgRunBlink2' arrays                                off                                  write data                                waited first 1000 ms to detect long press
LcdState* S::lgRunBlink2_connectedStates[] = { &lgRunBlink2, &lgRunBlink2,      &lgRunBlink3,     &lgRunBlink1,               &lgPause1,    &lgRunCountdown };
int S::lgRunBlink2_conditions[] =            { 200,          1000,              2000,             WRITE_FILE,                 SHORT_PRESS,  LONG_PRESS };
void (* S::lgRunBlink2_actions[])() =        { collectOff,   L::pLgRunBlinkOff, L::pLgRunScreen3, L::pLgRunScreenWriteValues, L::pLgPause1, L::p2Countdown };
#define LG_RUN_BLINK_2 (sizeof(lgRunBlink2_conditions) / sizeof(lgRunBlink2_conditions[0]))

// State 'lgRunBlink3' arrays                                off                                  write data                                waited first 1000 ms to detect long press
LcdState* S::lgRunBlink3_connectedStates[] = { &lgRunBlink3, &lgRunBlink3,      &lgRunBlink1,     &lgRunBlink1,               &lgPause1,    &lgRunCountdown };
int S::lgRunBlink3_conditions[] =            { 200,          1000,              2000,             WRITE_FILE,                 SHORT_PRESS,  LONG_PRESS };
void (* S::lgRunBlink3_actions[])() =        { collectOff,   L::pLgRunBlinkOff, L::pLgRunScreen1, L::pLgRunScreenWriteValues, L::pLgPause1, L::p2Countdown };
#define LG_RUN_BLINK_3 (sizeof(lgRunBlink3_conditions) / sizeof(lgRunBlink3_conditions[0]))

// State 'lgPause1' arrays                                                              waited first 1000 ms to detect long press
LcdState* S::lgPause1_connectedStates[] = { &lgPause1,  &lgPause2,    &lgRunBlink1,     &lgPauseCountdown};
int S::lgPause1_conditions[] =            { 200,        2000,         SHORT_PRESS,      LONG_PRESS };
void (* S::lgPause1_actions[])() =        { collectOff, L::pLgPause2, L::pLgRunScreen1, L::p2Countdown };
#define LG_PAUSE_1 (sizeof(lgPause1_conditions) / sizeof(lgPause1_conditions[0]))

// State 'lgPause2' arrays                                                              waited first 1000 ms to detect long press
LcdState* S::lgPause2_connectedStates[] = { &lgPause2,  &lgPause1,    &lgRunBlink1,     &lgPauseCountdown};
int S::lgPause2_conditions[] =            { 200,        2000,         SHORT_PRESS,      LONG_PRESS };
void (* S::lgPause2_actions[])() =        { collectOff, L::pLgPause1, L::pLgRunScreen1, L::p2Countdown };
#define LG_PAUSE_2 (sizeof(lgPause2_conditions) / sizeof(lgPause2_conditions[0]))

// State 'lgRunCountdown' arrays
LcdState* S::lgRunCountdown_connectedStates[] = { &lgRunCountdown, &lgRunCountdown, &lgReady1,          &lgRunBlink1 };
int S::lgRunCountdown_conditions[] =            { 1000,            2000,            2500,               RELEASED };
void (* S::lgRunCountdown_actions[])() =        { L::p1Countdown,  L::p0Countdown,  L::pLgReadyScreen1, L::pLgRunScreen1 };
#define LG_RUN_COUNTDOWN (sizeof(lgRunCountdown_conditions) / sizeof(lgRunCountdown_conditions[0]))

// State 'lgPauseCountdown' arrays
LcdState* S::lgPauseCountdown_connectedStates[] = { &lgPauseCountdown, &lgPauseCountdown, &lgReady1,          &lgPause1 };
int S::lgPauseCountdown_conditions[] =            { 1000,              2000,              2500,               RELEASED };
void (* S::lgPauseCountdown_actions[])() =        { L::p1Countdown,    L::p0Countdown,    L::pLgReadyScreen1, L::pLgPause1 };
#define LG_PAUSE_COUNTDOWN (sizeof(lgPauseCountdown_conditions) / sizeof(lgPauseCountdown_conditions[0]))


int LcdStatesDiagram::buttonShortCondFunction() {
  return SHORT_PRESS;
}

int LcdStatesDiagram::buttonLongCondFunction() {
  return LONG_PRESS;
}

int LcdStatesDiagram::buttonReleasedCondFunction() {
  L::resetOnces();
  return RELEASED;
}

int LcdStatesDiagram::writeCondFunction() {
  return WRITE_FILE;
}


void LcdStatesDiagram::initStates() {
  lgSplash.setArrays(lgSplash_connectedStates, lgSplash_conditions, lgSplash_actions, LG_SPLASH);
  lgWifi.setArrays(lgWifi_connectedStates, lgWifi_conditions, lgWifi_actions, LG_WIFI);
  lgReady1.setArrays(lgReady1_connectedStates, lgReady1_conditions, lgReady1_actions, LG_READY_1);
  lgReady2.setArrays(lgReady2_connectedStates, lgReady2_conditions, lgReady2_actions, LG_READY_2);
  lgReady3.setArrays(lgReady3_connectedStates, lgReady3_conditions, lgReady3_actions, LG_READY_3);
  lgReady4.setArrays(lgReady4_connectedStates, lgReady4_conditions, lgReady4_actions, LG_READY_4);
  lgPwrOff1.setArrays(lgPwrOff1_connectedStates, lgPwrOff1_conditions, lgPwrOff1_actions, LG_PWR_OFF_1);
  lgPwrOff2.setArrays(lgPwrOff2_connectedStates, lgPwrOff2_conditions, lgPwrOff2_actions, LG_PWR_OFF_2);
  lgPwrOff3.setArrays(lgPwrOff3_connectedStates, lgPwrOff3_conditions, lgPwrOff3_actions, LG_PWR_OFF_3);
  lgPwrOff4.setArrays(lgPwrOff4_connectedStates, lgPwrOff4_conditions, lgPwrOff4_actions, LG_PWR_OFF_4);
  lgRunBlink1.setArrays(lgRunBlink1_connectedStates, lgRunBlink1_conditions, lgRunBlink1_actions, LG_RUN_BLINK_1);
  lgRunBlink2.setArrays(lgRunBlink2_connectedStates, lgRunBlink2_conditions, lgRunBlink2_actions, LG_RUN_BLINK_2);
  lgRunBlink3.setArrays(lgRunBlink3_connectedStates, lgRunBlink3_conditions, lgRunBlink3_actions, LG_RUN_BLINK_3);
  lgPause1.setArrays(lgPause1_connectedStates, lgPause1_conditions, lgPause1_actions, LG_PAUSE_1);
  lgPause2.setArrays(lgPause2_connectedStates, lgPause2_conditions, lgPause2_actions, LG_PAUSE_2);
  lgRunCountdown.setArrays(lgRunCountdown_connectedStates, lgRunCountdown_conditions, lgRunCountdown_actions, LG_RUN_COUNTDOWN);
  lgPauseCountdown.setArrays(lgPauseCountdown_connectedStates, lgPauseCountdown_conditions, lgPauseCountdown_actions, LG_PAUSE_COUNTDOWN);
}


void LcdStatesDiagram::fsmExecute() {
  // Check whether button was pressed or released (interrupt)
  if (LcdButton::hasChanged()) {
    // Serial << F("changed\n");   // DEBUG
    // Button 'd'own:
    if (LcdButton::isDown()) {
      Serial << F("down\n");   // DEBUG
      // Start to measure button down time
      LcdButtonTimer::start();
    }
    // Button 'u'p:
    if (LcdButton::isUp()) {
      Serial << F("up\n");   // DEBUG
      // Terminate measuring button down time
      LcdButtonTimer::stop();
      // Serial << LcdButtonTimer::getElapsedTime() << endl;   // DEBUG
      // Avoid reaction on delayed 'u'p if not ...
      if (!LcdButtonTimer::haveElapsedEqualMore(1000)) {
        // Serial << F("less 1000\n");   // DEBUG
        // FSM::printCurrentState();   // DEBUG
        LcdFSM::setConditionFunction(buttonShortCondFunction);
        LcdFSM::nextState();
        LcdFSM::setConditionFunction(LcdFSM::millisCondFunction);
      }
      else {
        // Here are states tested which contain RELEASED events.
        // RELEASED allows to cancel an action which would take place after a long
        // button press in releasing the button before the action time is reached.
        // Serial << F("more than 1000 ms and state with RELEASED condition\n");   // DEBUG
        if (LcdFSM::getCurrentState() == &lgRunCountdown ||
            LcdFSM::getCurrentState() == &lgPauseCountdown ||
            LcdFSM::getCurrentState() == &lgWifi ||
            LcdFSM::getCurrentState() == &lgPwrOff1 ||
            LcdFSM::getCurrentState() == &lgPwrOff2 ||
            LcdFSM::getCurrentState() == &lgPwrOff3 ||
            LcdFSM::getCurrentState() == &lgPwrOff4
          )
        {
          LcdFSM::setConditionFunction(buttonReleasedCondFunction);
          LcdFSM::nextState();
          LcdFSM::setConditionFunction(LcdFSM::millisCondFunction);
        }
      }
    }
  }

  if (LcdButtonTimer::isActive()) {
    // Button is down but not up yet
    if (LcdButtonTimer::haveElapsedEqualMore(1000)) {
      Serial << F("more than 1000 ms\n");   // DEBUG
      LcdFSM::setConditionFunction(buttonLongCondFunction);
      LcdFSM::nextState();
      LcdFSM::setConditionFunction(LcdFSM::millisCondFunction);
      LcdButtonTimer::stop();
    }
  }

  if (LcdLogger::isEnabled()) {
    // Write data to file
    LcdFSM::setConditionFunction(writeCondFunction);
    LcdFSM::nextState();
    LcdFSM::setConditionFunction(LcdFSM::millisCondFunction);
    LcdLogger::disable();
  }

  // Go to next state if condition of condition function is met.
  LcdFSM::nextState();
}


void LcdStatesDiagram::setActive(bool flag) {
  active = flag;
}


bool LcdStatesDiagram::isActive() {
  return active;
}


void LcdStatesDiagram::collectOn() {
  collect = true;
}


void LcdStatesDiagram::collectOff() {
  collect = false;
}


bool LcdStatesDiagram::canCollect() {
  return collect;
}


void LcdStatesDiagram::on() {
  LcdText::wifiSymbol();
}


void LcdStatesDiagram::off() {
  LcdText::wifiNoSymbol();
}


void LcdStatesDiagram::offZeroMillis() {
  off();
  LcdFSM::resetMillis();
}


void LcdStatesDiagram::pwr() {
  LcdText::powerOffSymbol();
}


void LcdStatesDiagram::poweroff() {
  Serial << F("about to power off...\n");
  delay(500);
  LcdText::powerOffText();
  digitalWrite(Config::POWER_HOLD_PIN, LOW);
}
