/*
 * Logger.cpp
 *
 * Created on: 31 Jan 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "Logger.h"
#include <Arduino.h>
#include "LogComposer.h"
#include "Config.h"
#include <SD.h>
#include "Temperature.h"
#include "Turbidity.h"
#include "Conductivity.h"
#include "WbRTC.h"
#include "Commands.h"
#include "Streaming.h"


// Class variables
char Logger::buf[50];
File Logger::root;
Timer Logger::timer = Timer(PSTR("logger"));
String Logger::filename;
File Logger::logfile;
bool Logger::tempOn = false;
bool Logger::conductivityOn = false;
bool Logger::turbidityOn = false;


void Logger::begin(String defaultFilename) {
  setFilename(defaultFilename);
  LogComposer::initHeaders();
  Serial << F("Logger initialised\n");
}


bool Logger::isFilenameCorrect(String filename) {
  int n = filename.indexOf(PSTR("."));
  if (n != -1) {
    // Extension
    String ext = filename.substring(n);
    // Serial << ext << endl << ext.length() << endl;   // DEBUG
    if (ext.length() > 4) {
      Config::outStream->printf(PSTR("error: '%s' extension is > 3 chars long\n"), filename.c_str());
      return false;
    }
  }
  else {
    // No extension
    n = filename.length();
  }
  String name = filename.substring(0, n);
  if (name.length() < 1 || name.length() > 8) {
    Config::outStream->printf(PSTR("error: '%s' is < 1 or > 8 chars long\n"), filename.c_str());
    return false;
  }
  return true;
}


bool Logger::createFile(String filename) {
  int n = filename.indexOf(PSTR("."));
  if (n == -1) {
    n = filename.length();
  }

  // Format: name
  String name = filename.substring(0, n);
  // Format: .ext
  String dotExt = filename.substring(n, n + 4);

  if (n == 8) {
    name = name.substring(0, n - 2);
  }
  else if (n == 7) {
    if (isDigit(name.charAt(n - 1))) {
      name = name.substring(0, n - 1);
      if (isDigit(name.charAt(n - 2))) {
        name = name.substring(0, n - 2);
      }
    }
    else {
      name = name.substring(0, n - 1);
    }
  }
  else if (n > 2) {
    // Strip off up to two trailing digits
    if (isDigit(name.charAt(n - 1))) {
      name = name.substring(0, n - 1);
      if (isDigit(name.charAt(n - 2))) {
        name = name.substring(0, n - 2);
      }
    }
  }

  // Create a file with a new inexistent number < 100 like
  // /name00.ext .. /name09.ext, /name10.ext .. /name99.ext
  int i = 0;
  while (i < 100) {
    filename = name + (i < 10 ? "0" : "") + i + dotExt;
    // Increment for next loop, or to set i > 99
    i++;
    if (!SD.exists(String('/') + filename)) {
      // Found inexistent file name
      break;
    }
    else {
      // Serial << F("File exists: ") << filename << endl;   // DEBUG
      continue;
    }
  }
  // Filename candidate
  Logger::filename = filename;

  // Number out of range?
  if (i > 99) {
    Config::outStream->printf(PSTR("error: '%s' number > 99\n"), filename.c_str());
    return false;
  }

  // Try to create file
  logfile = SD.open(String('/') + filename, FILE_WRITE);
  if (!logfile) {
    // Cannot open: write eror
    Config::outStream->printf(PSTR("error: could not create '%s'\n"), filename.c_str());
    return false;
  }

   // Written, all ok
  logfile.close();
  return true;
}


void Logger::writeFileHeaders() {
  String headersString = LogComposer::getHeadersString();
  writeFile(headersString);
}


String Logger::getFilename() {
  return filename;
}


void Logger::setFilename(String filename) {
  Logger::filename = filename;
}


void Logger::printPartialFile() {
  File file = SD.open(String('/') + filename, FILE_READ);
  if (file) {
    Config::outStream->printf(PSTR("contents of %s:\n"), filename.c_str());

    if (file.size() < 100) {
      // Print whole file if small enough
      while (file.available())   Config::outStream->write(file.read());
      Config::outStream->println(PSTR("eof"));
      file.close();
      return;
    }

    char c;
    int nol = 0;  // number of lines
    int nuc = 0;  // number of chars in line 1
    // Print first 4 lines of file
    while (file.available() && nol < 4) {
      if (nol == 1)   nuc++;
      c = file.read();
      Config::outStream->write(c);
      if (c == '\n')   nol++;
    }

    // Print last 3 lines of file if possible
    long seekPos = file.size() - 4 * nuc;
    if (seekPos > 0L) {
      Config::outStream->println(PSTR("... (last lines)"));
      file.seek(seekPos);
      while (file.available() && file.read() != '\n')   ;

      nol = 0;
      while (file.available() && nol < 3) {
        c = file.read();
        Config::outStream->write(c);
        if (c == '\n')   nol++;
      }
    }
    file.close();
    Config::outStream->println(PSTR("eof"));
  }
  else {
    Config::outStream->printf(PSTR("could not open file %s\n"), filename.c_str());
  }
}


void Logger::printSdCardInfo() {
  uint8_t cardType = SD.cardType();
  if (cardType == CARD_NONE) {
    Config::outStream->println(PSTR("No SD card detected"));
    return;
  }

  Config::outStream->print(PSTR("SD card type: "));
  if (cardType == CARD_MMC)   Config::outStream->println(PSTR("MMC"));
  else if (cardType == CARD_SD)   Config::outStream->println(PSTR("SDSC"));
  else if (cardType == CARD_SDHC)   Config::outStream->println(PSTR("SDHC"));
  else   Config::outStream->println(PSTR("unknown"));
  uint64_t cardSize = SD.cardSize() / (1024 * 1024);
  Config::outStream->printf(PSTR("SD card size: %llu MB\n"), cardSize);
  Config::outStream->printf(PSTR("Total space: %llu MB\n"), SD.totalBytes() / (1024 * 1024));
  Config::outStream->printf(PSTR("Used space: %llu MB\n"), SD.usedBytes() / (1024 * 1024));
}


void Logger::writeFile(String data) {
  File file = SD.open(String('/') + filename, FILE_APPEND);
  if (file) {
    file.println(data);
    file.close();
  }
  else {
    Config::outStream->printf(PSTR("could not write to file %s\n"), filename.c_str());
  }
}


void Logger::writeFileValues() {
  LogComposer::initValues();
  // Current date and time added here
  WbRTC::saveCurrentDateTime();
  LogComposer::addValue(WbRTC::getSavedDateString());
  LogComposer::addValue(WbRTC::getSavedTimeString());

  if (tempOn) {
    LogComposer::addValue(Temperature::helper.readString(false));
  }
  if (conductivityOn) {
    LogComposer::addValue(Conductivity::helper.readString(false));
  }
  if (turbidityOn) {
    LogComposer::addValue(Turbidity::readString(false));
  }

  String valuesString = LogComposer::getValuesString();
  writeFile(valuesString);
}


void Logger::setTempOn(bool flag) {
  tempOn = flag;
}


bool Logger::isTempOn() {
  return tempOn;
}


void Logger::setConductivityOn(bool flag) {
  conductivityOn = flag;
}


bool Logger::isConductivityOn() {
  return conductivityOn;
}


void Logger::setTurbidityOn(bool flag) {
  turbidityOn = flag;
}


bool Logger::isTurbidityOn() {
  return turbidityOn;
}


void Logger::printTempSettings() {
  Config::outStream->print(PSTR("temperature logging on: "));
  Config::outStream->println(tempOn ? PSTR("yes") : PSTR("no"));
}


void Logger::printConductivitySettings() {
  Config::outStream->print(PSTR("conductivity logging on: "));
  Config::outStream->println(conductivityOn ? PSTR("yes") : PSTR("no"));
}


void Logger::printTurbiditySettings() {
  Config::outStream->print(PSTR("turbidity logging on: "));
  Config::outStream->println(turbidityOn ? PSTR("yes") : PSTR("no"));
}


void Logger::handleLogger() {
  if (timer.isRunning()) {
    if (timer.isWithinMeasuringPeriod()) {
      Config::outStream->print(PSTR("data logged on "));
      Commands::printCurrentTimeLn();
      // Write values of selected sensors to log file
      writeFileValues();
    }
  }
}


char* Logger::getDateTime(File file) {
  time_t t= file.getLastWrite();
  struct tm* tmstruct = localtime(&t);
  sprintf(buf, "%02d.%02d.%d  %02d:%02d:%02d",
    tmstruct->tm_mday, tmstruct->tm_mon + 1, tmstruct->tm_year + 1900,
    tmstruct->tm_hour, tmstruct->tm_min, tmstruct->tm_sec
  );
  return buf;
}


void Logger::directory() {
  // List all files on the card
  File root = SD.open("/");
  if (!root) {
    Config::outStream->printf(PSTR("failed to open SD root\n"));
    return;
  }

  File file = root.openNextFile();
  int rootFileCount = 0;
  String filename;
  while (file) {
    if (file.isDirectory()) {
      // Serial << F("  DIR : ") << file.name() << endl;   // DEBUG
    }
    else {
      filename = String(file.name()).substring(1);   // Strip off leading '/'
      Config::outStream->printf(PSTR("%s  %s  %i\n"), filename.c_str(), getDateTime(file), file.size());
      rootFileCount++;
    }
    file = root.openNextFile();
  }
  file.close();
  Config::outStream->printf(PSTR("number of files: %i\n"), rootFileCount);
  root.close();
}
