/*
 * OnePointCalibration.cpp
 *
 * Created on: 15 Apr 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "OnePointCalibration.h"
#include "WbPreferences.h"
#include "Temperature.h"


void OnePointCalibration::init(
  const char* rawId, const char* idealId, const char* calibTempId)
{
  raw = WbPreferences::getFloatValue(rawId, 0.0f);
  ideal = WbPreferences::getFloatValue(idealId, 0.0f);
  calibTemp = WbPreferences::getFloatValue(calibTempId, 0.0f);
}


float OnePointCalibration::voltageFormula(float voltage) {
  return (133.42 * voltage * voltage * voltage -
          255.86 * voltage * voltage +
          857.39 * voltage);
}


float OnePointCalibration::calculateValue(float rawValue) {
  // rawValue is the value read by the sensor and converted from
  // voltage to electrical conductivity, called via Conductivity::readRaw()
  //
  // One-point calibration (example):
  // Buffer solution in bottle with indicated value: ideal
  //   e.g.  1413 us/cm  (@ 25 °C)
  // measured solution value in bottle: raw
  //   e.g.  1390 us/cm  (@ 23 °C)
  //
  //                            ideal
  // ecValueCalib = rawValue * -------
  //                             raw
  //
  //
  // Temperature compensation factor:
  //                               1
  // compensation = --------------------------------------
  //                 1 + 0.02 * (currentTemp - calibTemp)
  //
  float ecValueCalib = rawValue * (ideal / raw);

  // Read current temperature
  float currentTempValue = Temperature::helper.readCalibratedValue();
  // Temperature compensation
  float ecValueCompensated = ecValueCalib / (1.0 + 0.02 * (currentTempValue - calibTemp));
  return ecValueCompensated;
}


float OnePointCalibration::getRaw() {
  return raw;
}


float OnePointCalibration::getIdeal() {
  return ideal;
}


float OnePointCalibration::getCalibTemp() {
  return calibTemp;
}


void OnePointCalibration::writeRaw(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  raw = value;
}


void OnePointCalibration::writeIdeal(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  ideal = value;
}


void OnePointCalibration::writeCalibTemp(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  calibTemp = value;
}
