/*
 * OnePointCalibrationHelper.cpp
 *
 * Created on: 15 Apr 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "OnePointCalibrationHelper.h"
#include "OnePointCalibration.h"
#include "Config.h"
#include "Temperature.h"
#include "WbRTC.h"


void OnePointCalibrationHelper::init(
  String name, const char* rawPrefId,
  const char* idealPrefId, const char* tempPrefId, const char* lastDatePrefId,
  float (*readRawFP)(), String (*formatStringFP)(float value, bool withUnit))
{
  this->name = name;
  this->rawPrefId = rawPrefId;
  this->idealPrefId = idealPrefId;
  this->lastDatePrefId = lastDatePrefId;
  calibration.init(rawPrefId, idealPrefId, tempPrefId);
  this->readRawFP = readRawFP;
  this->formatStringFP = formatStringFP;
  Config::outStream->print(F("All "));
  Config::outStream->print(name);
  Config::outStream->println(F(" values initialised"));
}


String OnePointCalibrationHelper::readDate() {
  return WbPreferences::getDateString(lastDatePrefId);
}


void OnePointCalibrationHelper::printFormattedLn(float value, bool withUnit) {
  String formattedValue = formatStringFP(value, withUnit);
  Config::outStream->println(formattedValue);
}


void OnePointCalibrationHelper::printSensorRaw() {
  Config::outStream->print(F("sensor raw:      "));
  printFormattedLn(calibration.getRaw(), true);
}


void OnePointCalibrationHelper::printSensorIdeal() {
  Config::outStream->print(F("ideal reference: "));
  printFormattedLn(calibration.getIdeal(), true);
}


void OnePointCalibrationHelper::printTemp() {
  Config::outStream->print(F("temperature:     "));
  Config::outStream->print(calibration.getCalibTemp(), 1);
  Config::outStream->println(F(" °C"));
}


void OnePointCalibrationHelper::printCalibValues() {
  Config::outStream->print(name);
  Config::outStream->print(F(" calibration, last changed: "));
  Config::outStream->println(readDate());
  printSensorRaw();
  printSensorIdeal();
  printTemp();
}


void OnePointCalibrationHelper::setCalibValues(float ideal) {
  float raw = readRawFP();
  // Write values
  writeCalib(raw, ideal);
  printSensorRaw();
  printSensorIdeal();
}


float OnePointCalibrationHelper::readCalibratedValue() {
  float rawValue = readRawFP();
  return calibration.calculateValue(rawValue);
}


String OnePointCalibrationHelper::readString(bool withUnit) {
  float idealValue = readCalibratedValue();
  return formatStringFP(idealValue, withUnit);
}


void OnePointCalibrationHelper::writeDate() {
  WbRTC::saveCurrentDateTime();
  String now = WbRTC::savedDay() < 10 ? "0" : "";
  now += WbRTC::savedDay();
  now += F("/");
  now += WbRTC::savedMonth() < 10 ? "0" : "";
  now += WbRTC::savedMonth();
  now += F("/");
  now += WbRTC::savedYear();
  WbPreferences::setDateString(lastDatePrefId, now);
}


void OnePointCalibrationHelper::writeCalibTemp(float temp) {
  calibration.writeCalibTemp(tempPrefId, temp);
  writeDate();
}


void OnePointCalibrationHelper::writeCalib(float raw, float ideal) {
  calibration.writeRaw(rawPrefId, raw);
  calibration.writeIdeal(idealPrefId, ideal);
  writeDate();
}


void OnePointCalibrationHelper::fakesetCalibValues(float raw, float ideal, float temp) {
  // Write values
  writeCalib(raw, ideal);
  printSensorRaw();
  printSensorIdeal();
  writeCalibTemp(temp);
  printTemp();
}
