/*
 * Precalibration.cpp
 *
 * Created on: 9 Jun 2023
 * Author: ruedi.mueller@fhnw.ch
 */

#include "Precalibration.h"
#include "Streaming.h"
#include "Temperature.h"
#include "Conductivity.h"
#include "Turbidity.h"
#include "WbPreferences.h"
#include "Config.h"
#include "WbRTC.h"
#include "Commands.h"


const String Precalibration::PRECALIB_FILE PROGMEM = "/precalib.txt";


void Precalibration::setInitialSensorPrefsCond() {
  // listRootDir(SD, PSTR("/"));   // DEBUG

  if (doesPrecalibFileExist()) {
    precalibFromFile();
    deletePrecalibFile();

    Config::outStream->println(F("endless loop activated: restart now"));
    while (true)   { };
  }
}


void Precalibration::listRootDir(fs::FS &fs, const char* dirname) {
  Serial << F("\nSD root directory: ") << dirname << endl;
  File root = fs.open(dirname);
  if (!root) {
     Serial << F("failed to open SD directory\n");
     return;
   }

  File file = root.openNextFile();
  String filename;
  while (file) {
    if (file.isDirectory()) {
      Serial << F("  DIR : ") << file.name() << endl;
    }
    else {
      // Name
      filename = String(file.name()).substring(1);   // Strip off leading '/'
      Serial << filename << endl;
    }
    file = root.openNextFile();
  }
  file.close();
  root.close();
}


boolean Precalibration::doesPrecalibFileExist() {
  Serial << PRECALIB_FILE << F(" exists: ");
  if (SD.exists(PRECALIB_FILE)) {
    Serial <<  F("yes\n");
    return true;
  }
  else {
    Serial <<  F("no\n");
    return false;
  }
}


void Precalibration::precalibFromFile() {
  Serial << F("Set RTC with compile-time date and time\n");
  WbRTC::setWithPreprocessorDateTime();
  Commands::printCurrentTimeLn();

  Serial << F("Set initial sensor values\n");
  float vals[5];
  int i = 0;
  File file = SD.open(PRECALIB_FILE, FILE_READ);
  if (file) {
    char c;
    while (file.available()) {
      c = file.read();
      if (c == '#') {
        // Jump over comment line
        // Serial << c;   // DEBUG
        while (file.available()) {
          c = file.read();
          // Serial << "." << c;   // DEBUG
          if (c == '\r')   break;
        }
      }
      else if (c == '\n' || c == '\r') {
      // Jump over empty line
      // Serial << "nl\n";   // DEBUG
      }
      else {
        // Get possible value
        // Serial << c;   // DEBUG
        String value = "";
        value += c;
        while (file.available()) {
          c = file.read();
          if (c != '\r') {
            // Serial << c;   // DEBUG
            value += c;
          }
          else {
            value.trim();
            if (value.isEmpty()) {
              // Ignore line having contained only spaces
              break;
            }

            // Convert from String to float
            vals[i++] = atof(value.c_str());
            // Serial << endl;   // DEBUG
            break;
          }
        }
      }
    }
    file.close();

    // Serial << vals[0] << " "  << vals[1] << " "  << vals[2] << " "    // DEBUG
    //       << vals[3] << " "  << vals[4] << endl;   // DEBUG
    
    // Set calibration values
    Temperature::helper.fakesetCalibLowValues(vals[0], vals[0]);
    Temperature::helper.fakesetCalibHighValues(vals[1], vals[1]);

    Conductivity::helper.fakesetCalibValues(vals[2], vals[2], vals[3]);
    WbPreferences::putFloatValue(COND_CALIB_TEMP_STR, vals[3]);   // NOTE: needed for unknown reasons

    WbPreferences::setWifiMethod(
      vals[4] == 0 ? ACCESSPOINT_METHOD : NETWORK_METHOD
    );

    WbPreferences::printPrefs();
  }
  else {
    Serial << F("could not open ") << PRECALIB_FILE << endl;
  }
}


void Precalibration::deletePrecalibFile() {
  SD.remove(PRECALIB_FILE);
  Serial << PRECALIB_FILE << F(" deleted\n");
}
