/*
 * Timer.cpp
 *
 * Created on: 18 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "Timer.h"
#include <Arduino.h>
#include "Config.h"
#include "WbRTC.h"
#include "WbPreferences.h"


char Timer::buf[50];


Timer::Timer(String name) : whereFrom(name), running(false) {
  reset();
}


void Timer::setPrefs() {
  interval = TimeSpan(0, 0, 0, WbPreferences::getTimerMinInterval());
}


void Timer::printRunningError() {
  Config::outStream->print(whereFrom);
  Config::outStream->println(F(" is running; halt first"));
}


void Timer::setStartTimeToday(int hh, int mm) {
  if (!running) {
    startDateTime = DateTime(WbRTC::savedYear(), WbRTC::savedMonth(), WbRTC::savedDay(), hh, mm);
    nextLogDateTime = startDateTime;
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" start: "));
    printDateTime(startDateTime);
  }
  else {
    printRunningError();
  }
}


void Timer::setStartDate(int dd, int mm, int yyyy) {
  if (!running) {
    startDateTime = DateTime(yyyy, mm, dd, startDateTime.hour(), startDateTime.minute());
    nextLogDateTime = startDateTime;
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" start: "));
    printDateTime(startDateTime);
  }
  else {
    printRunningError();
  }
}


void Timer::setStopTimeToday(int hh, int mm) {
  if (!running) {
    stopDateTime = DateTime(WbRTC::savedYear(), WbRTC::savedMonth(), WbRTC::savedDay(), hh, mm);
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" stop: "));
    printDateTime(stopDateTime);
  }
  else {
    printRunningError();
  }
}


void Timer::setStopDate(int dd, int mm, int yyyy) {
  if (!running) {
    stopDateTime = DateTime(yyyy, mm, dd, stopDateTime.hour(), stopDateTime.minute());
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" stop: "));
    printDateTime(stopDateTime);
  }
  else {
    printRunningError();
  }
}


void Timer::setInterval(int mm, int ss) {
  if (!running) {
    if (mm == 0 && ss < Config::TIMER_MIN_INTERVAL) {
      Config::outStream->print(whereFrom);
      Config::outStream->print(F(" interval set to "));
      Config::outStream->print(Config::TIMER_MIN_INTERVAL);
      Config::outStream->println(F(" secs"));
      ss = Config::TIMER_MIN_INTERVAL;
    }
    interval = TimeSpan(0, 0, mm, ss);
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" interval: "));
    printIntervalModif();
  }
  else {
    printRunningError();
  }
}


void Timer::setInterval(int hh, int mm, int ss) {
  if (!running) {
    if (hh == 0 && mm == 0 && ss < Config::TIMER_MIN_INTERVAL) {
      Config::outStream->print(whereFrom);
      Config::outStream->print(F(" interval set to "));
      Config::outStream->print(Config::TIMER_MIN_INTERVAL);
      Config::outStream->println(F(" secs"));
      ss = Config::TIMER_MIN_INTERVAL;
    }
    interval = TimeSpan(0, hh, mm, ss);
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" interval: "));
    printIntervalModif();
  }
  else {
    printRunningError();
  }
}


void Timer::setIntervalDays(int dd) {
  if (!running) {
    interval = TimeSpan(dd, interval.hours(), interval.minutes(), interval.seconds());
    Config::outStream->print(whereFrom);
    Config::outStream->print(F(" interval: "));
    printIntervalModif();
  }
  else {
    printRunningError();
  }
}


void Timer::reset() {
  startDateTime = DateTime(0, 0, 0, 0, 0);
  stopDateTime = DateTime(0, 0, 0, 0, 0);
  nextLogDateTime = startDateTime;
}


void Timer::resetTimes() {
  if (!running) {
    reset();
    printSettings();
  }
  else {
    printRunningError();
  }
}


void Timer::printSettings() {
  Config::outStream->print(whereFrom);
  Config::outStream->print(F(" start:    "));
  printDateTime(startDateTime);
  Config::outStream->print(whereFrom);
  Config::outStream->print(F(" stop:     "));
  printDateTime(stopDateTime);
  Config::outStream->print(whereFrom);
  Config::outStream->print(F(" interval: "));
  printInterval();
  Config::outStream->print(whereFrom);
  Config::outStream->print(F(" running: "));
  Config::outStream->println(running ? "yes" : "no");
}


TimeSpan Timer::readInterval() {
  interval = TimeSpan(0, 0, 0, WbPreferences::getTimerMinInterval());
  return interval;
}


void Timer::writeInterval(TimeSpan ts) {
  WbPreferences::setTimerMinInterval(ts.seconds());
  interval = ts;
}


void Timer::run() {
  WbRTC::saveCurrentDateTime();
  if (startDateTime >= WbRTC::getSavedCurrentDateTime()) {
    Config::outStream->print(whereFrom);
    Config::outStream->println(F(" is now running"));
    running = true;
  }
  else {
    // Avoid start as start time is in the past
    Config::outStream->print(F("start time missed; "));
    Config::outStream->print(whereFrom);
    Config::outStream->println(F(" not running"));
  }
}


void Timer::halt() {
  running = false;
}


bool Timer::isRunning() {
  return running;
}


bool Timer::isWithinMeasuringPeriod() {
  WbRTC::saveCurrentDateTime();
  if (startDateTime > WbRTC::getSavedCurrentDateTime()) {
    // Too early, start later
    return false;
  }
  if (stopDateTime < WbRTC::getSavedCurrentDateTime()) {
    // Stop, time is up
    halt();
    Config::outStream->print(whereFrom);
    Config::outStream->println(F(" halted"));
    return false;
  }

  // First time: nextLogDateTime == startDateTime
  if (nextLogDateTime <= WbRTC::getSavedCurrentDateTime()) {
    // Within limits (inclusive), time to measure
    nextLogDateTime = nextLogDateTime + interval;
    return true;
  }
  else {
    // Within limits, but too early to measure
    return false;
  }
}


void Timer::printDateTime(DateTime dateTime) {
  sprintf(
    buf,
    "%02d/%02d/%4d  %02d:%02d:%02d",
    dateTime.day(), dateTime.month(), dateTime.year(),
    dateTime.hour(), dateTime.minute(), dateTime.second()
  );
  Config::outStream->println(buf);
}


void Timer::printTime(DateTime dateTime) {
  sprintf(
    buf,
    "%02d:%02d:%02d",
    dateTime.hour(), dateTime.minute(), dateTime.second()
  );
  Config::outStream->println(buf);
}


void Timer::printInterval() {
  if (interval.days() != 0) {
    sprintf(
      buf,
      "  %3d days  %02d:%02d:%02d",
      interval.days(),
      interval.hours(), interval.minutes(), interval.seconds()
    );
  }
  else {
    sprintf(
      buf,
      "            %02d:%02d:%02d",
      interval.hours(), interval.minutes(), interval.seconds()
    );
  }
  Config::outStream->println(buf);
}


void Timer::printIntervalModif() {
  if (interval.days() != 0) {
    sprintf(
      buf,
      "%d days  %02d:%02d:%02d",
      interval.days(),
      interval.hours(), interval.minutes(), interval.seconds()
    );
  }
  else {
    sprintf(
      buf,
      "%02d:%02d:%02d",
      interval.hours(), interval.minutes(), interval.seconds()
    );
  }
  Config::outStream->println(buf);
}


DateTime Timer::getStartDateTime() {
  return startDateTime;
}


void Timer::setStartDateTime(DateTime startDateTime) {
  this->startDateTime = startDateTime;
}


DateTime Timer::getStopDateTime() {
  return stopDateTime;
}


void Timer::setStopDateTime(DateTime stopDateTime) {
  this->stopDateTime = stopDateTime;
}


TimeSpan& Timer::getInterval() {
  return interval;
}


void Timer::setInterval(TimeSpan interval) {
  this->interval = interval;
}


// LCD display mode only
void Timer::lcdRun() {
  running = true;
}


// LCD display mode only
void Timer::lcdHalt() {
  running = false;
}


// LCD display mode only
void Timer::setLogDateTimeNow() {
  WbRTC::saveCurrentDateTime();
  nextLogDateTime = WbRTC::getSavedCurrentDateTime();
  interval = TimeSpan(0, 0, 0, WbPreferences::getLcdTimerInterval());
}


// LCD display mode only
bool Timer::isTimeToLog() {
  WbRTC::saveCurrentDateTime();
  if (nextLogDateTime <= WbRTC::getSavedCurrentDateTime()) {
    nextLogDateTime = nextLogDateTime + interval;
    // Time to log
    return true;
  }
  else {
    // Too early to log
    return false;
  }
}
