/*
 * Turbidity.cpp
 *
 * Created on: 27 Jun 2023
 * Author: ruedi.mueller@fhnw.ch
 */

#include "Turbidity.h"
#include <Arduino.h>
#include "Config.h"
#include "Streaming.h"


const float Turbidity::TOO_HIGH = 999.9;
const float Turbidity::TOO_LOW = 0.0;

const float Turbidity::NEVER_USED = 777.7;


// Accuracy of turbidity measurement via Secchi disk: ca. 0.5 cm.
// Die Volt-Werte dazwischen jeweils linear interpolieren:
// visibility (cm)   wbox2 (V)
// ---------------   ---------
// > 121 cm          > 3.8 V
// 121.0             3.8
//  33.5             3.7
//  18.5             3.6
//  13.5             3.5   <--
//  11.0             3.5   <--
//   8.5             3.4
//   7.5             3.3
//   5.5             3.2
//   5.0             3.0
//   3.5             2.8
//   3.0             2.6
//   2.5             2.5
//   2.0             2.1
// < 2 cm            < 2.1 cm
float Turbidity::voltage[]    =  {   3.8,  3.7,  3.6,  3.5,  3.5, 3.4, 3.3, 3.2, 3.0, 2.8, 2.6, 2.5, 2.1 };
//                                highest              ***   ***                                     lowest value
float Turbidity::visibility[] =  { 121.0, 33.5, 18.5, 13.5, 11.0, 8.5, 7.5, 5.5, 5.0, 3.5, 3.0, 2.5, 2.0 };
//                                                       ~~~~  ~~~~
float Turbidity::interpolate[] = {   0.0,  0.0,  0.0, 12.5, 12.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
//                                                       ^^^^  ^^^^
//                                         13.5 + 11.0 / 2 ~= 12.5
const int Turbidity::TURB_TABLE_SIZE = sizeof(voltage) / sizeof(float);

char Turbidity::buf[20];


void Turbidity::begin() {
  // No actions needed
}


// Return array index where the volt value is greater equal the array value
int Turbidity::getIndex(float volt) {
  for (int i = 0; i < TURB_TABLE_SIZE; ++i) {
    if (volt >= voltage[i]) {
      return i;
    }
  }
  // This return is never used, cp. myMap(float)
  return NEVER_USED;
}


// Partially map (interpolate in a linear way) volt to visibility depth (Secchi disk).
// Test if value is too high or too low in which case an out-of-range value is answered.
// Answer a manually interpolated value for cases where consecutive identical voltages
// exist in the voltage array.
// In all other cases interpolate and answer the visibility value.
float Turbidity::myMap(float volt) {
  int id = getIndex(volt);
  // Serial << "volt: " << volt << endl;   // DEBUG
  // Serial << "id: " << id << endl;
  // Serial << "interpolate[id]: " << interpolate[id] << endl;
  // Serial << "volt - voltage[id]: " << volt - voltage[id] << endl;
  // Serial << "visibility[id]: " << visibility[id] << endl;

  if (volt > voltage[0]) {
    // Out of range: turbidity too high
    return TOO_HIGH;
  }
  else if (volt < voltage[TURB_TABLE_SIZE - 1]) {
    // Out of range: turbidity too low
    return TOO_LOW;
  }
  else if (volt == voltage[id] && interpolate[id] != 0.0) {
    // Consecutive voltages with identical values, but different visibilities (resolution issue)
    return interpolate[id];
  }
  else {
    return (volt - voltage[id]) * (visibility[id-1] - visibility[id]) / (voltage[id-1] - voltage[id]) + visibility[id];
  }
}



float Turbidity::readRaw() {
  int turbidityValue = analogRead(Config::TURBIDITY_PIN);
  
  // ESP32 ADC: (12 bits: 0 .. 4095) for voltages (0 .. 3.3V)
  //   
  //   +----------+          100 kOhm      220 kOhm
  //   | Turb.    |          _____         _____
  //   | sensor   |---------|_____|---+---|_____|----GND
  //   |          | x                 |                      +----------+
  //   +----------+ (0..4.5V)         |                      | ESP32    |       Output
  //                                  +----------------------| ADC      |-----> 0..4.5V
  //                                             x*(220/320) |          |
  //                                             0..3.1V     +----------+
  // Sensor output: x [V]
  // After voltage divider and at esp_adc: xV * 220 Ohm / 320 Ohm
  // (  Maximum input at ESP32 ADC: 4.5V * (220 / 320) = 3.09V  )
  // adc_value = x * (220 / 320) * (4095 / 3.3)
  // Mapping to a max of 4095: adc_value * MAPPING_FACTOR
  const float MAPPING_FACTOR = (3.3 * 320) / (5 * 220);

  // CAUTION: voltage set for Arduino compatible reading as documented at
  // https://wiki.dfrobot.com/Turbidity_sensor_SKU__SEN0189
  float voltage = 5.0 * turbidityValue * MAPPING_FACTOR / 4095;   // Arduino compatible reading
  return voltage; 
}


float Turbidity::depth() {
  float voltage = readRaw();
  return myMap(voltage); 
}


String Turbidity::formatString(float turbidity, bool withUnit) {
  // Print number as a positive float:  1.1  6.8  33.4
  // Workaround, because...
  // sprintf(buf, "  %.1f°C", read());  // ... does NOT work in Arduino
  int turbTimesTen = (int)(10 * turbidity);
  int prePoint = turbTimesTen / 10;
  int pastPoint = turbTimesTen % 10;
  sprintf(
    buf, "%i.%i%s",
    prePoint, pastPoint,
    withUnit ? " cm" : ""
  );
  return String(buf);
}


String Turbidity::readString(bool withUnit) {
  float visibility = depth();
  return formatString(visibility, withUnit);
}


void Turbidity::printArray(float* arr) {
  for (int i = 0; i < TURB_TABLE_SIZE; ++i) {
    int turbTimesTen = (int)(10 * arr[i]);
    int prePoint = turbTimesTen / 10;
    int pastPoint = turbTimesTen % 10;
    sprintf(buf, "%3i.%i ", prePoint, pastPoint);
    Config::outStream->print(String(buf));
  }
  Config::outStream->println();
}

void Turbidity::printVoltageArray() {
  printArray(voltage);
}

void Turbidity::printVisibilityArray() {
  printArray(visibility);
}
