/*
 * TwoPointCalibration.cpp
 *
 * Created on: 27 Feb 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "TwoPointCalibration.h"
#include "WbPreferences.h"
#include "Streaming.h"


void TwoPointCalibration::init(
  const char* rawLowId, const char* rawHighId,
  const char* idealLowId, const char* idealHighId)
{
  rawLow = WbPreferences::getFloatValue(rawLowId, 0.0f);
  rawHigh = WbPreferences::getFloatValue(rawHighId, 0.0f);
  idealLow = WbPreferences::getFloatValue(idealLowId, 0.0f);
  idealHigh = WbPreferences::getFloatValue(idealHighId, 0.0f);
}


float TwoPointCalibration::calculateValue(float rawValue) {
  // Two-point calibration:
  //
  //   ideal (i, y)
  //        ^
  //        |
  //     iH -                      .
  //        |
  //     cV -             ?
  //        |
  //     iL -    .
  //        |
  //        |
  //      --+----|--------|--------|-------> raw (r, x)
  //            rL        rV      rH
  //
  //  line:  y = c0 + m * (x - c1)
  //         y = i, x = r
  //         c0 = offset = iL
  //         c1 = rL
  //         m = slope = (iH - iL) / (rH - rL)
  //
  //                   iH - iL
  //  ==>    i = iL + --------- * (r - rL)  ;   if r = rL  then i = iL
  //                   rH - rL              ;   if r = rH  then i = iH
  //
  //  ==>    cV = (rV - rL) * (iH - iL) / (rH - rL) + iL
  //
  float calibratedValue =
    (rawValue - rawLow) * (idealHigh - idealLow) /
    (rawHigh - rawLow) +
    idealLow;
  return calibratedValue;
}


float TwoPointCalibration::getRawLow() {
  return rawLow;
}


float TwoPointCalibration::getRawHigh() {
  return rawHigh;
}


float TwoPointCalibration::getIdealLow() {
  return idealLow;
}


float TwoPointCalibration::getIdealHigh() {
  return idealHigh;
}


void TwoPointCalibration::writeRawLow(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  rawLow = value;
}


void TwoPointCalibration::writeIdealLow(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  idealLow = value;
}


void TwoPointCalibration::writeRawHigh(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  rawHigh = value;
}


void TwoPointCalibration::writeIdealHigh(const char* prefId, float value) {
  WbPreferences::putFloatValue(prefId, value);
  idealHigh = value;
}
