/*
 * TwoPointCalibrationHelper.cpp
 *
 * Created on: 2 Apr 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "TwoPointCalibrationHelper.h"
#include "TwoPointCalibration.h"
#include "Config.h"
#include "WbRTC.h"


void TwoPointCalibrationHelper::init(
  String name, const char* rawLowPrefId, const char* rawHighPrefId,
  const char* idealLowPrefId, const char* idealHighPrefId, const char* lastDatePrefId,
  float (*readRawFP)(), String (*formatStringFP)(float value, bool withUnit))
{
  this->name = name;
  this->rawLowPrefId = rawLowPrefId;
  this->rawHighPrefId = rawHighPrefId;
  this->idealLowPrefId = idealLowPrefId;
  this->idealHighPrefId = idealHighPrefId;
  this->lastDatePrefId = lastDatePrefId;
  calibration.init(rawLowPrefId, rawHighPrefId, idealLowPrefId, idealHighPrefId);
  this->readRawFP = readRawFP;
  this->formatStringFP = formatStringFP;
  
  Config::outStream->print(F("All "));
  Config::outStream->print(name);
  Config::outStream->println(F(" values initialised"));
}


String TwoPointCalibrationHelper::readDate() {
  return WbPreferences::getDateString(lastDatePrefId);
}


void TwoPointCalibrationHelper::printFormattedLn(float value, bool withUnit) {
  String formattedValue = formatStringFP(value, withUnit);
  Config::outStream->print(formattedValue);
  Config::outStream->println();
}


void TwoPointCalibrationHelper::printSensorRawLow() {
  Config::outStream->print(F("sensor raw low:       "));
  printFormattedLn(calibration.getRawLow(), true);
}


void TwoPointCalibrationHelper::printSensorAdcRawLow() {
  Config::outStream->print(F("sensor adc raw low:   "));
  printFormattedLn(calibration.getRawLow(), false);
}


void TwoPointCalibrationHelper::printSensorIdealLow() {
  Config::outStream->print(F("ideal reference low:  "));
  printFormattedLn(calibration.getIdealLow(), true);
}


void TwoPointCalibrationHelper::printSensorRawHigh() {
  Config::outStream->print(F("sensor raw high       "));
  printFormattedLn(calibration.getRawHigh(), true);
}


void TwoPointCalibrationHelper::printSensorAdcRawHigh() {
  Config::outStream->print(F("sensor adc raw high:  "));
  printFormattedLn(calibration.getRawHigh(), false);
}


void TwoPointCalibrationHelper::printSensorIdealHigh() {
  Config::outStream->print(F("ideal reference high: "));
  printFormattedLn(calibration.getIdealHigh(), true);
}


void TwoPointCalibrationHelper::printCalibValues() {
  Config::outStream->print(name);
  Config::outStream->print(F(" calibration, last changed: "));
  Config::outStream->println(readDate());
  printSensorRawLow();
  printSensorIdealLow();
  printSensorRawHigh();
  printSensorIdealHigh();
}


void TwoPointCalibrationHelper::printMixedCalibValues() {
  Config::outStream->print(name);
  Config::outStream->print(F(" calibration, last changed: "));
  Config::outStream->println(readDate());
  printSensorAdcRawLow();
  printSensorIdealLow();
  printSensorAdcRawHigh();
  printSensorIdealHigh();
}


void TwoPointCalibrationHelper::setCalibLowValues(float idealLow) {
  float rawLow = readRawFP();
  // Write low values
  writeCalibLow(rawLow, idealLow);
  printSensorRawLow();
  printSensorIdealLow();
}


void TwoPointCalibrationHelper::setMixedCalibLowValues(float idealLow) {
  float rawLow = readRawFP();
  // Write low values
  writeCalibLow(rawLow, idealLow);
  printSensorAdcRawLow();
  printSensorIdealLow();
}


void TwoPointCalibrationHelper::setCalibHighValues(float idealHigh) {
  float rawHigh = readRawFP();
  // Write high values
  writeCalibHigh(rawHigh, idealHigh);
  printSensorRawHigh();
  printSensorIdealHigh();
}


void TwoPointCalibrationHelper::setMixedCalibHighValues(float idealHigh) {
  float rawHigh = readRawFP();
  // Write high values
  writeCalibHigh(rawHigh, idealHigh);
  printSensorAdcRawHigh();
  printSensorIdealHigh();
}


float TwoPointCalibrationHelper::readCalibratedValue() {
  float rawValue = readRawFP();
  return calibration.calculateValue(rawValue);
}


String TwoPointCalibrationHelper::readString(bool withUnit) {
  float rawValue = readRawFP();
  float idealValue = calibration.calculateValue(rawValue);
  return formatStringFP(idealValue, withUnit);
}


void TwoPointCalibrationHelper::writeDate() {
  WbRTC::saveCurrentDateTime();
  String now = WbRTC::savedDay() < 10 ? "0" : "";
  now += WbRTC::savedDay();
  now += F("/");
  now += WbRTC::savedMonth() < 10 ? "0" : "";
  now += WbRTC::savedMonth();
  now += F("/");
  now += WbRTC::savedYear();
  WbPreferences::setDateString(lastDatePrefId, now);
}


void TwoPointCalibrationHelper::writeCalibLow(float rawLow, float idealLow) {
  calibration.writeRawLow(rawLowPrefId, rawLow);
  calibration.writeIdealLow(idealLowPrefId, idealLow);
  writeDate();
}


void TwoPointCalibrationHelper::writeCalibHigh(float rawHigh, float idealHigh) {
  calibration.writeRawHigh(rawHighPrefId, rawHigh);
  calibration.writeIdealHigh(idealHighPrefId, idealHigh);
  writeDate();
}


void TwoPointCalibrationHelper::fakesetCalibLowValues(float rawLow, float idealLow) {
  // Write low values
  writeCalibLow(rawLow, idealLow);
  printSensorRawLow();
  printSensorIdealLow();
}


void TwoPointCalibrationHelper::fakesetMixedCalibLowValues(float rawLow, float idealLow) {
  // Write low values
  writeCalibLow(rawLow, idealLow);
  printSensorAdcRawLow();
  printSensorIdealLow();
}


void TwoPointCalibrationHelper::fakesetCalibHighValues(float rawHigh, float idealHigh) {
  // Write high values
  writeCalibHigh(rawHigh, idealHigh);
  printSensorRawHigh();
  printSensorIdealHigh();
}


void TwoPointCalibrationHelper::fakesetMixedCalibHighValues(float rawHigh, float idealHigh) {
  // Write high values
  writeCalibHigh(rawHigh, idealHigh);
  printSensorAdcRawHigh();
  printSensorIdealHigh();
}
