/*
 * WbPreferences.cpp
 *
 * Created on: 26 Nov 2021
 * Author: ruedi.mueller@fhnw.ch
 */

#include "WbPreferences.h"
#include <Arduino.h>
#include <Preferences.h>
#include "Config.h"
#include "Temperature.h"
#include "Conductivity.h"
#include "Turbidity.h"

#define READ_WRITE false


Preferences WbPreferences::pref;


void WbPreferences::start() {
  pref.begin("wbox", READ_WRITE);
}


void WbPreferences::printPrefs() {
  Config::outStream->println(F("preferences"));
  Config::outStream->print(F("minimum logging interval: "));
  Config::outStream->print(getTimerMinInterval());
  Config::outStream->println(F(" sec"));
  Config::outStream->print(F("lcd log filename: "));
  Config::outStream->println(getLcdLogfileName());
  Config::outStream->print(F("lcd timer interval: "));
  Config::outStream->print(getLcdTimerInterval());
  Config::outStream->println(F(" sec"));
  Config::outStream->print(F("lcd backlight timeout: "));
  Config::outStream->print(getBacklightTimeout());
  Config::outStream->println(F(" min; < 0: no timeout"));
  printWifi();
  Config::outStream->print(F("Serial1 enabled: "));
  Config::outStream->println(getSerial1Switch() == SERIAL1_ENABLED ? F("yes") : F("no"));

  Config::outStream->println(TEMPERATURE_STR);
  Config::outStream->print(F("  calibration date: "));
  Config::outStream->println(getDateString(TEMP_LAST_DATE_STR));
  Config::outStream->print(F("  raw low: "));
  Config::outStream->println(getFloatValue(TEMP_RAW_LOW_STR, 0.0f));
  Config::outStream->print(F("  ideal low: "));
  Config::outStream->println(getFloatValue(TEMP_IDEAL_LOW_STR, 0.0f));
  Config::outStream->print(F("  raw high: "));
  Config::outStream->println(getFloatValue(TEMP_RAW_HIGH_STR, 0.0f));
  Config::outStream->print(F("  ideal high: "));
  Config::outStream->println(getFloatValue(TEMP_IDEAL_HIGH_STR, 0.0f));

  Config::outStream->println(CONDUCTIVITY_STR);
  Config::outStream->print(F("  calibration date: "));
  Config::outStream->println(getDateString(COND_LAST_DATE_STR));
  Config::outStream->print(F("  raw: "));
  Config::outStream->println(getFloatValue(COND_RAW_STR, 0.0f));
  Config::outStream->print(F("  ideal: "));
  Config::outStream->println(getFloatValue(COND_IDEAL_STR, 0.0f));
  Config::outStream->print(F("  calibration temp: "));
  Config::outStream->println(getFloatValue(COND_CALIB_TEMP_STR, 0.0f));

  Config::outStream->println(TURBIDITY_STR);
  Config::outStream->println(F("  hard coded calibration table:"));
  Config::outStream->print(F("  sensor voltage [V]:\n  "));
  Turbidity::printVoltageArray();
  Config::outStream->print(F("  Secchi disk visibility [cm]:\n  "));
  Turbidity::printVisibilityArray();
}

void WbPreferences::printWifi() {
  printNetCredentials();
  printApCredentials();
  printWifiMethod();
}

void WbPreferences::printNetCredentials() {
  Config::outStream->print(F("network ssid: "));
  Config::outStream->print(getNetSsid());
  Config::outStream->println(F(", network pwd: ?????????"));
}

void WbPreferences::printApCredentials() {
  Config::outStream->print(F("access point ssid: "));
  Config::outStream->print(getApSsid());
  Config::outStream->print(F(", access point pwd: "));
  Config::outStream->println(getApPwd());
}

void WbPreferences::printWifiMethod() {
  Config::outStream->print(F("wifi method [net, ap]: "));
  Config::outStream->print(getWifiMethod() == NETWORK_METHOD ? F("net") : F("ap"));
  Config::outStream->println(F("; method change active only after restart!"));
}


#include <nvs_flash.h>
void WbPreferences::clearPrefs() {
  // https://randomnerdtutorials.com/esp32-save-data-permanently-preferences/
  // You should download a new sketch to your board immediately after running the above,
  // or it will reformat the NVS partition every time it is powered up.
  nvs_flash_erase();   // erase the NVS partition and...
  nvs_flash_init();    // initialize the NVS partition
}


unsigned int WbPreferences::getInt(const char* key, const int defaultValue) {
  start();
  int value = pref.getInt(key, 0);
  if (value == 0) {
    pref.putInt(key, defaultValue);
    value = defaultValue;
  }
  pref.end();
  return value;
}
void WbPreferences::putInt(const char* key, int value) {
  start();
  pref.putInt(key, value);
  pref.end();
}


unsigned int WbPreferences::getUInt(const char* key, const unsigned int defaultValue) {
  start();
  int value = pref.getUInt(key, 0);
  if (value == 0) {
    pref.putUInt(key, defaultValue);
    value = defaultValue;
  }
  pref.end();
  return value;
}
void WbPreferences::putUInt(const char* key, unsigned int value) {
  start();
  pref.putUInt(key, value);
  pref.end();
}


String WbPreferences::getString(const char* key, String defaultStr) {
  start();
  String str = pref.getString(key, "");
  if (str.equals("")) {
    pref.putString(key, defaultStr);
    str = defaultStr;
  }
  pref.end();
  return str;
}
void WbPreferences::putString(const char* key, String str) {
  start();
  pref.putString(key, str);
  pref.end();
}


unsigned int WbPreferences::getTimerMinInterval() {
  return getUInt("minInterval", Config::TIMER_MIN_INTERVAL);
}
void WbPreferences::setTimerMinInterval(unsigned int seconds) {
  putUInt("minInterval", seconds);
}


String WbPreferences::getLcdLogfileName() {
  return getString("lcdFilename", Config::LCD_DEFAULT_LOGFILE_NAME);
}
void WbPreferences::setLcdLogfileName(String filename) {
  putString("lcdFilename", filename);
}


unsigned int WbPreferences::getLcdTimerInterval() {
  return getUInt("lcdInterval", Config::LCD_TIMER_INTERVAL);
}
void WbPreferences::setLcdTimerInterval(unsigned int seconds) {
  putUInt("lcdInterval", seconds);
}


int WbPreferences::getBacklightTimeout() {
  return getInt("lcdTimeout", Config::LCD_TIMER_INTERVAL);
}
void WbPreferences::setBacklightTimeout(int minutes) {
  putInt("lcdTimeout", minutes);
}


String WbPreferences::getNetSsid() {
  return getString("netSsid", "not set");
}
void WbPreferences::setNetSsid(String ssid) {
  putString("netSsid", ssid);
}

String WbPreferences::getNetPwd() {
  return getString("netPwd", "not set");
}
void WbPreferences::setNetPwd(String pwd) {
  putString("netPwd", pwd);
}

String WbPreferences::getApSsid() {
  return getString("apSsid", "wbox-ap");
}
void WbPreferences::setApSsid(String ssid) {
  putString("apSsid", ssid);
}

String WbPreferences::getApPwd() {
  return getString("apPwd", "123456789");
}
void WbPreferences::setApPwd(String pwd) {
  putString("apPwd", pwd);
}


int WbPreferences::getWifiMethod() {
  return getUInt("wifiMethod", NETWORK_METHOD);
}
void WbPreferences::setWifiMethod(int method) {
  putUInt("wifiMethod", method);
}


int WbPreferences::getSerial1Switch() {
  return getUInt("serial1", SERIAL1_DISABLED);
}
void WbPreferences::setSerial1Switch(int enabled) {
  putUInt("serial1", enabled);
}


float WbPreferences::getFloatValue(const char* key, const float defaultValue) {
  start();
  float value = pref.getFloat(key, 0.0f);
  if (value == 0.0f) {
    pref.putFloat(key, defaultValue);
    value = defaultValue;
  }
  pref.end();
  return value;
}
void WbPreferences::putFloatValue(const char* key, float value) {
  start();
  pref.putFloat(key, value);
  pref.end();
}


String WbPreferences::getDateString(const char* id) {
  return getString(id, "no date set");
}
void WbPreferences::setDateString(const char* id, String date) {
  putString(id, date);
}
