/*
 * WbServer.cpp
 *
 * Created on: 29 Mar 2022
 * Author: ruedi.mueller@fhnw.ch
 */

#include "WbServer.h"
#include <Arduino.h>
#include <ESPAsyncWebServer.h>
#include <AsyncTCP.h>
#include <ESPmDNS.h>
#include "Streaming.h"
#include "Config.h"
#include "Interpreter.h"
#include "WbSD.h"
#include "SD.h"
#include "WbPreferences.h"
#include "LcdText.h"
#include "Commands.h"
#include "Logger.h"
#include "Temperature.h"
#include "Conductivity.h"
#include "Turbidity.h"


// Maximum length of webCommand without parameters
#define MAX_WEBCOMMAND_LENGTH 40
// Readings time interval
#define TIMER_DELAY 2000L 


AsyncWebServer WbServer::server(HTTP_PORT);
AsyncWebSocket WbServer::ws(PSTR("/ws"));
IPAddress WbServer::myIP;
String WbServer::webCommand = "";
String WbServer::webpage = "";
StopWatch WbServer::stopWatch;
bool WbServer::active = false;
AsyncEventSource WbServer::events("/events");
bool WbServer::readingsPageActive = false;
String WbServer::temp;
String WbServer::cond;
String WbServer::turb;
unsigned long WbServer::lastReading = 0;


void WbServer::begin() {
  initWiFiMethod();
  // Add found IP to LCD display
  LcdText::printIP(myIP.toString());

  initSPIFFS();
  initWebSocket();
  initWebServer();
  stopWatch.start();
}


void WbServer::initWiFiMethod() {
  if (WbPreferences::getWifiMethod() == ACCESSPOINT_METHOD) {
    Serial << F("Setting soft-AP: ");
    bool result = 
      WiFi.softAP(WbPreferences::getApSsid().c_str(), WbPreferences::getApPwd().c_str());
    if (result == true) {
      Serial << F("ready\n");
    }
    else {
      Serial << F("failed!\n");
    }
    myIP = WiFi.softAPIP();
    Serial << F("AP IP address: ") << myIP << endl;
    active = true;
  }
  else if (WbPreferences::getWifiMethod() == NETWORK_METHOD) {
    Serial << F("Joining network: ");
    WiFi.begin(WbPreferences::getNetSsid().c_str(), WbPreferences::getNetPwd().c_str());
    // Serial <<  WbPreferences::getNetPwd() << endl;   // DEBUG
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial << F(".");
      LcdText::printRotatePattern();
    }
    myIP = WiFi.localIP();
    Serial << F("\nConnected to ") << WbPreferences::getNetSsid() << F(", IP address: ") << WiFi.localIP() << endl;
    active = true;
  }
  else {
    Serial << F("WiFi method problem. Check with 'wifi' command from Serial Monitor.\n");
  }
}


void WbServer::initSPIFFS() {
  if (!SPIFFS.begin()) {
    Serial << F("cannot mount SPIFFS volume...\n");
  }
  else {
    Serial << F("SPIFFS volume mounted\n");
  }
}


void WbServer::onRootRequest(AsyncWebServerRequest* request) {
  request->send(SPIFFS, PSTR("/index.html"), PSTR("text/html"), false);
}


// ======= DIRECTORY PAGE =======================================================================================
const char WbServer::tablestart_html[] PROGMEM = 
R"rawliteral(<table align='center'>
  <tr>
    <th>Filename</th>
    <th class='cn' style='width: 30'>Date, Time</th>
    <th class='sz' style='width: 15'>Size</th>
    <th class='cn' style='width: 19'>Downlad File</th>
    <th class='cn' style='width: 15'>Delete File</th>
  </tr>
)rawliteral";


void WbServer::addDateTime(File file) {
  char buf[50];
  time_t t= file.getLastWrite();
  struct tm* tmstruct = localtime(&t);
  sprintf(buf, "%02d.%02d.%d%s%02d:%02d:%02d",
    tmstruct->tm_mday, tmstruct->tm_mon + 1, tmstruct->tm_year + 1900, "&nbsp;&nbsp;&nbsp;",
    tmstruct->tm_hour, tmstruct->tm_min, tmstruct->tm_sec
  );

  webpage += F("<td class='cn'>");
  webpage +=                       String(buf) + F("</td>");
}


void WbServer::addConfirmFunction(String filename, int num) {
  webpage += F("<script>");
  webpage += F("function myConfirm");
  webpage +=                          String(num) + F("() {");
  webpage += F(  "if (confirm('Do you really want to delete ");
  webpage +=                                                    filename + F(" ?')) {");
  webpage += F(    "myDelete('");
  webpage +=                      filename + F("');");
  webpage += F(  "}");
  webpage += F("}");
  webpage += F("</script>\n");
}


void WbServer::listRootDir(fs::FS &fs, const char* dirname) {
  // Serial << F("\nListing directory: ") << dirname << endl;   // DEBUG
  File root = fs.open(dirname);
  if (!root) {
     Serial << F("failed to open directory\n");
     return;
   }

  File file = root.openNextFile();
  String filename;
  int num = 0;
  while (file) {
    if (file.isDirectory()) {
      // Serial << F("  DIR : ") << file.name() << endl;   // DEBUG
    }
    else {
      // Serial << filename << F(" ");   // DEBUG
      // Name
      filename = String(file.name()).substring(1);   // Strip off leading '/'
      webpage += F("  <tr><td>");
      webpage +=            filename + F("</td>");
      // Date, time
      addDateTime(file);
      // Size
      unsigned long filesize = file.size();
      webpage += F("<td class='sz'>");
      webpage +=            String(filesize) + F("</td>");
      // Download buttons
      webpage += F("<td class='cn'><form name='parameters' method='GET' action='/download' target='_parent'>");
      webpage += F("<input type='hidden' name='fn' value='");
      webpage +=                                           filename + F("'>");
      webpage += F(  "<input type='submit' name='submit' value='Download'>");
      webpage += F("</form></td>");
      // Delete Buttons
      webpage += F("<td class='cn'>");
      webpage += F("<input type='button' onclick='myConfirm");
      webpage +=                                               String(num) + F("()' value='Delete' />");
      webpage += F("</td></tr>");
      
      addConfirmFunction(filename, num);
      num++;
    }
    file = root.openNextFile();
  }
  file.close();
  webpage += F("</table><br>");
  webpage += F("<h4>Number of files found on SD card: ");
  webpage += String(num);
  webpage += F("</h4>");
  root.close();
}


// Replace placeholder in the web page.
// Only executed if webpage is refreshed.
String WbServer::processor(const String& var) {
  // Serial << var << endl;   // DEBUG
  if (var == F("DIR_PLACEHOLDER")) {
    webpage = tablestart_html;
    listRootDir(SD, PSTR("/"));
    return String(webpage.c_str());
  }
  if (var == F("URL_PLACEHOLDER")) {
    return myIP.toString();
  }
  return String();
}


void WbServer::fileDownload(AsyncWebServerRequest* request) {
  // <ESP_IP>/download?fn=<filename>&submit=<download>
  if (request->hasParam(PSTR("fn")) && request->hasParam(PSTR("submit"))) {
    String filename = "/" + request->getParam(PSTR("fn"))->value();
    // Serial << "File to download: " << filename << endl;   // DEBUG
    File download = SD.open(filename);
    AsyncWebServerResponse *response = request->beginResponse(SD, filename, String(), true);
    request->send(response);
    download.close();
    return;
  }
  else {
    Serial << F("Params 'fn' and/or 'submit' not found\n");
  }
}


void WbServer::fileDelete(AsyncWebServerRequest* request) {
  // <ESP_IP>/delete?df=<filename>
  if (request->hasParam(PSTR("df"))) {
    String filename = "/" + request->getParam(PSTR("df"))->value();
    // Serial << F("File to delete: ") << filename << endl;   // DEBUG
    if (SD.exists(filename)) {
      SD.remove(filename);
      Serial << F("Deleted file ") << filename << F(" on SD card\n");
    }
    else {
      Serial << F("File '") << filename << F("' not found\n");
    }
    // Show table without deleted file
    Serial << F("update file table\n");
    request->redirect(PSTR("/dir"));
    Serial << F("update done\n");
  }
  else {
    Serial << F("Param 'df' not found\n");
  }
}


// ======= TERMINATE ==== =======================================================================================
const char WbServer::terminate_html[] PROGMEM = 
R"rawliteral(<!DOCTYPE html>
<html>
<head>
<title>Waterbox Command and File Server</title>
<style>
  body{max-width:47%;margin:0 auto;font-family:arial;font-size:105%;text-align:center;color:blue;background-color:#40C384;}
  h2,p{font-family:arial;}
</style>
</head>
<body>
  <h2>Waterbox Connection Terminated</h2>
  <p>Restart Waterbox and rejoin Access Point or Network.</p>
  <p>Afterwards reconnect with <a href='http://wbox.local'>wbox.local</a> or using this <a href='http://%URL_PLACEHOLDER%'>link</a></p>
  <p>(Do not activate location bar.)</p>
  <br><br><br><br><br><br>
  <hr>
  <p>Waterbox Project, <a href='https://www.wassernetz.ch/'>wassernetz.ch</a>, developed by <a href='https://makerstudio.fhnw.ch/'>Maker Studio FHNW</a>, 2021-2023</p>
</body>
</html>)rawliteral";


void WbServer::terminate(AsyncWebServerRequest* request) {
  // <ESP_IP>/terminate
  webpage = "";
  request->send_P(200, PSTR("text/html"), terminate_html);
  webpage = "";
}


// ======= PAGE NOT FOUND =======================================================================================
void WbServer::pageNotFound(AsyncWebServerRequest* request) {
  webpage =  F("Page not found\n\n");
  webpage += F("URI: ");
  webpage += request->url();
  webpage += F("\nMethod: ");
  webpage += (request->method() == HTTP_GET) ? F("GET") : F("POST");
  webpage += F("\nArguments: ");
  webpage += request->args();
  webpage += F("\n");
  for (uint8_t i = 0; i < request->args(); i++) {
    webpage += F(" ");
    webpage += request->argName(i) + F(": ") + request->arg(i) + F("\n");
  }
  request->send_P(404, PSTR("text/plain"), webpage.c_str());
}
// ==============================================================================================================


void WbServer::initWebServer() {
  // Server Commands
  server.on("/", onRootRequest);
  // Show directory page
  server.on(PSTR("/dir"), HTTP_GET, [](AsyncWebServerRequest* request) {
    readingsPageActive = false;
    Serial << F("about to send show dir\n");

    // Logger::directory();  // DEBUG

    request->send(SPIFFS, PSTR("/dir.html"), PSTR("text/html"), false, processor);
    Serial << F("show dir sent\n");
  });
  server.serveStatic("/", SPIFFS, "/");
  // Download file
  server.on(PSTR("/download"), HTTP_GET, [] (AsyncWebServerRequest* request) {
    readingsPageActive = false;
    fileDownload(request);
  });
  // Delete file if confirmed
  server.on(PSTR("/delete"), HTTP_GET, [] (AsyncWebServerRequest* request) {
    readingsPageActive = false;
    fileDelete(request);
  });
  // Terminate session
  server.on(PSTR("/terminate"), HTTP_GET, [] (AsyncWebServerRequest* request) {
    readingsPageActive = false;
    // request->send(SPIFFS, PSTR("/terminate.html"), PSTR("text/html"), false, processor);
    request->send_P(200, PSTR("text/html"), terminate_html, processor);
    Serial << F("about to power off via net/ap\n");
    Commands::toPowerOff = true;
  });
  // Hook to SD card: file must be present on SD card
  server.on(PSTR("/wbox"), HTTP_GET, [](AsyncWebServerRequest* request) {
    readingsPageActive = false;
    request->send(SD, PSTR("/wbox/entry.html"), PSTR("text/html"));
  });
  server.onNotFound(pageNotFound);

  // Sensor readings
  server.on("/readings", HTTP_GET, [](AsyncWebServerRequest* request) {
    readingsPageActive = true;
    request->send_P(200, PSTR("text/html"), readings_html, readingsProcessor);
    // Serial << F("show /readings page\n");
  });
  onEventConnect();
  server.addHandler(&events);

  server.begin();
  Serial << F("web server started\n");

  // Set up mDNS responder
  if (!MDNS.begin("wbox")) {
    Serial << F("Error setting up mDNS responder!\n");
    Serial << F("mDNS responder not started\n");
  }
  else {
    Serial << F("mDNS responder started\n");
  }

  // Set up mDNS responder
  MDNS.addService(PSTR("http"), PSTR("tcp"), 80);
  Serial << F("mDNS service started\n");
}


// Define an event listener to handle the different asynchronous steps of
// the WebSocket protocol.
void WbServer::onEvent(AsyncWebSocket* server, AsyncWebSocketClient* client,
             AwsEventType type, void* arg, uint8_t* data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT:
      // Client has logged in
      Serial << F("web socket client ") << client->id() << F(" connected from ")
             << client->remoteIP().toString().c_str() << endl;
      break;
    case WS_EVT_DISCONNECT:
      // Client has logged out
      Serial << F("web socket client ") << client->id() << F(" disconnected\n");
      break;
    case WS_EVT_DATA:
      // Receive data packet from the client, here via JS onKeyDown()
      handleWebSocketMessage(arg, data, len);
      break;
    case WS_EVT_PONG:
      // Response to a ping request
    case WS_EVT_ERROR:
      // Error is received from the client
      break;
  }
}


void WbServer::handleWebSocketMessage(void* arg, uint8_t* data, size_t len) {
  AwsFrameInfo* info = (AwsFrameInfo*)arg;
  // Ensure that the data are received in full and in one piece, and that they
  // correspond to textual data.
  if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
    char buf[MAX_WEBCOMMAND_LENGTH + 1];
    memcpy(buf, data, len);
    buf[len] = '\0';
    webCommand = String(buf);
    // Serial << "webCommand: " << webCommand << endl;   // DEBUG
  }
}


// Send data to web socket clients
void WbServer::notifyClients() {
  // Send the same message to all clients
  if (Config::pb.isFull()) {
    ws.textAll(PSTR("error: printbuffer full"), 23);   // length: 23
  }
  else {
    ws.textAll(Config::pb(), Config::pb.length());
  }
  // Serial << "Config::pb(): " << String(Config::pb()) << endl;         // DEBUG
  // Serial << "Config::pb.length(): " << Config::pb.length() << endl;   // DEBUG
}


void WbServer::initWebSocket() {
  ws.onEvent(onEvent);
  server.addHandler(&ws);
  Serial << F("web socket started\n");
}


void WbServer::cleanupClients() {
  if (stopWatch.hasReached(1000L)) {   // 1000 millis
    stopWatch.start();                 // restart
    ws.cleanupClients();
    // Serial << "restart\n";   // DEBUG
  }
}


void WbServer::interpret() {
  // It may happen that the connection is lost for various reasons, or closed
  // intentionally. Browsers sometimes do not correctly close the WebSocket
  // connection, even when the close() function is called in Javascript. This
  // will eventually exhaust the web server’s resources and will cause the server
  // to crash. 
  // To avoid this, the AsyncWebSocket class provides an efficient cleaning method:
  // cleanupClients(). Periodically calling the cleanClients() function from the
  // main loop() limits the number of clients by closing the oldest client when
  // the maximum number of clients has been exceeded.
  // This can be called in every cycle, however, if you wish to use less power, then
  // calling as infrequently as once per second is sufficient.
  cleanupClients();

  if (webCommand != "") {
    // Pass command to Interpreter
    Interpreter::setInput(webCommand);
    // which in turn executes command and collects answer in PrintBuffer pb.
    // This is the text which will be displayed in the textarea.
    Config::pb.clear();
    Config::outStream = &Config::pb;
    Interpreter::interpret();

    // Update AND notify clients
    notifyClients();
    webCommand = "";
    Config::pb.clear();
  }

  // Display messages like monitor state updates etc.
  if (Config::pb.length() > 0) {
    ws.textAll(Config::pb(), Config::pb.length());
    Config::pb.clear();
  }
}


// ======= SENSOR READINGS =====================================================================================
const char WbServer::readings_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html>
<head>
  <title>Watebox Web Server</title>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <style>
    html {font-family: arial; text-align: center; color: blue; display: inline-block;}
    p {font-size: 1.2rem;}
    body {margin: 0;}
    .topnav {overflow: hidden; background-color: #40C384; color: blue; font-size: 1.7rem;}
    .content {padding: 20px;}
    .card {background-color: white; box-shadow: 2px 2px 12px 1px rgba(140,140,140,.9);}
    .cards {max-width: 250px; margin: 0 auto; display: grid; grid-gap: 2rem; grid-template-columns: repeat(auto-fit, minmax(300px, 1fr));}
    .reading {font-size: 2.8rem;}
    .card.sensor {color: blue;}
  </style>
</head>
<body>
  <div class="topnav">
    <h3>Waterbox Sensor Readings</h3>
  </div>
  <div class="content">
    <div class="cards">
      <div class="card sensor">
        <h4><i class="fas sensors"></i> TEMPERATURE</h4>
        <p><span class="reading"><span id="tempId">%TEMP_PLACEHOLDER%</span> &deg;C</span></p>
      </div>
      <div class="card sensor">
        <h4><i class="fas sensors"></i> CONDUCTIVITY</h4>
        <p><span class="reading"><span id="condId">%COND_PLACEHOLDER%</span> &micro;S/cm</span></p>
      </div>
      <div class="card sensor">
        <h4><i class="fas sensors"></i> TURBIDITY</h4>
        <p><span class="reading"><span id="turbId">%TURB_PLACEHOLDER%</span> cm</span></p>
      </div>
    </div>
  </div>
    <br><br><br>
  <hr>
  <p>Waterbox Project, <a href='https://www.wassernetz.ch/'>wassernetz.ch</a>, developed by <a href='https://makerstudio.fhnw.ch/'>Maker Studio FHNW</a>, 2021-2023</p>
<script>
if (!!window.EventSource) {
  var source = new EventSource('/events');

  source.addEventListener('open', function(e) {
   console.log("Events connected");
  }, false);
  source.addEventListener('error', function(e) {
    if (e.target.readyState != EventSource.OPEN) {
      console.log("Events disconnected");
    }
  }, false);
  source.addEventListener('message', function(e) {
    console.log("message", e.data);
  }, false);
  source.addEventListener('tempId', function(e) {
    console.log("temperature", e.data);
    document.getElementById("tempId").innerHTML = e.data;
  }, false);
  source.addEventListener('condId', function(e) {
    console.log("conductivity", e.data);
    document.getElementById("condId").innerHTML = e.data;
  }, false);
  source.addEventListener('turbId', function(e) {
    console.log("turbidity", e.data);
    document.getElementById("turbId").innerHTML = e.data;
  }, false);
}
</script>
</body>
</html>)rawliteral";


String WbServer::readingsProcessor(const String& var) {
  Serial.println(var);
  if (var == "TEMP_PLACEHOLDER") {
    return temp;   // Value must be passed from main loop!!
  }
  if (var == "COND_PLACEHOLDER") {
    return cond;
  }
  if (var == "TURB_PLACEHOLDER") {
    return turb;
  }
  return String();
}


void WbServer::onEventConnect() {
    // Handle web server events
  events.onConnect([](AsyncEventSourceClient* client) {
    if (client->lastId()) {
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    // Send event with message "hello!", id current millis
    // and set reconnect delay to 10 seconds
    client->send("hello!", NULL, millis(), 10000);
  });
}


void WbServer::sendEvent(String sensorReading, String id, long millis) {
  events.send("ping", NULL, millis);
  events.send(sensorReading.c_str(), id.c_str(), millis);
}


void WbServer::updateReadingsIfActive() {
  if (WbServer::readingsPageActive) {
    if ((millis() - lastReading) > TIMER_DELAY) {
      temp = Temperature::helper.readString(false); 
      sendEvent(temp, "tempId", millis());

      cond = Conductivity::helper.readString(false);
      sendEvent(cond, "condId", millis());      

      turb = Turbidity::readString(false);

      sendEvent(turb, "turbId", millis());

      lastReading = millis();
    }
  }
}
