/*
 * Target: ESP32 Huzzah Feather (Adafruit)
 * 
 * DATA LOGGER "WATERBOX"
 * This is the main program and entry point for the 'waterbox' project.
 * The project is maintained under Visual Studio Code.
 *
 * The waterbox project consists in its core of an ESP32 board, an adalogger
 * board (SD and RTC), a custom-made pcb, an LCD display and button, and 3
 * sensors (temperature, electrical conductivity and turbidity).
 * 
 * The 'waterbox' provides two methods to measure data:
 * (A) Displaying/logging measured data via commands (Serial, Serial1, WiFi)
 * (B) Displaying measured data on LCD display, logging invoked via button
 *
 * Method (A)
 * Using the commandline (Serial Monitor, PuTTY, ...):
 * Type  help  to show more help commands. From there discover more commands.
 * Consult documentation.
 *
 * Method (B)
 * Using LCD display and button:
 * After reset the splash screen is shown on the LCD display. With a button
 * press (before any other communication via Serial) the lcd display is
 * animated and button presses will be answered.
 * Consult separate document.
 *
 * History
 * 27 Jun 2023  rm   New turbidity measurement and display (Secchi disk)
 * 19 Jun 2023  rm   SD hookup button on website (optically) removed
 * 19 Jun 2023  rm   SD card precalib.txt used to set date/time and initial sensor values
 *  1 May 2023  rm   sw under new version control
 * 16 Jan 2023  rm   hardware change: LCD display with I2C
 * 19 Dec 2022  rm   lcd button blocking after logging stop
 * 30 Mar 2022  rm   revamped server architecture
 *  9 Jan 2022  rm   integrated async web server and socket
 *  1 Dec 2021  rm   ported from Arduino MEGA project
 */

#include <Arduino.h>
#include "WbSD.h"
#include "Config.h"
#include "WbPreferences.h"
#include "Commands.h"
#include "Interpreter.h"
#include "InputStreamReader.h"
#include "WbServer.h"
#include "WbRTC.h"
#include "WbLCD.h"
#include "LcdButton.h"
#include "LcdLogger.h"
#include "LcdFSM.h"
#include "LcdStatesDiagram.h"
#include "Temperature.h"
#include "Turbidity.h"
#include "Conductivity.h"
#include "Logger.h"
#include "Monitor.h"
#include "Streaming.h"
#include "Precalibration.h"


void setup() {
  Config::setPinMode();

  // SERIAL
  InputStreamReader::begin('\n');
  Serial.begin(115200);
  delay(300);
  // VERSION: Output always to Serial
  Commands::printVersionInstitution();

  // BUTTON: init button interrupt etc.
  LcdButton::begin();

  // LCD DISPLAY, BACKLIGHT ON AND SYSTEM POWER HOLD
  WbLCD::begin();

  // SERIAL1
  if (WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) {
    Serial << F("Serial1 enabled. Make sure Serial1 is connected.") << endl;
    Serial1.begin(115200);
  }
  else   Serial << F("Serial1 disabled") << endl;

  // SD card
  WbSD::begin();

  // RTC (Set up Wire library)
  if (!WbRTC::begin()) {
    // No error message as SD card and Wire integrated integrated
    // in one hardware unit: adalogger
  }
  Serial << F("Current RTC time: ");
  Commands::printCurrentTimeLn();

  // LCDLOGGER: init logger and fsm
  LcdLogger::begin();

  // LOGGER (for activation via commandline the following different logger is used)
  Logger::begin(Config::DEFAULT_LOGFILE_NAME);

  // SENSOR INITIALISATION
  Temperature::begin();
  Conductivity::begin();
  Turbidity::begin();

  // Check if initial sensor values have to be set
  Precalibration::setInitialSensorPrefsCond();

  // CONNECTION TO SERIAL, SERIAL1 OR LCD: first come first serve
  Serial.flush();
  if (WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) {
    Serial1.flush();
  }

  WbLCD::backlightOn();  // Set timeout timer
  // Wait for input
  while (true) {
    // Test if LCD button was pressed...
    if (LcdButton::hasChanged()) {
      // Enable FSM
      LcdStatesDiagram::setActive(true);
      Commands::printConnectedTo(F("LCD display and button"));
      LcdFSM::resetMillis();
      break;
    }
    // ...or Serial active...
    else if (Serial.available() > 0) {
      Commands::setSerial();
      break;
    }
    // ...or Serial1 active (if enabled)
    if ((WbPreferences::getSerial1Switch() == SERIAL1_ENABLED) && Serial1.available() > 0) {
      Commands::setSerial1();
      break;
    }
    WbLCD::backlightCheck();
  }
}


// NOTE for LCD button mode:
// loop is reached as soon as the LCD button is pressed (and therefore touched).
void loop() {
  // Handle LCD display and button presses
  LcdLogger::handleLcdDisplayMode();
  
  // Command interpreter (Serial, Serial1)
  Interpreter::readAndInterpret();

  // Logger if enabled
  Logger::handleLogger();

  // Handle monitor if enabled
  Monitor::handleMonitor();

  // WiFi mode
  // WiFi can be invoked in LcdText::pLgWifiScreen()
  // by pressing the button > 10 seconds.
  WbServer::interpret();
  // Waterbox in WiFi mode can be powered off by pressing button > 10 seconds.
  LcdButton::powerOffIfWbServerActive();
  // Send update to /readings webpage if active
  WbServer::updateReadingsIfActive();

  // Test if power off is pending evoked in LCD display mode
  if (Commands::toPowerOff) {
    LcdStatesDiagram::poweroff();
  }
}
